package algorithm;

import java.util.Arrays;
import java.util.BitSet;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.lang3.mutable.MutableInt;

import algorithm.dataStructures.CompactDegreeSequence;
import algorithm.dataStructures.Triple;
import benchmarks.MainCLIAnonymizerAndAlternativeRealizer;

public class CostWaster
{

	public static final MutableInt COSTWASTER_UPPERBOUND_NO_SOLUTIONS = new MutableInt(2);

	private static final long DYN_PROG_TABLE_UPPER_BOUND = 1000 * 1000;

	/**
	 * 
	 * Returns the number of trivial degree one increases which are possible in any (optimal) outcome of the dynamic programming.
	 * 
	 * @param minAddCost
	 * @param k
	 * @return
	 */
	public static boolean canYouWasteDegreeOneVertices(final CompactDegreeSequence seq, final int minAddCost, final int k)
	{
		final int f = (2 * k) - 1;
		final boolean[] remainingBlocks = new boolean[seq.getMaxDegree() + 1];

		for (int i = 0; i <= seq.getMaxDegree(); i++)
		{
			if (i == seq.getMaxDegree())
			{
				remainingBlocks[i] = seq.get(i) >= k;
			} else
			{
				remainingBlocks[i] = (seq.get(i) >= f) || ((seq.get(i + 1) >= f) && (seq.get(i) >= k));
			}
		}

		int c = 0;

		for (int i = 0; i < seq.getMaxDegree(); i++)
		{
			if ((seq.get(i) >= f) && remainingBlocks[i + 1])
			{
				c += seq.get(i) - f;
			}
			if (c >= minAddCost)
				return true;
		}
		return false;
	}

	private CompactDegreeSequence anonSequence;

	private final int cost;

	private final int costSoFar;

	private final boolean findRealizableSequence;

	private boolean foundRealizable = false;

	private final int k;

	private int maxDegree;

	private final int maxFoundSolutions; // 500

	private final CompactDegreeSequence originalSequence;

	private final int[] solIncrease;

	private Set<CompactDegreeSequence> solutions = null;

	// public CostWaster(CompactDegreeSequence originalSequence, Triple[]
	// triples,
	// CompactDegreeSequence anonSequence, int additionalCosts, int k)
	// {
	// this.k = k;
	// this.originalSequence = originalSequence;
	// this.anonSequence = anonSequence;
	// this.triples = triples;
	// this.cost = additionalCosts;
	// this.maxDegree = anonSequence.getMaxValue();
	// }

	private BitSet[][] table;

	public CostWaster(final CompactDegreeSequence compactDegreeSequence, final CompactDegreeSequence anonSequence, final int addCost, final int k)
	{
		this(compactDegreeSequence, anonSequence, addCost, k, true, CostWaster.COSTWASTER_UPPERBOUND_NO_SOLUTIONS.intValue());
	}

	public CostWaster(final CompactDegreeSequence compactDegreeSequence, final CompactDegreeSequence anonSequence, final int addCost, final int k,
			final boolean findRealizableSequence, final int maxStoredSolutions)
	{
		this.k = k;
		this.originalSequence = compactDegreeSequence;
		this.anonSequence = anonSequence;
		this.costSoFar = anonSequence.difference(compactDegreeSequence).getTotalDegreeSum();
		this.maxDegree = anonSequence.getMaxNonZeroDegree();
		this.solIncrease = new int[this.maxDegree + 1];
		this.solIncrease[0] = compactDegreeSequence.get(0) - anonSequence.get(0);
		for (int i = 1; i < this.maxDegree; i++)
		{
			this.solIncrease[i] = (compactDegreeSequence.get(i, 0) - anonSequence.get(i)) + this.solIncrease[i - 1];
		}
		this.solIncrease[this.maxDegree] = 0;

		// this.triples = triples;
		this.cost = addCost;
		this.findRealizableSequence = findRealizableSequence;
		this.solutions = new HashSet<CompactDegreeSequence>();
		this.maxFoundSolutions = maxStoredSolutions;
	}

	/**
	 * Heuristic that increases the degrees by more than one. Tries to keep the maximum degree-increase small.
	 * 
	 * Runs in O(maxDegree^2) time.
	 * 
	 * @return
	 */
	private int determineBalancedIncreasesHeuristic(final boolean constructSolution)
	{
		final Triple[] wasteTriples = new Triple[this.maxDegree + 1];

		int maxIncreasedDegrees = 0;

		for (int degIncLevel = 2; degIncLevel < (this.maxDegree - 1); degIncLevel++)
		{
			for (int i = 0; i < wasteTriples.length; i++)
			{
				wasteTriples[i] = new Triple(this.maxDegree, 0, 0);
			}

			int increasedDegrees = 0; // number of vertices that might be increased
										// by degInc without destroying one block
			int remainingIncrease = 0;
			for (int i = 0; i <= (this.maxDegree - degIncLevel); i++)
			{
				remainingIncrease = this.cost - increasedDegrees;
				if (remainingIncrease == 0)
				{
					break;
				}
				if (remainingIncrease < 0)
					throw new IllegalAccessError("This case should never occur!! Line " + DebugHelper.getLineNumber() + " in class "
							+ DebugHelper.getClassName());

				final int noOfIncreasableDegrees = this.anonSequence.get(i)
						- Math.max(this.solIncrease[i] - (i > 0 ? this.solIncrease[i - 1] : 0), this.k);

				if (noOfIncreasableDegrees > 0) // can we increase something?
				{
					int noOfActuallyIncreased = 0;
					int actuallIncLevel = 0;

					// determine how far how many degrees from level i get increased
					if ((noOfIncreasableDegrees * degIncLevel) <= remainingIncrease)
					{ // wasting all I can waste does exceed the waste-cost
						if ((this.anonSequence.get(i + degIncLevel) > 0) || (noOfIncreasableDegrees >= this.k))
						{ // either enough (>= k) vertices wasted or there are vertices at level i + degIncLevel
							noOfActuallyIncreased = noOfIncreasableDegrees;
							actuallIncLevel = degIncLevel;
						} else
						{ // cannot increase noOfIncreasableDegrees degrees by degIncLevel
							// => Check whether I can increase noOfIncreasableDegrees degrees by some j < degIncLevel.
							int j = i + degIncLevel;
							while ((j > i) && (this.anonSequence.get(j) == 0))
							{
								j--;
							}
							if (this.anonSequence.get(j) > 0)
							{
								noOfActuallyIncreased = noOfIncreasableDegrees;
								actuallIncLevel = j - i;
							}
						}
					} else
					{ // wasting all I can waste exceeds cost => I have to waste less!
						if (noOfIncreasableDegrees >= remainingIncrease)
						{
							if ((this.anonSequence.get(i + 1) > 0) || (remainingIncrease >= this.k))
							{
								// If one can waste the remaining cost by increase some degrees by one => do it!
								noOfActuallyIncreased = remainingIncrease;
								actuallIncLevel = 1;
							}
						} else if ((noOfIncreasableDegrees >= this.k)
								|| (this.anonSequence.get(i + (remainingIncrease / noOfIncreasableDegrees)) > 0))
						{
							noOfActuallyIncreased = noOfIncreasableDegrees;
							actuallIncLevel = remainingIncrease / noOfIncreasableDegrees;
						} else
						{
							int j = i + (remainingIncrease / noOfIncreasableDegrees);
							while ((j > i) && (this.anonSequence.get(j) == 0))
							{
								j--;
							}
							if (this.anonSequence.get(j) > 0)
							{
								noOfActuallyIncreased = noOfIncreasableDegrees;
								actuallIncLevel = j - i;
							}
						}
					}

					// do the wasting
					increasedDegrees += noOfActuallyIncreased * actuallIncLevel;
					for (int j = i; j < (i + actuallIncLevel); j++)
					{
						final int actualIncrease = Math.min(noOfIncreasableDegrees, remainingIncrease);
						wasteTriples[j].setT(wasteTriples[j].getT() + actualIncrease);
					}
				}
			}
			maxIncreasedDegrees = Math.max(maxIncreasedDegrees, increasedDegrees);
			if ((increasedDegrees >= this.cost) && constructSolution)
			{
				// System.out.println("Solution found; Single increases: " +
				// singleIncreases);
				this.examineSequence(wasteTriples);
				if (this.solutions.size() >= this.maxFoundSolutions)
					return increasedDegrees;
			}
		}
		if ((this.solutions.size() >= 1) || !this.findRealizableSequence)
			return maxIncreasedDegrees;
		else
			return 0;
	}

	/**
	 * checks whether vertex degrees can be increased by a total of <code>cost</code>. Produces all solutions, not just optimal ones.
	 * 
	 * @param i
	 * @param remainingCost
	 * @param wasteTriples
	 * @return
	 */
	private boolean determineIncreases(final int i, final int remainingCost, final Triple[] wasteTriples)
	{
		if (this.solutions.size() >= this.maxFoundSolutions)
			return true;

		final int currentBlockSize = this.anonSequence.get(i) + wasteTriples[i - 1].getT();

		if ((i <= 0) || (i > this.maxDegree))
			throw new IllegalAccessError("This case should never occur!! Line " + DebugHelper.getLineNumber() + " in class "
					+ DebugHelper.getClassName());
		// return false;
		else if (remainingCost == 0)
		{
			if ((currentBlockSize < this.k) && (currentBlockSize > 0))
				return false;
			for (int j = i; j <= this.maxDegree; j++)
			{
				wasteTriples[j] = new Triple(0, 0, 0);
			}
			return this.examineSequence(wasteTriples);
		} else if (i == this.maxDegree)
			return false;
		else
		{
			boolean solFound = false;
			final int max = Math.min(remainingCost, currentBlockSize);

			wasteTriples[i] = new Triple(0, 0, 0);
			for (int j = max; j >= 0; j--)
			{
				if (((currentBlockSize - j) == 0) || ((currentBlockSize - j) >= this.k))
				{
					wasteTriples[i].setT(j);
					wasteTriples[i].setC(j);
					solFound = this.determineIncreases(i + 1, remainingCost - j, wasteTriples) || solFound;

					if (this.solutions.size() >= this.maxFoundSolutions)
						return solFound;
				}
			}
			return solFound;
		}
	}

	/**
	 * Dynamic Programming Procedure to check whether <code>cost</code> vertex-degrees can be increased by exactly one. If such an increase exist,
	 * then it will be found.
	 * 
	 * Runs in O(MaxDegree * cost^2) time
	 * 
	 * @return
	 */
	private boolean determineOneIncreasesDynProg()
	{
		final int minValue = Math.min(this.cost, this.originalSequence.getMaxVerticesWithSameDegree());

		if ((1.0d * minValue * this.cost * this.maxDegree) > CostWaster.DYN_PROG_TABLE_UPPER_BOUND) // table would be too large
			return false;

		this.table = new BitSet[this.maxDegree + 1][];
		for (int i = 0; i <= this.maxDegree; i++)
		{
			this.table[i] = new BitSet[minValue + 1];
			for (int j = 0; j <= minValue; j++)
			{
				this.table[i][j] = new BitSet(this.cost + 1);
			}
		}

		int mint = 0;
		int maxt = 0;

		for (int i = 0; i <= this.maxDegree; i++)
		{

			for (int t = 0; t <= Math.min(this.cost, this.originalSequence.get(i, 0)); t++)
			{

				for (int c = 0; c <= this.cost; c++)
				{
					if (i == 0)
					{
						if ((c == 0) && ((this.anonSequence.get(i) >= (this.k + t)) || (this.anonSequence.get(i) == t)))
						{
							this.table[i][t].set(c);
						} else
						{
							this.table[i][t].clear(c);
						}
					} else
					{
						mint = 0;
						if (this.anonSequence.get(i - 1) == 0)
						{
							maxt = 0;
						} else
						{
							maxt = Math.min(this.originalSequence.get(i - 1, 0) - this.solIncrease[i - 1], c);
						}

						this.table[i][t].clear(c);
						for (int tt = mint; tt <= maxt; tt++)
						{
							if ((((this.anonSequence.get(i) - t) + tt) == 0) || (((this.anonSequence.get(i) - t) + tt) >= this.k))
							{
								if (this.table[i - 1][tt].get(c - tt))
								{
									this.table[i][t].set(c);
								}
							}
						}
					}
				}
			}
			if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 8)
			{
				System.out.println(Arrays.toString(this.table[i]));
			}
		}
		final int maxI = Math.min(this.originalSequence.get(this.maxDegree, 0), this.cost);
		boolean foundSolution = false;

		for (int i = 0; i <= maxI; i++)
		{
			if (((i == 0) || (i == this.anonSequence.get(this.maxDegree)) || (((this.anonSequence.get(this.maxDegree) - i) >= this.k) && (i >= this.k)))
					&& this.table[this.maxDegree][i].get(this.cost - i))
			{
				foundSolution = true;
				final Triple[] wasteTriples = new Triple[this.maxDegree + 1];
				wasteTriples[this.maxDegree] = new Triple(this.maxDegree, i, this.cost - i);
				this.recursivelyBuildAllSolutionSequences(this.maxDegree, wasteTriples);
			}
		}
		return foundSolution;
	}

	public int determineOneIncreasesHeuristic()
	{
		return this.determineOneIncreasesHeuristic(false);
	}

	/**
	 * Counts the number of untouched vertices in each block that might be increased by one without destroying the block. If this number is larger
	 * than <code>cost</code>, then a solution increasing exactly cost vertex-degrees is build. This is a heuristic. It might fail to find an existing
	 * solution!
	 * 
	 * Runs in O(maxDegree) time.
	 * 
	 * @return
	 */
	private int determineOneIncreasesHeuristic(final boolean constructSolution)
	{
		final Triple[] wasteTriples = new Triple[this.maxDegree + 1];
		wasteTriples[this.maxDegree] = new Triple(this.maxDegree, 0, 0);
		int singleIncreases = 0; // number of vertices that might be increased
									// by one without destroying one block
		int maxRemainingIncrease = 0;
		for (int i = 0; i < this.maxDegree; i++)
		{
			wasteTriples[i] = new Triple(i, 0, 0);
			maxRemainingIncrease = Math.max(0, this.cost - singleIncreases);

			if ((this.anonSequence.get(i) > this.k) && (this.anonSequence.get(i + 1) > 0))
			{
				final int inc = this.anonSequence.get(i) - Math.max(this.solIncrease[i] - (i > 0 ? this.solIncrease[i - 1] : 0), this.k);
				singleIncreases += inc;
				wasteTriples[i].setT(Math.min(inc, maxRemainingIncrease));
			}
		}

		if ((singleIncreases >= this.cost) && constructSolution)
		{
			if (!this.examineSequence(wasteTriples))
			{ // for some strange reason the new sequence is not realizable
				singleIncreases = 0;
			}
		}
		return singleIncreases;
	}

	private boolean examineSequence(final Triple[] wasteTriples)
	{
		if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 5)
		{
			System.out.println("===");
			System.out.println("seq:\t" + this.anonSequence);
			System.out.println("trip:\t" + Arrays.toString(wasteTriples));
		}

		final int[] newsequence = new int[this.maxDegree + 2];
		Arrays.fill(newsequence, 0);

		newsequence[this.maxDegree + 1] = wasteTriples[this.maxDegree].getT();
		for (int i = 0; i <= this.maxDegree; i++)
		{
			newsequence[i] += (this.anonSequence.get(i) - wasteTriples[i].getT()) + (i > 0 ? wasteTriples[i - 1].getT() : 0);
		}
		final CompactDegreeSequence seq = new CompactDegreeSequence(newsequence, false);
		final CompactDegreeSequence differenceToOrignal = seq.difference(this.originalSequence);
		this.foundRealizable = this.foundRealizable || differenceToOrignal.isRealizable();

		if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 5)
		{
			System.out.println("===");
			System.out.println("newSeq:\t" + seq);
			System.out.println("diff:\t" + differenceToOrignal);
		}

		if ((differenceToOrignal.isRealizable() && seq.isRealizable()) || !this.findRealizableSequence)
		{
			this.solutions.add(seq);
			if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 5)
			{
				System.out.println("NrOdSeqs:\t" + this.solutions.size());
			}
			return true;
		}
		// else
		// {
		// System.out.println("Nr:\t" + this.solutions.size());
		// System.out.println("\tFinal Seq:\t" + seq);
		// System.out.println("\tdiff Seq:\t" + differenceToOrignal);
		// System.out.println("real:\t" + differenceToOrignal.isRealizable() + "\t real2:\t" + seq.isRealizable());
		// }

		if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 5)
		{
			System.out.println("===");
		}

		return false;
	}

	public Set<CompactDegreeSequence> getAllRealizableSequences()
	{
		this.solutions = new HashSet<CompactDegreeSequence>();

		if ((((this.costSoFar + this.cost) % 2) == 1) && this.findRealizableSequence)
			return this.solutions;

		if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 4)
		{
			System.out.println("============Begin CostWaster==============");
			System.out.println("OrigSeq: " + this.originalSequence + "; " + Arrays.toString(this.solIncrease));
			System.out.println("cost: " + this.cost + "\t anonSeq: " + this.anonSequence);
		}

		this.iterateOverAllDegreeSequences();

		if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 4)
		{
			System.out.println("Realizable Solution found: " + this.foundRealizable);
			System.out.println("============End CostWaster==============");
		}
		return this.solutions;
	}

	public boolean isThereARealizableSequences()
	{
		if ((((this.costSoFar + this.cost) % 2) == 1) && this.findRealizableSequence)
			return false;

		if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 5)
		{
			System.out.println("============Begin CostWaster==============");
			System.out.println("OrigSeq: " + this.originalSequence + "; " + Arrays.toString(this.solIncrease));
			System.out.println("anonSeq: " + this.anonSequence + "; cost: " + this.cost);
		}

		this.iterateOverAllDegreeSequences();

		if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 5)
		{
			System.out.println("Realizable Solution found: " + this.foundRealizable);
			System.out.println("============End CostWaster==============");
		}
		return this.foundRealizable;
	}

	/**
	 * Given a k-anonymous sequence <code>anonSequence</code> obtained from <code>orignalSequence</code> (<code>triples</code> gives the trace in the
	 * dynamic programming table), this method computes all (Erdös-Gallai optimal) k-anon sequences obtainable from <code>anonSequence</code> by the
	 * cost of <code>additionalCosts</code>.
	 * 
	 * @param orignalSequence
	 * @param triples
	 * @param anonSequence
	 * @param additionalCosts
	 */
	public boolean iterateOverAllDegreeSequences()
	{
		// first try to increase the degree of additionalCosts many vertices by
		// exactly one
		// (Erdös-Gallai optimum)
		boolean solFound = false;
		final int increases = this.determineOneIncreasesHeuristic(true);
		if (increases >= this.cost)
		{
			solFound = true;
		}

		if (!solFound)
		{
			// try to find increase by one solution via dynamic programming
			solFound = this.determineOneIncreasesDynProg();
		}

		if (!solFound)
		{
			// try to find increase by one solution via dynamic programming
			solFound = (this.determineBalancedIncreasesHeuristic(true) > this.cost);
		}

		if (!solFound)
		{
			if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 5)
			{
				System.out.println("start exact cost-wasting");
			}
			// TODO no solution with just degree-one increases;...

			this.anonSequence = new CompactDegreeSequence(this.anonSequence, this.anonSequence.getMaxDegree() + 1 + (this.cost / this.k));
			this.maxDegree = this.anonSequence.getMaxDegree();
			final int max = Math.min(this.cost, this.anonSequence.get(0));
			final Triple[] wasteTriples = new Triple[this.maxDegree + 1];
			boolean tmp = false;

			wasteTriples[0] = new Triple(0, 0, 0);
			for (int i = max; i >= 0; i--)
			{
				if (((this.anonSequence.get(0) - i) == 0) || ((this.anonSequence.get(0) - i) >= this.k))
				{
					wasteTriples[0].setT(i);
					wasteTriples[0].setC(i);
					tmp = this.determineIncreases(1, this.cost - i, wasteTriples);
					if (tmp)
					{
						solFound = true;
					}
				}
			}
		}

		return solFound;
	}

	private void recursivelyBuildAllSolutionSequences(final int i, final Triple[] wasteTriples)
	{
		if (this.solutions.size() >= this.maxFoundSolutions)
			return;

		if (i <= 0)
		{
			this.examineSequence(wasteTriples);
		} else
		{
			final int t = wasteTriples[i].getT();
			final int c = wasteTriples[i].getC();

			int mint = 0;
			int maxt = 0;

			mint = 0;
			if (this.anonSequence.get(i - 1) == 0)
			{
				maxt = 0;
			} else
			{
				maxt = Math.min(this.originalSequence.get(i - 1) - this.solIncrease[i - 1], c);
			}

			for (int tt = mint; tt <= maxt; tt++)
			{
				if ((((this.anonSequence.get(i) - t) + tt) == 0) || (((this.anonSequence.get(i) - t) + tt) >= this.k))
				{
					if (this.table[i - 1][tt].get(c - tt))
					{
						wasteTriples[i - 1] = new Triple(i - 1, tt, c - tt);
						this.recursivelyBuildAllSolutionSequences(i - 1, wasteTriples);
						wasteTriples[i - 1] = null;
					}
				}
			}
		}
	}
}
