package algorithm;

import java.util.Arrays;
import java.util.BitSet;
import java.util.Set;

import algorithm.dataStructures.CompactDegreeSequence;
import algorithm.dataStructures.DegreeSequence;
import algorithm.dataStructures.GraphInterface;
import benchmarks.LogStringBuilder;
import benchmarks.MainCLIAnonymizerAndAlternativeRealizer;
import dataStructures.Store;

/**
 * Implementation of the new dynamic programming algorithm for the Degree Sequence Anonymization problem.
 * 
 * @author sepphar
 * 
 */
public class DegreeAnonymizer
{
	public static int logDR = 0;

	public static int cost(final CompactDegreeSequence degreeSequence, final int i, final int t, final int j)
	{
		if (t == 0)
			return 0;

		int uB = i;
		int lB = 0;
		int l = Math.max(0, i - 1);

		int noOfVerticesLtillI = degreeSequence.getNumberOfVerticesInBags(l, i);
		while ((noOfVerticesLtillI < t) || ((noOfVerticesLtillI - degreeSequence.get(l)) >= t))
		{
			if (noOfVerticesLtillI < t)
			{
				uB = l - 1;
			} else
			{
				lB = l + 1;
			}
			if (lB > uB)
			{
				break;
			}
			l = lB + ((uB - lB) / 2);
			noOfVerticesLtillI = degreeSequence.getNumberOfVerticesInBags(l, i);
		}

		final int no = l < i ? degreeSequence.getNumberOfVerticesInBags(l + 1, i) : 0;
		if ((no + Math.min(degreeSequence.get(l), t - no)) < t)
			return Integer.MAX_VALUE;

		final int degreeSum = l < i ? (no * j) - degreeSequence.getDegreeSumsInBags(l + 1, i) : 0;
		return degreeSum + ((t - no) * (j - l));

	}

	/**
	 * Returns the smallest index i such that i>=startIndex and the chain of full blocks with end i have in total at least k-1 vertices that can be
	 * increased to higher degrees (than i) without harming the fact that they are full.
	 * 
	 * @param k
	 * @param startIndex
	 * @param degreeSequence
	 * @return
	 */
	private static int getNextFullBag(final int k, final int startIndex, final CompactDegreeSequence degreeSequence)
	{
		int c = 0;
		int cIndex = -1;
		for (int t = startIndex; t <= degreeSequence.getMaxDegree(); t++)
		{
			if (degreeSequence.get(t) >= k)
			{
				if (cIndex == -1)
				{
					cIndex = t;
				}

				c += degreeSequence.get(t) - k;
			} else
			{
				cIndex = -1;
				c = 0;
			}
			if (c >= (k - 1))
				return cIndex;
		}
		return -1;
	}

	/**
	 * 
	 * @param sequence
	 * @param table table from dyn program
	 * @param k anon param
	 * @param i current degree
	 * @param t how many vertices raised from degree i
	 * @param c cost
	 * @return
	 */
	public static boolean getTableEntry(final CompactDegreeSequence sequence, final BitSet[][] table, final int k, final int i, final int t,
			final int c)
	{
		// countGetTableEntries++;
		if (c < 0)
			return false;
		if ((i < 0) || (t < 0) || (i > sequence.getMaxDegree()))
			throw new IllegalArgumentException();

		if (t == sequence.get(i))
		{
			if (i == 0)
				if (c == 0)
					return true;
				else
					return false;// throw new IllegalArgumentException();
			else
			{
				int j = i - 1;
				while ((sequence.get(j) == 0) && (j > 0))
				{
					j--;
				}
				return DegreeAnonymizer.getTableEntry(sequence, table, k, j, 0, c);
			}
		}

		if (t >= k)
			throw new IllegalArgumentException();

		final boolean b = table[i][t].get(c);
		return b;

	}

	/**
	 * Checks whether there is a chance to reduce the sequence
	 * 
	 * @param parSequence
	 * @return
	 */
	private static boolean promisingToApplyRule(final CompactDegreeSequence parSequence, final int k)
	{
		for (int deg = 0; deg <= parSequence.getMaxDegree(); deg++)
		{
			if ((parSequence.get(deg) < k) && (parSequence.get(deg) > 0))
			{
				final boolean increasePossible = ((deg + 1) <= parSequence.getMaxDegree())
						&& ((parSequence.get(deg) + parSequence.get(deg + 1)) >= k);
				final boolean increaseLower = (deg > 0) && ((parSequence.get(deg) + parSequence.get(deg - 1)) >= k);
				if (!increaseLower && increasePossible)
					// System.out.println("abort:" + parSequence);
					return false;
			}
		}
		// System.out.println("pass:" + parSequence);
		return true;
	}

	private final boolean applyDR;

	private final CompactDegreeSequence degreeSequence;

	private final boolean findRealizableSequence;

	private final GraphInterface graph;

	private final int k;

	public int lastExaminedCost;

	// private long c1 = 0, c2 = 0, c3 = 0; //three counters for debugging

	private int opt = Integer.MAX_VALUE;

	private final CompactDegreeSequence orginalSequence;

	public final int preprocessCost;

	public SolutionDegreeIterator solutionDegreeIterator;

	private final int solutionsStoredInCostWaster;
	
	public DegreeAnonymizer(final GraphInterface graph, final CompactDegreeSequence degreeSequence, final int k)
	{
		this(graph, degreeSequence, k, true, true, 2);
	}

	public DegreeAnonymizer(final GraphInterface graph, final CompactDegreeSequence degreeSequence, final int k,
			final boolean findRealizableSolution, final boolean applyDataReduction)
	{
		this(graph, degreeSequence, k, findRealizableSolution, applyDataReduction, 2);
	}

	public DegreeAnonymizer(final GraphInterface graph, final CompactDegreeSequence degreeSequence, final int k,
			final boolean findRealizableSolution, final boolean applyDataReduction, final int solutionsStoredInCostWaster)
	{
		// if ( findRealizableSolution && !degreeSequence.isRealizable() )
		// throw new
		// IllegalArgumentException("input sequence is not realizable!");

		this.graph = graph;
		this.orginalSequence = degreeSequence;
		this.k = k;
		this.applyDR = applyDataReduction;
		if (applyDataReduction)
		{
			final long tmpTimeStamp = DegreeAnonymizer.logDR > 0 ? System.currentTimeMillis() : 0;
			this.degreeSequence = this.iniDataReduction(degreeSequence);
			this.preprocessCost = this.degreeSequence.difference(this.orginalSequence).getTotalDegreeSum();
			if (DegreeAnonymizer.logDR > 0)
			{
				System.out.println("DR #mods:" + this.preprocessCost + " (" + ((System.currentTimeMillis() - tmpTimeStamp) / 1000d) + ")");
			}
		} else
		{
			this.degreeSequence = degreeSequence;
			this.preprocessCost = 0;
		}

		this.findRealizableSequence = findRealizableSolution;
		this.solutionsStoredInCostWaster = solutionsStoredInCostWaster;
	}

	public DegreeAnonymizer(final GraphInterface graph, final int k)
	{
		this(graph, new CompactDegreeSequence(DegreeSequence.calcDegreeSequence(graph), true), k);
	}

	public DegreeAnonymizer(final GraphInterface graph, final int k, final boolean findRealizableSolution, final boolean applyDataReduction)
	{
		this(graph, new CompactDegreeSequence(DegreeSequence.calcDegreeSequence(graph), true), k, findRealizableSolution, applyDataReduction, 2);
	}

	public Set<CompactDegreeSequence> getAllkAnonRealSequences()
	{
		final BitSet[][] table = new BitSet[this.degreeSequence.getMaxDegree() + 1][];
		this.solutionDegreeIterator = new SolutionDegreeIterator(this.graph, table, this.orginalSequence, this.degreeSequence, this.k,
				this.findRealizableSequence, !this.applyDR, this.solutionsStoredInCostWaster);

		{// ini table
			for (int i = 0; i <= this.degreeSequence.getMaxDegree(); i++)
			{
				final BitSet[] e = new BitSet[this.k];
				table[i] = e;
				// table.add(e);
				for (int t = 0; t < this.k; t++)
				{
					e[t] = new BitSet();
				}
			}
		}

		final LiuTerziDegreeAnonymizer minimumDegreeAnonymizer = new LiuTerziDegreeAnonymizer(this.degreeSequence, this.k);
		minimumDegreeAnonymizer.getMinSolSize();

		for (int c = 0;; c++)
		{
			if ((MainCLIAnonymizerAndAlternativeRealizer.logLevel > 2) && ((c % 500) == 0))
			{
				System.out.println("dyn prog for c:" + (c + this.preprocessCost)
				// + " \tcalls: " +
				// myFormatter.format(DegreeAnonymizer.countGetTableEntries)
				// + "\tc1: " + myFormatter.format(c1)
				// + "\tc2: " + myFormatter.format(c2)
				// + "\tc3: " + myFormatter.format(c3)
						+ "\t timeSoFar:" + ((System.currentTimeMillis() - MainCLIAnonymizerAndAlternativeRealizer.timeStamp) / 1000d) + "sec"
						+ "\t ltSolverSize:" + minimumDegreeAnonymizer.getMinSolSize()
						);
			}
			for (int i = 0; i <= this.degreeSequence.getMaxDegree(); i++)
			{
				final BitSet[] subTable = table[i];

				for (int t = 0; t < this.k; t++)
				{
					final BitSet bitSet = subTable[t];
					if (minimumDegreeAnonymizer.getMinAnonCosts(i, t) <= c)
					{
						if (i == 0)
						{
							if ((c == 0) && (((this.degreeSequence.get(i) - t) >= this.k) || ((this.degreeSequence.get(i) - t) == 0)))
							{
								// if (minimumDegreeAnonymizer.getMinAnonCosts(i, t) > c)
								// throw new IllegalStateException("k: " + this.k + "\ti: " + i + "\t t: " + t + "\t c: " + c + "\tminSol:"
								// + minimumDegreeAnonymizer.getMinAnonCosts(i, t) + "\t seq" + this.degreeSequence.toString());

								bitSet.set(c);
							}
						} else if (this.degreeSequence.get(i) > t)
						{
							if ((this.degreeSequence.get(i) - t) >= this.k)
							{
								if (table[i - 1][0].get(c))
								{
									// if (minimumDegreeAnonymizer.getMinAnonCosts(i, t) > c)
									// throw new IllegalStateException("k: " + this.k + "\ti: " + i + "\t t: " + t + "\t c: " + c + "\tminSol:"
									// + minimumDegreeAnonymizer.getMinAnonCosts(i, t) + "\t seq" + this.degreeSequence.toString());

									bitSet.set(c);
								} else
								{
									int j = i - 1;
									int counter = 0;
									while ((j >= 0) && ((counter += this.degreeSequence.get(j)) < this.k)
											&& (DegreeAnonymizer.cost(this.degreeSequence, i - 1, counter, i) <= c))
									{
										// c1++;
										if (DegreeAnonymizer.getTableEntry(this.degreeSequence, table, this.k, j, this.degreeSequence.get(j), c
												- DegreeAnonymizer.cost(this.degreeSequence, i - 1, counter, i)))
										{
											// if (minimumDegreeAnonymizer.getMinAnonCosts(i, t) > c)
											// throw new IllegalStateException("k: " + this.k + "\ti: " + i + "\t t: " + t + "\t c: " + c
											// + "\tminSol:" + minimumDegreeAnonymizer.getMinAnonCosts(i, t) + "\t seq"
											// + this.degreeSequence.toString());

											bitSet.set(c);
											break;
										}
										j--;
									}
								}
							}
							// 0 < degreeSequence.get(i) - t < k
							else if (DegreeAnonymizer.cost(this.degreeSequence, i - 1, this.k - (this.degreeSequence.get(i) - t), i) <= c)
							{
								int j = i - 1;
								int counter = 0;
								while ((j >= 0) && ((counter + this.degreeSequence.get(j)) < (this.k - (this.degreeSequence.get(i) - t))))
								{
									counter += this.degreeSequence.get(j);
									j--;
								}
								if (j >= 0)
								{
									final int remainingFromJ = this.k - ((counter + this.degreeSequence.get(i)) - t); // #degrees taken from level j
									// c2++;
									if (DegreeAnonymizer.getTableEntry(this.degreeSequence, table, this.k, j, remainingFromJ,
											c - DegreeAnonymizer.cost(this.degreeSequence, i - 1, this.k - (this.degreeSequence.get(i) - t), i)))
									{
										// if (minimumDegreeAnonymizer.getMinAnonCosts(i, t) > c)
										// throw new IllegalStateException("k: " + this.k + "\ti: " + i + "\t t: " + t + "\t c: " + c + "\tminSol:"
										// + minimumDegreeAnonymizer.getMinAnonCosts(i, t) + "\t seq" + this.degreeSequence.toString());

										bitSet.set(c);
									} else
									{
										int u = this.degreeSequence.get(j) - remainingFromJ;
										while ((u < this.k)
												&& (DegreeAnonymizer.cost(this.degreeSequence, i - 1,
														(u + this.k) - (this.degreeSequence.get(i) - t), i) <= c))
										{
											// c3++;
											if (DegreeAnonymizer.getTableEntry(
													this.degreeSequence,
													table,
													this.k,
													j,
													this.degreeSequence.get(j),
													c
															- DegreeAnonymizer.cost(this.degreeSequence, i - 1,
																	(u + this.k) - (this.degreeSequence.get(i) - t), i)))
											{
												// if (minimumDegreeAnonymizer.getMinAnonCosts(i, t) > c)
												// throw new IllegalStateException("k: " + this.k + "\ti: " + i + "\t t: " + t + "\t c: " + c
												// + "\tminSol:" + minimumDegreeAnonymizer.getMinAnonCosts(i, t) + "\t seq"
												// + this.degreeSequence.toString());

												bitSet.set(c);
												break;
											}

											j--;
											if (j >= 0)
											{
												u += this.degreeSequence.get(j);
											} else
											{
												break;
											}
										}
									}
								}
							}
						} else
						{
							bitSet.set(c, table[i - 1][t - this.degreeSequence.get(i)].get(c));
						}
					}
				}
			}
			if (table[this.degreeSequence.getMaxDegree()][0].get(c))
			{
				if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 0)
				{
					System.out.println("found dyn prog solution for c:" + (c + this.preprocessCost) + "\tcurrentUpperBound:"
							+ this.solutionDegreeIterator.getCurrentUpperbound() + "\t timeSoFar:"
							+ ((System.currentTimeMillis() - MainCLIAnonymizerAndAlternativeRealizer.timeStamp) / 1000d) + "sec");
				}
				this.opt = Math.min(c, this.opt);
				this.lastExaminedCost = c;
			}
			if (this.opt != Integer.MAX_VALUE)
			{
				LogStringBuilder.getLogger().addEntry("lastExaminedCost", this.lastExaminedCost + this.preprocessCost);
				LogStringBuilder.getLogger().addTimeEntry(LogStringBuilder.LogKey_TIME_FOR_GET_ALL_K_ANON_REAL_SEQUENCES,false);

				this.lastExaminedCost = c;

				Set<CompactDegreeSequence> s;
				if ((s = this.solutionDegreeIterator.getAllRealizableSequences(c)) != null)
					return s;
				if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 3)
				{
					System.out.println("test solution of size:" + c + "(recInstances:"
							+ this.solutionDegreeIterator.getNumberOfRecConsideredInstances() + ")");
				}
			}
		}
	}

	public SolutionDegreeIterator getLastSolutionDegreeIterator()

	{
		return this.solutionDegreeIterator;
	}

	/**
	 * Returns the set of k-anon sequences with the smallest difference c to the input sequence such that the "increase" from the given to the
	 * resulting instance is graphically-realizable (Erdös-Gallai).
	 * 
	 * @return
	 */
	public int getSizeOfMinimumSolution()
	{
		final Set<CompactDegreeSequence> s = this.getAllkAnonRealSequences();
		if ((s != null) && (s.size() > 0))
			return s.iterator().next().difference(this.orginalSequence).getTotalDegreeSum();
		return -1;
	}

	private CompactDegreeSequence iniDataReduction(final CompactDegreeSequence degreeSequence)
	{
		final int[] newSequence = degreeSequence.getDegreeCopy();

		final Store<Integer, CompactDegreeSequence> upperBoundForKAnon = UpperBound.getFastUpperBoundForKAnon(degreeSequence, this.k);
		// TODO check whether this bound should be indeed the advanced Erdös
		// bound
		final int minAddCost = upperBoundForKAnon.getS2().difference(degreeSequence).noOfAddDegreeOneVerticesForBeingRealizableWRSimpleErdös();

		if (CostWaster.canYouWasteDegreeOneVertices(degreeSequence, minAddCost, this.k))
		{
			int lastFullBlock = 0;

			// iterate over all blocks/degrees in ascending order
			for (int currentBlock = 0; currentBlock <= degreeSequence.getMaxDegree(); currentBlock++)
			{
				if (degreeSequence.get(currentBlock) >= ((this.k << 1) - 1))
				{
					lastFullBlock = currentBlock;
				} else if (degreeSequence.get(currentBlock) > 0)
				{// currentBlock is to small but not empty
					if (currentBlock < degreeSequence.getMaxDegree())
					{
						final int nextFullBlock = DegreeAnonymizer.getNextFullBag(this.k, currentBlock + 1, degreeSequence);

						if ((nextFullBlock > 0) && ((lastFullBlock > 0) || (nextFullBlock < degreeSequence.getMaxDegree())))
						{
							// create subsequence from lastFullBlock to
							// nextFullBlock
							final CompactDegreeSequence parSequence = new CompactDegreeSequence(Arrays.copyOfRange(newSequence, lastFullBlock,
									nextFullBlock + 1), false);
							if (DegreeAnonymizer.promisingToApplyRule(parSequence, this.k))
							{
								final Set<CompactDegreeSequence> parSolutions = new DegreeAnonymizer(this.graph, parSequence, this.k, false, false,
										this.solutionsStoredInCostWaster).getAllkAnonRealSequences();

								for (final CompactDegreeSequence parSolution : parSolutions)
								{
									final CompactDegreeSequence difference = parSolution.difference(parSequence);

									if (((lastFullBlock == 0) || (parSolution.get(0) >= this.k))
											// first block is either full or it is degree 0
											&& ((difference.getMaxDegree() == 1) || (difference.getNumberOfVerticesInBags(2,
													difference.getMaxDegree()) <= 0)))
												// max raise is 1
									{// sol found, copy solution back
										for (int r = 0; r <= parSolution.getMaxDegree(); r++)
										{
											newSequence[lastFullBlock + r] = parSolution.get(r);
										}

										break;
									}
								}

								lastFullBlock = nextFullBlock;
								currentBlock = nextFullBlock;
							} else
							{// range not promising; skip it
								lastFullBlock = nextFullBlock;
								currentBlock = nextFullBlock;
							}
						}
					} else
					{
						if (degreeSequence.get(currentBlock) < k)
						{
							// currentBlock is max degree block and it needs to be
							// filled up to k
							final CompactDegreeSequence t = new CompactDegreeSequence(newSequence, false);
							int index = t.getMaxDegree() - 1;
							while (t.getNumberOfVerticesInBags(index, t.getMaxDegree()) < this.k)
							{
								newSequence[index] = 0;
								index--;
							}
							newSequence[t.getMaxDegree()] = this.k;
							newSequence[index] = newSequence[index] - (this.k - t.getNumberOfVerticesInBags(index + 1, t.getMaxDegree()));
						}
					}
				}
			}

		}

		// return degreeSequence2;
		return new CompactDegreeSequence(newSequence, false);

	}
}
