package algorithm;

import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import algorithm.alternativeRealizer.BlockRaiseListComputer;
import algorithm.alternativeRealizer.RaiseSequenceEntry;
import algorithm.dataStructures.CompactDegreeSequence;
import algorithm.dataStructures.GraphInterface;

public class ErdosGallaiAdvanced implements ErdosGallaiInterface
{

	private GraphInterface graph;

	private CompactDegreeSequence init;

	public long time1 = 0, time2 = 0, noOfEGCalls = 0, maxRFail = 0;

	public ErdosGallaiAdvanced(final GraphInterface graph, final CompactDegreeSequence init)
	{
		this.setGraph(graph, init);
	}

	/**
	 * 
	 * @param init initial compact degree sequence
	 * @param sol solution compact degree sequence
	 * @param fixAll fix all vertices, even the ones with degree raise = 1
	 * @return Realizability of solution degree sequence
	 */
	@Override
	public int computeFeasibility(final CompactDegreeSequence sol, final boolean fixAll)
	{
		this.noOfEGCalls++;

		// For each vertex whose degree has to be raised: Store original vertex
		// degree (first) and solution vertex degree (second), ordered
		// descending
		final List<RaiseSequenceEntry> degreeRaiseList = new ArrayList<>(1000);

		degreeRaiseList.add(null);

		final TIntSet solutionDegreeSet = new TIntHashSet();
		final List<RaiseSequenceEntry> degreeBlockRaiseList = new BlockRaiseListComputer().generateBlockRaiseListBottomUp(this.init, sol);
		for (final RaiseSequenceEntry t : degreeBlockRaiseList)
		{

			for (int i = 0; i < t.getThird(); i++)
			{

				if (solutionDegreeSet.contains(t.getFirst()))
				{
					degreeRaiseList.add(new RaiseSequenceEntry(t.getFirst(), t.getSecond(), 0));
				} else
				{
					degreeRaiseList.add(new RaiseSequenceEntry(t.getFirst(), t.getSecond(), 1));
				}
				solutionDegreeSet.add(t.getSecond());

			}

		}

		Collections.sort(degreeRaiseList, new Comparator<RaiseSequenceEntry>()
		{
			@Override
			public int compare(final RaiseSequenceEntry o1, final RaiseSequenceEntry o2)
			{
				if (o1 == null)
					return -1;
				if (o2 == null)
					return 1;
				if ((o2.getSecond() - o2.getFirst()) != (o1.getSecond() - o1.getFirst()))
					return o1.getSecond() - o1.getFirst() - (o2.getSecond() - o2.getFirst());
				else
					return o2.getSecond() - o1.getSecond();
			}

		});

		int sum1 = 0;
		int minimalSumSoFar = 0;

		for (int r = 1; r < (degreeRaiseList.size() - 1); r++)
		{
			long tmp = System.currentTimeMillis();

			// calculate left sum
			final RaiseSequenceEntry t = degreeRaiseList.get(r);
			sum1 += t.getSecond() - t.getFirst();

			// calculate (r*(r-1)) first. Then iterate over first r entries in
			// degree raise list (index j), for each j iterate over entries
			// j+1,...,r entries. If for a pair of entries t1, t2 there is only
			// one vertex with vertex degree stored in entry t1 and only one
			// vertex with vertex degree stored in entry t2 AND these vertices
			// are adjacent, substract 2 from the product. In order to fit to
			// the data reduction rules from Degree Anonymization, we only do
			// this if a vertex degree has to be raised by more than 1.
			int mul = r * (r - 1);
			if (mul > 0)
			{
				int j = r;
				{
					final RaiseSequenceEntry t1 = degreeRaiseList.get(j);
					if ((fixAll || ((t1.getSecond() - t1.getFirst()) != 1)) && (t1.getThird() == 1))
					{
						int minimalSum = Integer.MAX_VALUE;
						for (final int nodet1 : this.graph.getVerticesWithDegree(t1.getFirst()))
						{
							int sum = 0;
							final TIntSet alreadyTaken = new TIntHashSet();
							int l = 1;
							while (l < j)
							{
								final RaiseSequenceEntry t2 = degreeRaiseList.get(l);

								if ((fixAll || ((t2.getSecond() - t2.getFirst()) != 1)) && (t2.getThird() == 1))
								{
									boolean substract = true;
									for (final int nodet2 : this.graph.getVerticesWithDegree(t2.getFirst()))
									{
										if (!this.graph.isAdjacent(nodet1, nodet2))
										{
											if (alreadyTaken.add(nodet2))
											{
												substract = false;
												break;
											}

										}
									}
									if (substract)
									{
										sum += 2;
									}

								}
								++l;
							}
							minimalSum = Math.min(minimalSum, sum);
						}
						if (minimalSum != Integer.MAX_VALUE)
						{

							minimalSumSoFar += minimalSum;
							mul -= minimalSumSoFar;
						}

					}

					++j;
				}
			}
			this.time1 += System.currentTimeMillis() - tmp;
			tmp = System.currentTimeMillis();

			// Calculation of the right side sum.
			int sum2 = 0;
			int i = r + 1;
			while (i < degreeRaiseList.size())
			{

				final RaiseSequenceEntry t1 = degreeRaiseList.get(i);
				// just what's the algorithm of erdos & gallai
				sum2 += Math.min(r, t1.getSecond() - t1.getFirst());

				// cause of the data reduction rules in DegAnon..
				if ((fixAll || ((t1.getSecond() - t1.getFirst()) != 1)) && (t1.getThird() == 1))
				{
					final int[] nodesOneDegree = this.graph.getVerticesWithDegree(t1.getFirst());

					// Here we save how much we can substract in the end for the
					// i-th vertex
					int maxNonAdjacentNodes = 0;

					// inspect all vertices which come into consideration for
					// the i-th raise
					for (int l = 0; (l < nodesOneDegree.length) && (maxNonAdjacentNodes <= Math.min(r, t1.getSecond() - t1.getFirst())); l++)
					{
						final int nodet1 = nodesOneDegree[l];

						int nonAdjacentNodes = 0;
						// inspect the first r vertices ...
						for (int x = 1; x <= r; x++)
						{
							final RaiseSequenceEntry t2 = degreeRaiseList.get(x);
							if (t2.getThird() == 0)
							{
								nonAdjacentNodes = Integer.MAX_VALUE;
								break;
							} else
							{
								if (fixAll || ((t2.getSecond() - t2.getFirst()) != 1))
								{
									// where for each vertex, we look at all
									// that come into consideration
									// while (nodeTwoDegreeIterator.hasNext()) {
									for (final int nodet2 : this.graph.getVerticesWithDegree(t2.getFirst()))
									{
										// in this case, we can connect node i
										// to a node x with specific vertex
										// degree
										if (!this.graph.isAdjacent(nodet1, nodet2))
										{
											nonAdjacentNodes++;
											break;
										}

									}

								}
							}

						}

						maxNonAdjacentNodes = Math.max(maxNonAdjacentNodes, nonAdjacentNodes);

					}

					if (r <= (t1.getSecond() - t1.getFirst()))
					{
						if ((r - maxNonAdjacentNodes) > 0)
						{
							sum2 -= r - maxNonAdjacentNodes;
						}
					} else
					{
						if ((t1.getSecond() - t1.getFirst() - maxNonAdjacentNodes) > 0)
						{
							sum2 -= t1.getSecond() - t1.getFirst() - maxNonAdjacentNodes;
						}
					}

				}

				++i;

			}
			this.time2 += System.currentTimeMillis() - tmp;

			if (sum1 > (mul + sum2))
			{
				this.maxRFail = Math.max(this.maxRFail, r);
				return sum1 - mul - sum2;
			}

		}
		return 0;
	}

	@Override
	public int computeFeasibility(final CompactDegreeSequence sol, final boolean fixAll, final boolean getMaxDiff)
	{
		return this.computeFeasibility(sol, fixAll);
	}

	@Override
	public int computeFeasibility(final CompactDegreeSequence sol, final boolean fixAll, final boolean getMaxDiff, final boolean checkPartialSolution)
	{
		return this.computeFeasibility(sol, fixAll);
	}

	public long getMaxRFail()
	{
		return this.maxRFail;
	}

	public long getNoOfEGCalls()
	{
		return this.noOfEGCalls;
	}

	public long getTime1()
	{
		return this.time1;
	}

	public long getTime2()
	{
		return this.time2;
	}

	@Override
	public void setGraph(final GraphInterface graph, final CompactDegreeSequence init)
	{
		this.graph = graph;
		this.init = init;
	}
}
