package algorithm;

import gnu.trove.map.hash.TObjectIntHashMap;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import java.util.SortedMap;
import java.util.TreeMap;

import algorithm.alternativeRealizer.DegreesChangeList;
import algorithm.dataStructures.CompactDegreeSequence;
import algorithm.dataStructures.GraphInterface;
import dataStructures.Store;

public class FastErdosGallai implements ErdosGallaiInterface
{

	private class MyIntArrayKey
	{
		private final int[] data;

		private final int hash;

		public MyIntArrayKey(final int[] values)
		{
			this.data = values;
			this.hash = Arrays.hashCode(this.data);
		}

		@Override
		public boolean equals(final Object obj)
		{
			if ((obj != null) && (obj instanceof MyIntArrayKey))
			{
				final MyIntArrayKey s = (MyIntArrayKey) obj;
				return Arrays.equals(s.data, this.data);
			}
			return false;
		}

		@Override
		public int hashCode()
		{
			return this.hash;
		}

		@Override
		public String toString()
		{
			return "MyIntArrayKey [data=" + Arrays.toString(this.data) + "]";
		}

	}

	private DegreesChangeList degreeRaiseList;

	private GraphInterface graph;

	private CompactDegreeSequence init;

	private final TObjectIntHashMap<MyIntArrayKey> noOfEdgesBetweenTwoBlocks;

	private final TObjectIntHashMap<MyIntArrayKey> SolDependingResults;

	public FastErdosGallai(final GraphInterface graph, final CompactDegreeSequence init)
	{
		this.noOfEdgesBetweenTwoBlocks = new TObjectIntHashMap<MyIntArrayKey>(10, 0.5f, -1);
		this.SolDependingResults = new TObjectIntHashMap<MyIntArrayKey>();
		this.setGraph(graph, init);
	}

	@Override
	public int computeFeasibility(final CompactDegreeSequence sol, final boolean fixAll)
	{
		return this.computeFeasibility(sol, fixAll, true);
	}

	@Override
	public int computeFeasibility(final CompactDegreeSequence sol, final boolean fixAll, final boolean getMaxDiff)
	{
		return this.computeFeasibility(sol, fixAll, true, false);
	}

	@Override
	public int computeFeasibility(final CompactDegreeSequence sol, final boolean fixAll, final boolean getMaxDiff, final boolean checkPartialSolution)
	{
		if (this.graph == null)
			return 0;

		int maxDiff = 0;
		this.SolDependingResults.clear();
		// List of changes
		this.degreeRaiseList = sol.differenceWithDegrees(this.init);
		final TreeMap<Integer, Store<Integer, TreeMap<Integer, Integer>>> degreeRaiseTree = this.degreeRaiseList.getTreeMap();

		long leftsum = 0, rightsum = 0;
		int r = 0; // the index from the Erdös Gallai Inequalities

		int edgesWithinTheCliquePart = 0, edgesBetweenParts = 0;
		int lastFromIncrease = -1;

		for (final Entry<Integer, Store<Integer, TreeMap<Integer, Integer>>> entry : degreeRaiseTree.descendingMap().entrySet())
		// iterating over all increases (of vertex degrees) in descending order
		{
			final int increase = entry.getKey();
			final TreeMap<Integer, Integer> value = entry.getValue().getS2();
			final int amountByIncrease = entry.getValue().getS1();

			SortedMap<Integer, Integer> tmpSmallMap;
			SortedMap<Integer, Store<Integer, TreeMap<Integer, Integer>>> tmpMap, tmpMap2;
			r += amountByIncrease;
			edgesBetweenParts = 0; // do not reset edgesWithinTheCliquePart!

			leftsum += increase * amountByIncrease;
			rightsum = r * (r - 1);

			// count the number of edges within the "clique-part"
			if (r > 1)
			{
				// iterating over vertex degrees (in ascending order) getting
				// raised by increase
				for (final Entry<Integer, Integer> innerEntry : value.descendingMap().entrySet())
				{
					final int origDegree = innerEntry.getKey();
					if (!this.degreeRaiseList.getDegreeRaisedTo().contains(origDegree) && (fixAll || (increase > 1)))
					{
						final int amountOfVertices = innerEntry.getValue();

						tmpMap = degreeRaiseTree.tailMap(increase + 1);
						edgesWithinTheCliquePart += this.edgesBetweenBigMap(increase, origDegree, amountOfVertices, tmpMap, fixAll);

						tmpSmallMap = value.tailMap(origDegree);
						edgesWithinTheCliquePart += this.edgesBetweenSmallMap(increase, origDegree, amountOfVertices, tmpSmallMap, increase, fixAll);
					}
				}
			}

			tmpMap = degreeRaiseTree.headMap(increase);
			tmpMap2 = degreeRaiseTree.tailMap(increase);
			final int fromIncrease = tmpMap2.firstKey();
			final int toIncrease = tmpMap2.lastKey();
			// count the number of edges between the two parts
			for (final Entry<Integer, Store<Integer, TreeMap<Integer, Integer>>> entry2 : tmpMap.entrySet())
			{
				final int increase2 = entry2.getKey();
				final int amountByIncrease2 = entry2.getValue().getS1();
				rightsum += amountByIncrease2 * Math.min(r, increase2);
				for (final Entry<Integer, Integer> entry3 : entry2.getValue().getS2().entrySet())
				{
					final int origDegree2 = entry3.getKey();
					if (!this.degreeRaiseList.getDegreeRaisedTo().contains(origDegree2) && (fixAll || (increase2 > 1)))
					{
						final int amountOfVertices2 = entry3.getValue();

						int tmpEdgesBetweenParts = 0;

						final int[] val = { increase2, origDegree2, amountOfVertices2, fromIncrease, toIncrease };
						final MyIntArrayKey key = new MyIntArrayKey(val);

						final int[] val2 = { increase2, origDegree2, amountOfVertices2, lastFromIncrease, toIncrease };
						final MyIntArrayKey key2 = new MyIntArrayKey(val2);
						if (this.SolDependingResults.containsKey(key2))
						{
							tmpEdgesBetweenParts = this.SolDependingResults.get(key2);
							tmpEdgesBetweenParts += this.edgesBetweenSmallMap(increase2, origDegree2, amountOfVertices2, tmpMap2.get(fromIncrease)
									.getS2(), fromIncrease, fixAll);
							this.SolDependingResults.put(key, tmpEdgesBetweenParts);
						} else
						{
							tmpEdgesBetweenParts = this.edgesBetweenBigMap(increase2, origDegree2, amountOfVertices2, tmpMap2, fixAll);
							this.SolDependingResults.put(key, tmpEdgesBetweenParts);
						}

						if ((r * amountOfVertices2) < (tmpEdgesBetweenParts + (amountOfVertices2 * Math.min(r, increase2))))
						{
							edgesBetweenParts += (tmpEdgesBetweenParts + (amountOfVertices2 * Math.min(r, increase2))) - (r * amountOfVertices2);
						}
					}
				}
			}
			lastFromIncrease = fromIncrease;

			final long diff;
			if (checkPartialSolution)
			{
				// for partial solutions there can be other high-degree vertices not known so far => cannot count clique-edges twice
				diff = (leftsum + edgesWithinTheCliquePart + edgesBetweenParts) - rightsum;
			} else
			{
				diff = (leftsum + (2 * edgesWithinTheCliquePart) + edgesBetweenParts) - rightsum;
			}
			if (diff > 0)
			{
				if (getMaxDiff)
				{
					maxDiff = Math.max(maxDiff, (int) diff);
				} else
					return (int) diff;
			}

			// the r * (r-1) part is large enough to satisfy the inequality...
			if ((r * (r - 1)) > (2 * leftsum))
				return maxDiff;

		}
		return maxDiff;
	}

	private int edgesBetweenBigMap(final int increase, final int origDegree, final int amountOfVertices,
			final SortedMap<Integer, Store<Integer, TreeMap<Integer, Integer>>> tailMap, final boolean fixAll)
	{
		int sum = 0;
		for (final Entry<Integer, Store<Integer, TreeMap<Integer, Integer>>> amountEntry : tailMap.entrySet())
		{
			final int increase2 = amountEntry.getKey();
			if (!this.degreeRaiseList.getDegreeRaisedTo().contains(origDegree) && (fixAll || (increase2 > 1)))
			{
				final TreeMap<Integer, Integer> value2 = amountEntry.getValue().getS2();
				sum += this.edgesBetweenSmallMap(increase, origDegree, amountOfVertices, value2, increase2, fixAll);
			}
		}
		return sum;
	}

	private int edgesBetweenSmallMap(final int increase, final int origDegree, final int amountOfVertices, final SortedMap<Integer, Integer> value2,
			final int increase2, final boolean fixAll)
	{
		int sum = 0;
		for (final Entry<Integer, Integer> innerEntry2 : value2.entrySet())
		{
			final int origDegree2 = innerEntry2.getKey();
			final int amountOfVertices2 = innerEntry2.getValue();
			if (!this.degreeRaiseList.getDegreeRaisedTo().contains(origDegree) && !this.degreeRaiseList.getDegreeRaisedTo().contains(origDegree2)
					&& (fixAll || ((increase2 > 1) && (increase > 1))))
			{
				sum += this.getAmountOfEdgesBetweenDegreeBlocks(origDegree, origDegree2, amountOfVertices, amountOfVertices2);
			}
		}
		return sum;
	}

	public int getAmountOfEdgesBetweenDegreeBlocks(final int degree1, final int degree2, final int noOfDeg1Vertices, final int noOfDeg2Vertices)
	{
		if (((this.graph.getNumberOfVerticesWithDegree(degree1) - noOfDeg1Vertices) >= (degree2 * noOfDeg2Vertices))
				|| ((this.graph.getNumberOfVerticesWithDegree(degree2) - noOfDeg2Vertices) >= (degree1 * noOfDeg1Vertices)))
			return 0;

		final int[] myArray = { degree1, degree2, noOfDeg1Vertices, noOfDeg2Vertices };
		final MyIntArrayKey key = new MyIntArrayKey(myArray);

		final int noEdges;
		if ((noEdges = this.noOfEdgesBetweenTwoBlocks.get(key)) != -1)
			return noEdges;
		else
		{
			int minNoOfEdges = 0;

			final List<Integer> noOfBadNeighbors = new ArrayList<Integer>(this.graph.getNumberOfVerticesWithDegree(degree1));

			for (final int node1 : this.graph.getVerticesWithDegree(degree1))
			{
				int noOfAdjacentVertices = 0;

				int[] vertices;
				if (this.graph.getDegree(node1) < this.graph.getNumberOfVerticesWithDegree(degree2))
				{
					vertices = this.graph.getNeighbors(node1);
				} else
				{
					vertices = this.graph.getVerticesWithDegree(degree2);
				}
				for (final int node2 : vertices)
				{
					if (((this.graph.getDegree(node1) < this.graph.getNumberOfVerticesWithDegree(degree2)) && (this.graph.getDegree(node2) == degree2))
							|| ((this.graph.getDegree(node1) >= this.graph.getNumberOfVerticesWithDegree(degree2)) && this.graph.isAdjacent(node1,
									node2)))
					{
						noOfAdjacentVertices++;
					}
				}
				noOfBadNeighbors.add(noOfAdjacentVertices);
			}
			Collections.sort(noOfBadNeighbors);

			if (this.graph.getNumberOfVerticesWithDegree(degree1) < noOfDeg1Vertices)
			{
				System.err.println("ohje; deg: " + degree1 + ", " + this.graph.getNumberOfVerticesWithDegree(degree1) + "\tno: " + noOfDeg1Vertices);
			}

			final Iterator<Integer> iterator = noOfBadNeighbors.iterator();
			for (int i = 1; i <= noOfDeg1Vertices; i++)
			{
				final int no = iterator.next();
				if ((no > 0) && ((this.graph.getNumberOfVerticesWithDegree(degree2) - noOfDeg2Vertices) < no))
				{
					// problems during realization process only occur if a
					// vertex with degree1 is adjacent to more than
					// (this.getAmountOfNodesWithDegree(degree2) - tmp)
					// vertices of degree2
					minNoOfEdges += no - (this.graph.getNumberOfVerticesWithDegree(degree2) - noOfDeg2Vertices);
				}
			}

			if (degree1 == degree2) // same degree => edges counted twice
			{
				minNoOfEdges = minNoOfEdges / 2;
			}
			this.noOfEdgesBetweenTwoBlocks.put(key, minNoOfEdges);
			return minNoOfEdges;
		}
	}

	public int getHashMapSize()
	{
		return this.noOfEdgesBetweenTwoBlocks.size();
	}

	@Override
	public void setGraph(final GraphInterface graph, final CompactDegreeSequence init)
	{
		this.graph = graph;
		this.init = init;
		this.noOfEdgesBetweenTwoBlocks.clear();
		this.SolDependingResults.clear();
	}
}
