package algorithm;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import algorithm.dataStructures.CompactDegreeSequence;

/**
 * Class for computing minimum (NOT minimal!) solutions for anonymizing the degree sequences (pure number problem)
 * 
 * @author nichterlein
 * 
 */
public class LiuTerziDegreeAnonymizer
{
	private CompactDegreeSequence blockSequence;
	private int blockSeqAnonymity;

	private int[] compressedDegreeSequence;

	private int k;

	private int[] lastDegreeIndex; // stores for each degree the last index in the <code>compressedDegreeSequence</code>

	private int maxNumberOfStoredSolutions = 50; // parameter can be adjusted by choosing the corresponding constructor
	private Set<CompactDegreeSequence> sol;

	private boolean solutionsComputed = false;

	private int[] table; // table for dynamic programming

	private boolean tableComputed = false;

	/**
	 * 
	 * @param seq the block-sequence of the input
	 * @param k
	 */
	public LiuTerziDegreeAnonymizer(final CompactDegreeSequence seq, final int k)
	{
		this.initialize(seq, k);
	}

	/**
	 * 
	 * @param seq the block-sequence of the input
	 * @param k
	 * @param numberOfStoredSolutions upper bound on the number of computed solutions (that are k-anonymous block sequences)
	 */
	public LiuTerziDegreeAnonymizer(final CompactDegreeSequence seq, final int k, final int numberOfStoredSolutions)
	{
		this.initialize(seq, k, numberOfStoredSolutions);
	}

	/**
	 * perform simple dynamic program on degree sequence
	 */
	private void anonymize()
	{
		for (int i = 0; i < this.table.length; i++)
		{
			if (i < (this.k - 1))
			{
				this.table[i] = Integer.MAX_VALUE;
			} else if (i < ((2 * this.k) - 1))
			{
				this.table[i] = this.cost(0, i);
			} else
			{
				int min = Integer.MAX_VALUE;
				for (int t = Math.max(this.k - 1, i - (2 * this.k)); t <= (i - this.k); t++)
				{
					min = Math.min(min, this.table[t] + this.cost(t + 1, i));
				}
				this.table[i] = min;
			}
		}
		this.tableComputed = true;
	}

	private void computeChanges(final int i, final int j, final int[] changesPerDegree)
	{
		for (int l = i; l < j; l++)
		{
			changesPerDegree[this.compressedDegreeSequence[j]]++;
			changesPerDegree[this.compressedDegreeSequence[l]]--;
		}
	}

	private void computeSolutions(final int index, final int[] changesPerDegree)
	{
		if (this.sol.size() > this.maxNumberOfStoredSolutions)
			return;

		if (index < (this.k - 1))
			throw new IllegalAccessError();
		else if (index < ((2 * this.k) - 1))
		{
			if (this.table[index] == this.cost(0, index))
			{
				this.computeChanges(0, index, changesPerDegree);
				this.examineSequence(changesPerDegree);
			} else
				throw new IllegalAccessError();
		} else
		{
			boolean foundSol = false;
			for (int t = Math.max(this.k - 1, index - (2 * this.k)); t <= (index - this.k); t++)
			{
				final int c = this.cost(t + 1, index);
				if (((this.table[t] + c) == this.table[index]) && (!foundSol || (c > 0)))
				{
					if (c == 0)
					{
						foundSol = true;
					}
					final int[] tmp = Arrays.copyOf(changesPerDegree, changesPerDegree.length);
					this.computeChanges(t + 1, index, tmp);
					this.computeSolutions(t, tmp);
				}
			}
		}
	}

	/**
	 * compute the cost to bring all degrees in <code>compressedDegreeSequence</code> in from index <code>i</code> to index <code>j</code> to the same
	 * degree, that is <code>compressedDegreeSequence[j]</code>
	 * 
	 * @param i
	 * @param j
	 * @return
	 */
	private int cost(final int i, final int j)
	{
		int cost = 0;
		for (int l = i; l < j; l++)
		{
			cost += this.compressedDegreeSequence[j] - this.compressedDegreeSequence[l];
		}
		return cost;
	}

	private void examineSequence(final int[] changesPerDegree)
	{
		final int[] newBlockSeq = new int[this.blockSequence.getMaxDegree() + 1];
		for (int i = 0; i <= this.blockSequence.getMaxDegree(); i++)
		{
			newBlockSeq[i] = this.blockSequence.get(i) + changesPerDegree[i];
			if (newBlockSeq[i] < 0)
				throw new IllegalAccessError();
		}
		this.sol.add(new CompactDegreeSequence(newBlockSeq, false));
	}

	public int getK()
	{
		return this.k;
	}

	/**
	 * Computes the anonymization cost for k-anonymizing the block-sequence containing all blocks from degree zero to degree <code>degree</code> minus
	 * the <code>t</code> highest degrees.
	 * 
	 * @param degree
	 * @param t
	 * @return
	 */
	public int getMinAnonCosts(final int degree, final int t)
	{
		if ((degree < 0) || (degree > this.blockSequence.getMaxDegree()))
			throw new IllegalArgumentException("Degree " + degree + " is either smaller than 0 or greater than MaxDegree");

		if ((this.lastDegreeIndex[degree] - t) == -1)
			return 0; // nothing to do; considered block sequence is empty

		if ((this.lastDegreeIndex[degree] - t) < (this.k - 1))
			return Integer.MAX_VALUE; // more than 0 but less than k degrees => impossible to anonymize

		if ((this.k == 1) || (this.blockSeqAnonymity >= this.k))
			return 0; // nothing to do

		if (!this.tableComputed)
		{
			this.anonymize();
		}

		return this.table[this.lastDegreeIndex[degree] - t];
	}

	/**
	 * compute a set with all minimum solutions
	 * 
	 * @return
	 */
	public Set<CompactDegreeSequence> getMinimumSolutions()
	{
		if (this.getMinSolSize() == Integer.MAX_VALUE)
			return null;

		if (!this.solutionsComputed)
		{
			if (this.getMinSolSize() == 0)
			{
				this.sol.add(this.blockSequence);
			} else
			{
				final int[] changesPerDegree = new int[this.blockSequence.getMaxDegree() + 1];
				for (int i = 0; i < changesPerDegree.length; i++)
				{
					changesPerDegree[i] = 0;
				}
				this.computeSolutions(this.table.length - 1, changesPerDegree);
			}
			this.solutionsComputed = true;
		}

		return this.sol;
	}

	/**
	 * get the size of a minimum solution
	 * 
	 * @return
	 */
	public int getMinSolSize()
	{
		return this.getMinAnonCosts(this.blockSequence.getMaxDegree(), 0);
	}

	/**
	 * Initialize with new sequence (multiple initializations with different block-sequences possible; last initialized block-sequence is considered)
	 * 
	 * @param seq
	 * @param k
	 */
	public void initialize(final CompactDegreeSequence seq, final int k)
	{
		this.initialize(seq, k, this.maxNumberOfStoredSolutions); // do not change maxNumberOfStoredSolutions
	}

	/**
	 * Initialize with new sequence (multiple initializations with different block-sequences possible; last initialized block-sequence is considered)
	 * 
	 * @param seq
	 * @param k
	 * @param maxNumberOfSolution
	 */
	public void initialize(final CompactDegreeSequence seq, final int k, final int maxNumberOfSolution)
	{
		this.maxNumberOfStoredSolutions = maxNumberOfSolution;
		this.k = k;
		this.blockSequence = new CompactDegreeSequence(seq.getDegreeCopy(), false);
		this.blockSeqAnonymity = this.blockSequence.getLevelOfAnonymity();
		this.lastDegreeIndex = new int[this.blockSequence.getMaxDegree() + 1];

		int compressedSize = 0;

		for (int i = 0; i <= this.blockSequence.getMaxDegree(); i++)
		{
			compressedSize += Math.min(3 * k, this.blockSequence.get(i));
		}
		this.compressedDegreeSequence = new int[compressedSize];
		this.table = new int[compressedSize];

		int newIndex = 0;
		for (int i = 0; i <= this.blockSequence.getMaxDegree(); i++)
		{
			for (int j = 0; j < Math.min(3 * k, this.blockSequence.get(i)); j++)
			{
				this.compressedDegreeSequence[newIndex] = i;
				newIndex++;
			}
			this.lastDegreeIndex[i] = newIndex - 1;
		}
		this.tableComputed = false;
		this.solutionsComputed = false;
		this.sol = new HashSet<CompactDegreeSequence>();
	}

}
