package algorithm;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.Arrays;
import java.util.BitSet;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.lang3.mutable.MutableInt;

import algorithm.dataStructures.CompactDegreeSequence;
import algorithm.dataStructures.GraphInterface;
import algorithm.dataStructures.Triple;
import benchmarks.LogStringBuilder;
import benchmarks.MainCLIAnonymizerAndAlternativeRealizer;
import dataStructures.Store;

public class SolutionDegreeIterator
{
	/**
	 * HopeTest is not applied if currently considered degree in dynamic program is less than Max (maxDegree / applyHopeTestLowerBound,
	 * applyHopeTestAbsoluteLowerBound) <br>
	 * used if <code>hopeTestApplyMode<code> is in {1,3}.
	 * <p>
	 * reasonable values: 10,15,20,...,100
	 */
	public static final MutableInt applyHopeTestAbsoluteLowerBound = new MutableInt(40);

	/**
	 * HopeTest is not applied if currently considered degree in dynamic program is less than Max (maxDegree / applyHopeTestLowerBound,
	 * applyHopeTestAbsoluteLowerBound)
	 * <p>
	 * reasonable values: 2,3,...,10,20,30,50,100
	 */
	public static final MutableInt applyHopeTestLowerBound = new MutableInt(10);

	/**
	 * the HopeAdvancedTest is applied in every {@code constantAdvancedInHope}^th hopeTest  <br>
	 * used if <code>hopeTestApplyMode<code> is in {1,3,4}.
	 * <p>
	 * reasonable values: 1,2,....,10,15,20,50
	 */
	public static final MutableInt constantAdvancedInHope = new MutableInt(100);

	/**
	 * In which depth is HopeTest applied?!
	 * used if <code>hopeTestApplyMode<code> is in {1,3}.
	 * <p>
	 * reasonable values: 1,2,...,20,25,30,40,50
	 */
	public static final MutableInt constantHowOftenHopeForIsApplied = new MutableInt(5); // Integer.MAX_VALUE;

	/**
	 * different modi for deciding, when to apply the HopeTest
	 * 
	 * <p>
	 * Values:<br>
	 * 1 - depth dependent. Controlled via <code>constantHowOftenHopeForIsApplied</code> and <code>constantAdvancedInHope</code> <br>
	 * 2 - time adaptive mode. When the HopeTest consumes most time, then it is disabled. When the examineSequence takes more time, then the HopeTest
	 * is applied <br>
	 * 3 - mixture of first two modi where the depth and the time must fit <br>
	 * 4 - dependent of the degree sequence 
	 */
	public static final MutableInt hopeTestApplyMode = new MutableInt(4);

	/**
	 * Flag to enable or disable the last HopeTest successful mechanics
	 */
	private static final boolean enableLastStepSuccessfulOption = true;

	/**
	 * should the advanced hope test be performed when a considered instance could 
	 * not improve the upper bound?
	 */
	private static final boolean setAdvancedStepSuccessfulIfExaminFailed = false;

	/**
	 * only used if <code>hopeTestApplyMode<code> = 4.
	 * reasonable values: 2,3,...,1000
	 */
	public static final MutableInt maxHopeTestsInOneBranch = new MutableInt(3);

	public static final MutableInt noOfStoredSolutions = new MutableInt(200);

	public static boolean printLogForEachCostValue = false;
	
	public static String showNonEmptyEntries(final int[] arr)
	{
		final StringBuffer out = new StringBuffer();
		boolean first = true;
		int sum = 0;
		out.append("[ ");
		for (int i = 0; i < arr.length; i++)
		{
			if (arr[i] > 0)
			{
				out.append((!first ? ", " : "") + i + ": " + arr[i]);
				first = false;
				sum += arr[i];
			}
		}
		out.append(" ], sum: " + sum);
		return out.toString();
	}
	
	private boolean lastAdvancedStepSuccessful = false;

	private boolean lastStepSuccessful = false;

	private int alreadyConsiderCosts;

	/**
	 * stored Depths where HopeTest is applied.
	 */
	private Set<Integer> applyHopeTestInWhichDepths, applyAdvancedHopeTestInWhichDepths;

	private final ErdosGallaiInterface erdosGallai;

	private final boolean findRealizableSequence;

	private final boolean fixDegreeOneVerticesInErdösGallai;

	private final GraphInterface graph;

	/**
	 * arrays for counting for testing purpose
	 */
	private final int[] hopeDestroyedAtDepth, hopeDestroyedAtDepthAdvanced;

	private final int k;

	private int minHopeDestroyedDepth = Integer.MAX_VALUE, minHopeDestroyedDepthAdvanced = Integer.MAX_VALUE;

	private long noOfAEGHopeDestroyed = 0;

	private long noOfAEGHopeTests = 0;

	private long noOfAEGnegatives = 0;

	private long numberOfRecursivelyConsideredInstances = 0;

	private final CompactDegreeSequence originalSequence;

	private final int preProcessCosts;

	private final CompactDegreeSequence preProcessedSequence;

	private final int solutionsStoredInCostWaster;

	private final BitSet[][] table;

	private long timeForExamineSequence = 0, timeForHopeTest = 0, timeForAdvancedHopeTest = 0;

	private Store<Integer, Set<CompactDegreeSequence>> upperBound;
	
	private long timeStampOfLastLog = 0;

	public SolutionDegreeIterator(final GraphInterface graph, final BitSet[][] table2, final CompactDegreeSequence orginalSequence,
			final CompactDegreeSequence preProcessedSequence, final int k, final boolean findRealizableSequence, final boolean fixDegreeOn,
			final int solutionsStoredInCostWaster)
	{
		this.graph = graph;
		this.table = table2;
		this.originalSequence = orginalSequence;
		this.preProcessedSequence = preProcessedSequence;
		this.preProcessCosts = this.preProcessedSequence.difference(orginalSequence).getTotalDegreeSum();
		this.k = k;
		this.alreadyConsiderCosts = -1;
		this.upperBound = new Store<Integer, Set<CompactDegreeSequence>>(Integer.MAX_VALUE, null);
		this.findRealizableSequence = findRealizableSequence;
		this.fixDegreeOneVerticesInErdösGallai = fixDegreeOn;
		this.erdosGallai = new FastErdosGallai(graph, orginalSequence);
		this.hopeDestroyedAtDepth = new int[orginalSequence.getMaxDegree() + 1];
		this.hopeDestroyedAtDepthAdvanced = new int[orginalSequence.getMaxDegree() + 1];
		for (int i = 0; i < this.hopeDestroyedAtDepth.length; i++)
		{
			this.hopeDestroyedAtDepth[i] = 0;
			this.hopeDestroyedAtDepthAdvanced[i] = 0;
		}
		this.initializeDepthValuesForHopeTest();
		this.solutionsStoredInCostWaster = solutionsStoredInCostWaster;
	}

	private int[] constructSequenceFromTriples(final Triple[] triples)
	{
		final int[] newsequence = new int[this.preProcessedSequence.getMaxDegree() + 1];
		Arrays.fill(newsequence, 0);

		int c = 0;
		for (int i = 0; i <= this.preProcessedSequence.getMaxDegree(); i++)
		{
			if (triples[i] != null)
			{
				if (triples[i].getT() >= this.preProcessedSequence.get(i))
				{
					c += this.preProcessedSequence.get(i);
				} else
				{
					newsequence[i] = (this.preProcessedSequence.get(i) - triples[i].getT()) + c;
					c = triples[i].getT();
				}
			} else
			{
				if (c > 0)
					throw new IllegalArgumentException("illegal triple");
				newsequence[i] = this.preProcessedSequence.get(i);
			}
		}
		return newsequence;
	}

	/**
	 * Given a list of triples (a "chain" in the dynamic programming table pointing to a certain raw solution), computes all solution degree
	 * sequences.
	 * 
	 * @param triples
	 */
	private void examineSequence(final Triple[] triples)
	{
		this.numberOfRecursivelyConsideredInstances++;
		final long time = System.currentTimeMillis();

		int storedSolutions = 0;
		if (this.upperBound.getS2() != null)
		{
			storedSolutions = this.upperBound.getS2().size();
		}

		final CompactDegreeSequence newCompactSequence = new CompactDegreeSequence(this.constructSequenceFromTriples(triples), false);
		final CompactDegreeSequence differenceSeq = newCompactSequence.difference(this.originalSequence);
		final int degreeSum = differenceSeq.getTotalDegreeSum();
		int minAddCost;

		if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 4)
		{
			System.out.println("============Begin CostWaster==============");
			System.out.println("OrigSeq: " + this.preProcessedSequence);
			System.out.println(Arrays.toString(triples));
			System.out.println("anonSeq:\t" + newCompactSequence);
			System.out.println("cost so far:\t" + degreeSum);
			System.out.println("upper bound:\t" + this.upperBound);
		}

		if (!this.findRealizableSequence
				|| (differenceSeq.isRealizable() && ((this.graph == null) || (this.erdosGallai.computeFeasibility(newCompactSequence,
						this.fixDegreeOneVerticesInErdösGallai) <= 0))))
		{
			if (degreeSum < this.upperBound.getS1())
			{
				final HashSet<CompactDegreeSequence> s2 = new HashSet<CompactDegreeSequence>();
				this.upperBound = new Store<Integer, Set<CompactDegreeSequence>>(degreeSum, s2);
				s2.add(newCompactSequence);
			} else if ((degreeSum == this.upperBound.getS1()) && this.storeFurtherSolution())
			{
				this.upperBound.getS2().add(newCompactSequence);
			}

		} else if ((degreeSum + (minAddCost = differenceSeq.noOfAddDegreeOneVerticesForBeingRealizableWRSimpleErdös())) <= (this.upperBound.getS1() - this
				.reduceDueReachedSolutionsLimit()))
		{
			for (int addCost = minAddCost; (addCost + degreeSum) <= (this.upperBound.getS1() - this.reduceDueReachedSolutionsLimit()); addCost++)
			{
				// System.out.println("addcost:"+addCost);
				final CostWaster cw = new CostWaster(this.originalSequence, newCompactSequence, addCost, this.k, true,
						this.solutionsStoredInCostWaster);
				Set<CompactDegreeSequence> t;
				if (((t = cw.getAllRealizableSequences()) != null) && (t.size() > 0))
				{
					boolean found = false;
					for (final CompactDegreeSequence sol : t)
					{
						// System.out.println(differenceSeq.getTotalDegreeSum()
						// + addCost);
						long erdosGallaiAdvanced = 1;
						if ((this.graph == null)
								|| ((erdosGallaiAdvanced = this.erdosGallai.computeFeasibility(sol, this.fixDegreeOneVerticesInErdösGallai)) == 0))
						{
							found = true;
							if (this.upperBound.getS1() > (differenceSeq.getTotalDegreeSum() + addCost))
							{
								final HashSet<CompactDegreeSequence> s = new HashSet<CompactDegreeSequence>();
								s.add(sol);
								this.upperBound = new Store<Integer, Set<CompactDegreeSequence>>(differenceSeq.getTotalDegreeSum() + addCost, s);
							} else if (this.storeFurtherSolution())
							{
								this.upperBound.getS2().add(sol);
							}
						} else
						{
							this.noOfAEGnegatives++;
							addCost += erdosGallaiAdvanced - 1; // -1 due to
																// increase of
																// addCost in
																// next run
																// through the
																// loop
						}

					}
					if (found)
					{
						break;
					}
				}
			}
		}

		if ((this.upperBound.getS2() != null) && (this.upperBound.getS2().size() == storedSolutions))
		{ // no further solutions stored in current step => apply hope-test
			// again!
			this.lastStepSuccessful = true;
			if (setAdvancedStepSuccessfulIfExaminFailed)
				this.lastAdvancedStepSuccessful = true;
		}

		this.timeForExamineSequence += System.currentTimeMillis() - time;
	}

	public Set<CompactDegreeSequence> getAllRealizableSequences(final int maximumCosts)
	{
		if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 1)
		{
			final DecimalFormat myFormatter = new DecimalFormat("###,###,###.###");
			System.out.println("maxCost:" + maximumCosts
//					+ "\talreadyKnownUpperBound:" + this.upperBound.getS1()
					+ "\tStoredSolutions: "	+ (Integer.MAX_VALUE > this.upperBound.getS1() ? this.upperBound.getS2().size() : "0")
					// + "\talreadyConsideredCosts:" + this.alreadyConsiderCosts
					+ "\tinstances considered: " + myFormatter.format(this.numberOfRecursivelyConsideredInstances) 
					+ "\thope tests: "+ myFormatter.format(this.noOfAEGHopeTests) 
					+ "\tdropped: " + myFormatter.format(this.noOfAEGHopeDestroyed) 
//					+ "\tAEG negatives: " + myFormatter.format(this.noOfAEGnegatives) 
					+ "\tEx-Time: " + myFormatter.format(this.timeForExamineSequence / 1000d) + "s"
					+ "\tHope-Time: " + myFormatter.format(this.timeForHopeTest / 1000d) + "s" + "\tAHope-Time: "
					+ myFormatter.format(this.timeForAdvancedHopeTest / 1000d) + "s");
			// System.out.println("AEG-Calls: " +
			// myFormatter.format(this.erdosGallai.getNoOfEGCalls()) +
			// "\tTime1: ("
			// + myFormatter.format(this.erdosGallai.getTime1() / 1000d) +
			// "sec) \tTime2: ("
			// + myFormatter.format(this.erdosGallai.getTime2() / 1000d) +
			// "sec) \tTime3: ("
			// + myFormatter.format(this.erdosGallai.getTime3() / 1000d) +
			// "sec) \tHashMapSize: "
			// + myFormatter.format(this.erdosGallai.getHashMapSize()) +
			// "\tlookups: "
			// + myFormatter.format(this.erdosGallai.getNoOfLookUps()) +
			// "\toverall: "
			// + myFormatter.format(this.erdosGallai.getEdgesBetweenCalls()));
			// System.out.println("SolDepComps: " +
			// myFormatter.format(this.erdosGallai.getNoOfSolDependingComputations())
			// + "\tSolDepLookUps: " +
			// myFormatter.format(this.erdosGallai.getNoOfSolDependingLookUps())
			// + "\tSolDepAll: "
			// + myFormatter.format(this.erdosGallai.getSolDependingChecks()) +
			// "\tSolDepMaxSize: "
			// +
			// myFormatter.format(this.erdosGallai.getMaxSolDepHashMapSize()));
			System.out.println("");

			// LogStringBuilder.getLogger().addEntry("minDepth",
			// minHopeDestroyedDepth);
			// if (minHopeDestroyedDepth > 0 && minHopeDestroyedDepth <
			// Integer.MAX_VALUE)
			// {
			// LogStringBuilder.getLogger().addEntry("#vertices1",
			// preProcessedSequence.getNumberOfVerticesInBags(minHopeDestroyedDepth,
			// preProcessedSequence.getMaxValue()));
			// LogStringBuilder.getLogger().addEntry(
			// "ratio1",
			// 100.0f *
			// preProcessedSequence.getNumberOfVerticesInBags(minHopeDestroyedDepth,
			// preProcessedSequence.getMaxValue())
			// / (1.0f * preProcessedSequence.getTotalNumberOfVertices()));
			// }
			//
			// LogStringBuilder.getLogger().addEntry("minDepthAdvanced",
			// minHopeDestroyedDepthAdvanced);
			// if (minHopeDestroyedDepthAdvanced > 0 &&
			// minHopeDestroyedDepthAdvanced < Integer.MAX_VALUE)
			// {
			// LogStringBuilder.getLogger().addEntry("#vertices2",
			// preProcessedSequence.getNumberOfVerticesInBags(minHopeDestroyedDepthAdvanced,
			// preProcessedSequence.getMaxValue()));
			// LogStringBuilder.getLogger().addEntry(
			// "ratio2",
			// 100.0f
			// *
			// preProcessedSequence.getNumberOfVerticesInBags(minHopeDestroyedDepthAdvanced,
			// preProcessedSequence.getMaxValue()) / (1.0f *
			// preProcessedSequence.getTotalNumberOfVertices()));
			// }
			//
			// LogStringBuilder.getLogger().addEntry("maxDegPreprocessed",
			// preProcessedSequence.getMaxValue());
			// LogStringBuilder.getLogger()
			// .addEntry("hopeDestroyedAtDepth",
			// SolutionDegreeIterator2.showNonEmptyEntries(hopeDestroyedAtDepth));
			// LogStringBuilder.getLogger().addEntry("hopeDestroyedAtDepthAdvanced",
			// SolutionDegreeIterator2.showNonEmptyEntries(hopeDestroyedAtDepthAdvanced));
			// try {
			// LogStringBuilder.getLogger().printOutput(System.out);
			// } catch (IOException e) {
			// e.printStackTrace();
			// }
		}

		// TODO dead code??
		if ((this.alreadyConsiderCosts >= maximumCosts) && ((this.upperBound.getS1() - this.preProcessCosts) <= maximumCosts))
			return this.upperBound.getS2();
		if (this.alreadyConsiderCosts >= maximumCosts)
			throw new IllegalAccessError();

		while (this.alreadyConsiderCosts < maximumCosts)
		{
			this.alreadyConsiderCosts++;
			int cost;
			if ((cost = this.table[this.preProcessedSequence.getMaxDegree()][0].nextSetBit(this.alreadyConsiderCosts)) >= 0)
			{
				if (cost > maximumCosts)
				{
					this.alreadyConsiderCosts = maximumCosts;
					break;
				} else
				{
					this.alreadyConsiderCosts = cost;
					final Triple[] triples = new Triple[this.preProcessedSequence.getMaxDegree() + 1];
					// this.constantHowOftenHopeForIsApplied = (triples.length +
					// 4) / 3;
					triples[this.preProcessedSequence.getMaxDegree()] = new Triple(this.preProcessedSequence.getMaxDegree(), 0,
							this.alreadyConsiderCosts);
					this.recursivelyBuildAndExamineAllSolutionSequences(this.preProcessedSequence.getMaxDegree(), triples);
				}
			}
		}

		LogStringBuilder.getLogger().addEntry("totalgetNoOfAEGHopeTests", this.noOfAEGHopeTests);
		LogStringBuilder.getLogger().addEntry("instancesConsidered", this.numberOfRecursivelyConsideredInstances);
		LogStringBuilder.getLogger().addEntry("hopeforSuccussRate", this.getHopeTestSuccessRate());
		LogStringBuilder.getLogger().addEntry("lastUpperBoundInSolutionIterator", this.getCurrentUpperbound());

		
		// only log when trying to realize and log only all 5 seconds
		if (printLogForEachCostValue && this.findRealizableSequence && System.currentTimeMillis() - timeStampOfLastLog > 5000) 
		{
			try {
				LogStringBuilder.getLogger().printOutput(System.out);
				
				final FileOutputStream outStream = new FileOutputStream(new File("singleStep.log"), true);
				LogStringBuilder.getLogger().printOutput(outStream);
				outStream.close();
				timeStampOfLastLog = System.currentTimeMillis();
//				this.noOfAEGHopeTests = 0;
//				this.noOfAEGHopeDestroyed = 0;
//				this.numberOfRecursivelyConsideredInstances = 0;
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		
		if ((this.upperBound.getS1() - this.preProcessCosts) <= maximumCosts)
			return this.upperBound.getS2();
		return null;
	}

	public float getHopeTestSuccessRate()
	{
		if (this.noOfAEGHopeTests > 1)
			return 100f * ((float) this.noOfAEGHopeDestroyed / this.noOfAEGHopeTests);
		return 0;
	}
	
	public int getCurrentUpperbound()
	{
		return this.upperBound.getS1();
	}

	public long getNoOfAEGHopeDestroyed()
	{
		return this.noOfAEGHopeDestroyed;
	}

	public long getNoOfAEGHopeTests()
	{
		return this.noOfAEGHopeTests;
	}

	public long getNumberOfRecConsideredInstances()
	{
		return this.numberOfRecursivelyConsideredInstances;
	}

	private boolean hopeForDecreasingUpperBound(final Triple[] triples, final int depth)
	{
		if (!this.performHopeTest(depth, false))
			return true;

		long time = System.currentTimeMillis();

		// if ( !this.findRealizableSequence || !this.lastStepSuccessful &&
		// Integer.numberOfLeadingZeros(depth) +
		// Integer.numberOfTrailingZeros(depth) > 30 )
		// return true;

		final CompactDegreeSequence newCompactSequence = new CompactDegreeSequence(this.constructSequenceFromTriples(triples), false);
		final CompactDegreeSequence differenceSeq = newCompactSequence.difference(this.originalSequence);

		final int degreeSum = differenceSeq.getTotalDegreeSum();
		int t = degreeSum + differenceSeq.noOfAddDegreeOneVerticesForBeingRealizableWRSimpleErdös();
		boolean hopeForImprovingUpperbound = t <= (this.upperBound.getS1() - this.reduceDueReachedSolutionsLimit());

		this.noOfAEGHopeDestroyed += !hopeForImprovingUpperbound ? 1 : 0;
		this.noOfAEGHopeTests++;

		// TODO successfull rate!!!add

		this.lastStepSuccessful = !hopeForImprovingUpperbound;
		this.timeForHopeTest += System.currentTimeMillis() - time;

		if (!hopeForImprovingUpperbound)
		{
			this.hopeDestroyedAtDepth[depth]++;
			this.minHopeDestroyedDepth = Math.min(this.minHopeDestroyedDepth, depth);
		}

		if (!hopeForImprovingUpperbound || (triples[depth].getT() != 0) || !this.performHopeTest(depth, true))
			// || depth % (constantAdvancedInHope *
			// constantHowOftenHopeForIsApplied) != 0)
			return hopeForImprovingUpperbound;

		time = System.currentTimeMillis();

		final int advancedAEGbound = this.erdosGallai.computeFeasibility(newCompactSequence, this.fixDegreeOneVerticesInErdösGallai, true, true);
		t = degreeSum + advancedAEGbound;
		hopeForImprovingUpperbound = t <= (this.upperBound.getS1() - this.reduceDueReachedSolutionsLimit());
		this.noOfAEGHopeDestroyed += !hopeForImprovingUpperbound ? 1 : 0;

		this.lastAdvancedStepSuccessful = !hopeForImprovingUpperbound;
		this.timeForAdvancedHopeTest += System.currentTimeMillis() - time;

		if (!hopeForImprovingUpperbound)
		{
			this.erdosGallai.computeFeasibility(newCompactSequence, this.fixDegreeOneVerticesInErdösGallai);
			this.hopeDestroyedAtDepthAdvanced[depth]++;
			this.minHopeDestroyedDepthAdvanced = Math.min(this.minHopeDestroyedDepthAdvanced, depth);
		}
		return hopeForImprovingUpperbound;

	}

	private void initializeDepthValuesForHopeTest()
	{
		if (!this.findRealizableSequence)
			return;
		this.applyHopeTestInWhichDepths = new HashSet<Integer>();
		this.applyAdvancedHopeTestInWhichDepths = new HashSet<Integer>();
		int counter = 1;

		int lastValue = -1;

		final float lowerBoundOnDepthForApplyHopeTest = Math.max(this.preProcessedSequence.getMaxDegree()
				/ (float) SolutionDegreeIterator.applyHopeTestLowerBound.intValue(),
				SolutionDegreeIterator.applyHopeTestAbsoluteLowerBound.intValue());

		final int lowerBound = (int) lowerBoundOnDepthForApplyHopeTest;
		if (lowerBound > this.preProcessedSequence.getMaxDegree())
			return;

		final int noOfVerticesInRange = this.preProcessedSequence.getNumberOfVerticesInBags(lowerBound, this.preProcessedSequence.getMaxDegree());
		int noOfVerticesConsidered = 0;

		for (int i = this.preProcessedSequence.getMaxDegree(); i > lowerBoundOnDepthForApplyHopeTest; i--)
		{
			if (this.preProcessedSequence.get(i) > 0)
			{
				if ((SolutionDegreeIterator.hopeTestApplyMode.intValue() == 1) || (SolutionDegreeIterator.hopeTestApplyMode.intValue() == 3))
				{
					lastValue = i;
					if ((counter % SolutionDegreeIterator.constantHowOftenHopeForIsApplied.intValue()) == 0)
					{
						this.applyHopeTestInWhichDepths.add(i);
					}
					if ((counter % (SolutionDegreeIterator.constantHowOftenHopeForIsApplied.intValue() * SolutionDegreeIterator.constantAdvancedInHope
							.intValue())) == 0)
					{
						this.applyAdvancedHopeTestInWhichDepths.add(i);
					}
					counter++;
				} else if (SolutionDegreeIterator.hopeTestApplyMode.intValue() == 4)
				{
					noOfVerticesConsidered += this.preProcessedSequence.get(i);
					if (((noOfVerticesInRange * counter) / SolutionDegreeIterator.maxHopeTestsInOneBranch.intValue()) <= noOfVerticesConsidered)
					{
						this.applyHopeTestInWhichDepths.add(i);
						if ((counter % SolutionDegreeIterator.constantAdvancedInHope.intValue()) == 0)
						{
							this.applyAdvancedHopeTestInWhichDepths.add(i);
						}
						counter++;
					}
				}
			}
		}
		if (lastValue > 0)
		{
			this.applyHopeTestInWhichDepths.add(lastValue);
			this.applyAdvancedHopeTestInWhichDepths.add(lastValue);
		}
		// System.out.println(applyHopeTestInWhichDepths);
		// System.out.println(applyAdvancedHopeTestInWhichDepths);
		// System.out.println(preProcessedSequence);
		// System.out.println(preProcessedSequence.getNumberOfVerticesInBags(lastValue,preProcessedSequence.getMaxValue()));
		// System.out.println(preProcessedSequence.getTotalNumberOfVertices());
	}

	private boolean performHopeTest(final int depth, final boolean advanced)
	{
		if (!this.findRealizableSequence) // independent of the
											// hopeTestApplyMode
			return false;

		if (this.enableLastStepSuccessfulOption && (this.lastAdvancedStepSuccessful || this.lastStepSuccessful))
			return true;

		if ((SolutionDegreeIterator.hopeTestApplyMode.intValue() == 1) || (SolutionDegreeIterator.hopeTestApplyMode.intValue() == 3)
				|| (SolutionDegreeIterator.hopeTestApplyMode.intValue() == 4))
		{
			if ((this.preProcessedSequence.getMaxDegree() > (SolutionDegreeIterator.applyHopeTestLowerBound.intValue() * depth))
					&& (SolutionDegreeIterator.applyHopeTestAbsoluteLowerBound.intValue() > depth))
				return false;

			final boolean performTest = (advanced && this.applyAdvancedHopeTestInWhichDepths.contains(depth))
					|| (!advanced && this.applyHopeTestInWhichDepths.contains(depth));

			if ((SolutionDegreeIterator.hopeTestApplyMode.intValue() == 1) || (SolutionDegreeIterator.hopeTestApplyMode.intValue() == 4)
					|| !performTest)
				return performTest; // mode 1 and 4 just depend on depth

			if (advanced)
				return this.timeForAdvancedHopeTest < (3 * this.timeForExamineSequence);
			// perform test when not much time spent on test so far
			else
				return this.timeForHopeTest < (3 * this.timeForExamineSequence);
			// perform test when not much time spent on test so far
		} else if (SolutionDegreeIterator.hopeTestApplyMode.intValue() == 2)
		{
			if (advanced)
				return this.timeForAdvancedHopeTest < (3 * this.timeForExamineSequence);
			// perform test when not much time spent on test so far
			else
				return this.timeForHopeTest < (3 * this.timeForExamineSequence);
			// perform test when not much time spent on test so far
		}
		return false;
	}

	private void recursivelyBuildAndExamineAllSolutionSequences(final int i, final Triple[] triples)
	{
		// for (int p = 0; p <= i; p++)
		// System.out.print(" ");
		// System.out.println(i);
		if (i <= 0)
		{
			this.examineSequence(triples);
		} else if (this.hopeForDecreasingUpperBound(triples, i))
		{
			final int t = triples[i].getT();
			final int c = triples[i].getC();

			if (this.preProcessedSequence.get(i) > t)
			{
				if ((this.preProcessedSequence.get(i) - t) >= this.k)
				{
					if (this.table[i - 1][0].get(c))
					{
						triples[i - 1] = new Triple(i - 1, 0, c);
						this.recursivelyBuildAndExamineAllSolutionSequences(i - 1, triples);
						triples[i - 1] = null;
					}

					int j = i - 1;
					int counter = 0;
					int oldCounter = 0;
					while ((j >= 0) && ((counter += this.preProcessedSequence.get(j)) < this.k)
							&& (DegreeAnonymizer.cost(this.preProcessedSequence, i - 1, counter, i) <= c))
					{
						final int remCost = c - DegreeAnonymizer.cost(this.preProcessedSequence, i - 1, counter, i);
						triples[j] = new Triple(j, this.preProcessedSequence.get(j), -1);
						if ((counter > oldCounter)
								&& DegreeAnonymizer.getTableEntry(this.preProcessedSequence, this.table, this.k, j, this.preProcessedSequence.get(j),
										remCost))
						{
							if (j > 0)
							{
								triples[j - 1] = new Triple(j - 1, 0, remCost);
							}
							this.recursivelyBuildAndExamineAllSolutionSequences(j - 1, triples);
						}
						j--;
						oldCounter = counter;
					}
					Arrays.fill(triples, 0, i, null);
				}
				// 0 < sequence.get(i) - t < k
				else if (DegreeAnonymizer.cost(this.preProcessedSequence, i - 1, this.k - (this.preProcessedSequence.get(i) - t), i) <= c)
				{
					int j = i - 1;
					final int iniJ = j;
					int counter = 0;
					final int noElementsToFillUpBag_i = this.k - (this.preProcessedSequence.get(i) - t);
					while ((j >= 0) && ((counter + this.preProcessedSequence.get(j)) < noElementsToFillUpBag_i))
					{
						counter += this.preProcessedSequence.get(j);
						triples[j] = new Triple(j, this.preProcessedSequence.get(j), -1);
						j--;
					}
					if (j >= 0)
					{
						final int neededElementsFromBag_j = noElementsToFillUpBag_i - counter;
						if (DegreeAnonymizer.getTableEntry(this.preProcessedSequence, this.table, this.k, j, neededElementsFromBag_j, c
								- DegreeAnonymizer.cost(this.preProcessedSequence, i - 1, noElementsToFillUpBag_i, i)))
						{
							triples[j] = new Triple(j, neededElementsFromBag_j, c
									- DegreeAnonymizer.cost(this.preProcessedSequence, i - 1, this.k - (this.preProcessedSequence.get(i) - t), i));
							this.recursivelyBuildAndExamineAllSolutionSequences(j, triples);
							triples[j] = null;

						}

						int u = this.preProcessedSequence.get(j) - neededElementsFromBag_j;
						int oldU = 0;
						while ((u < this.k)
								&& (DegreeAnonymizer.cost(this.preProcessedSequence, i - 1, u + neededElementsFromBag_j + counter, i) <= c))
						{
							triples[j] = new Triple(j, this.preProcessedSequence.get(j), -1);
							if ((u > oldU)
									&& DegreeAnonymizer.getTableEntry(this.preProcessedSequence, this.table, this.k, j,
											this.preProcessedSequence.get(j),
											c - DegreeAnonymizer.cost(this.preProcessedSequence, i - 1, u + neededElementsFromBag_j + counter, i)))
							{
								if (j > 0)
								{
									triples[j - 1] = new Triple(j - 1, 0, c
											- DegreeAnonymizer.cost(this.preProcessedSequence, i - 1, u + neededElementsFromBag_j + counter, i));
								}
								this.recursivelyBuildAndExamineAllSolutionSequences(j - 1, triples);
							}

							j--;
							if (j >= 0)
							{
								oldU = u;
								u += this.preProcessedSequence.get(j);
							} else
							{
								break;
							}
						}
						Arrays.fill(triples, 0, iniJ, null);
					}
				}
			}
			// sequence.get(i) - t <= 0
			else
			{
				triples[i - 1] = new Triple(i - 1, t - this.preProcessedSequence.get(i), c);
				this.recursivelyBuildAndExamineAllSolutionSequences(i - 1, triples);
				triples[i - 1] = null;
			}

		}
	}

	private int reduceDueReachedSolutionsLimit()
	{
		if ((this.upperBound.getS2() == null) || (SolutionDegreeIterator.noOfStoredSolutions.intValue() > this.upperBound.getS2().size()))
			return 0;

		// if (Long.numberOfLeadingZeros(numberOfRecursivelyConsideredInstances)
		// + Long.numberOfTrailingZeros(numberOfRecursivelyConsideredInstances)
		// > 57)
		// return 0;
		return 1;
	}

	private boolean storeFurtherSolution()
	{
		if (this.reduceDueReachedSolutionsLimit() == 0)
		{
			LogStringBuilder.getLogger().addEntry("LastStoredSolInCacheAtCost", this.alreadyConsiderCosts);
			return true;
		}
		return false;
	}
}
