package algorithm.alternativeRealizer;

import gnu.trove.set.TIntSet;

import java.util.Map;

import algorithm.dataStructures.CompactDegreeSequence;
import algorithm.dataStructures.GraphInterface;
import algorithm.dataStructures.Vertex;
import benchmarks.MainCLIAnonymizerAndAlternativeRealizer;

/**
 * Graph constrcutor class for the k-anonymity problem. Various strategies are implemented to transfer a compact degree sequence to a graph structure.
 * If realization is successful, the resulting graph's vertex degrees correspond to the entries within the compact degree solution sequence.
 * Otherwise, the input graph remains unchanged.
 * 
 * For power-law distributed graphs, we recommend using SELECTION_STRATEGY_RANDOM with 20 passes and PROCESSING_STRATEGY_MAXDIFF_REFINED for the best
 * trade-off between efficiency and effectivity.
 * 
 * @author Clemens Hoffmann
 * 
 */
public class GreedyGraphConstructor
{

	/**
	 * Add edges between vertices with largest difference between initial and target degree first
	 */
	public final static int PROCESSING_STRATEGY_MAXDIFF_FIRST = 1;
	/**
	 * Add edges between vertices with largest difference between initial and target degree first (recompute order after adding an edge)
	 */
	public final static int PROCESSING_STRATEGY_MAXDIFF_REFINED = 4;
	/**
	 * Add edges between vertices with most common neighbors within selection first
	 */
	public final static int PROCESSING_STRATEGY_MAXNEIGHBORS_FIRST = 2;
	/**
	 * Add edges in random order
	 */
	public final static int PROCESSING_STRATEGY_RANDOM = 3;
	/**
	 * Number of maximum passes for PROCESSING_STRATEGY_RANDOM.
	 */
	public static int RANDOM_PROCESSING_PASSES = 20;
	/**
	 * Number of maximum passes for SELECTION_STRATEGY_RANDOM.
	 */
	public static int RANDOM_SELECTION_PASSES = 20;
	/**
	 * Vertices within selection share fewest edges as possible
	 */
	public final static int SELECTION_STRATEGY_LEAST_COMMON_NEIGHBORS = 1;
	/**
	 * Vertices within selection share fewest edges as possible (multi-pass)
	 */
	public final static int SELECTION_STRATEGY_LEAST_COMMON_NEIGHBORS_REFINED = 4;
	/**
	 * Vertices' degree within selection is minimized
	 */
	public final static int SELECTION_STRATEGY_LEAST_NEIGHBOR_DEGREE_SUM = 2;
	/**
	 * Vertices are selected randomly
	 */
	public final static int SELECTION_STRATEGY_RANDOM = 3;

	private final GraphInterface graph;

	/**
	 * 
	 * @param graph Initial, unmodified graph
	 */
	public GreedyGraphConstructor(final GraphInterface graph)
	{
		this.graph = graph;
	}

	/**
	 * Applies the differences between the initialSeq and solutionSeq compact degree sequences to the input graph.
	 * 
	 * @param initialSeq The compact degree sequence corresponding to the input graph.
	 * @param solutionSeq The k-anonymous solution degree sequence corresponding to the input graph.
	 * @param simulate Simulation mode reverts all changes made to the graph structure after construction.
	 * @param processingStrategy Determines the order in which vertices are processed during graph construction. e.g. which vertices are connected
	 *            first.
	 * @param selectionStrategy Determines the strategy to select vertices of certain degrees from the input graph, if the number of vertices of that
	 *            degree is larger than the amount of vertices which need to be modified.
	 * @return The edges which have been added to the graph during graph construction process (non-simulation-mode) or which need to be added to the
	 *         graph (simulation mode). If graph construction was not successful, the return value is null.
	 * 
	 */
	public Map<Integer, TIntSet> constructGraph(final CompactDegreeSequence initialSeq, final CompactDegreeSequence solutionSeq,
			final boolean simulate, final int processingStrategy, final int selectionStrategy)
	{

		final ModifyList modifyList = new ModifyList(this.graph);

		if (selectionStrategy == GreedyGraphConstructor.SELECTION_STRATEGY_RANDOM) // Multi-pass random selection
		{
			for (int j = 0; j < GreedyGraphConstructor.RANDOM_SELECTION_PASSES; j++)
			{
				modifyList.computeAndGetModifyList(initialSeq, solutionSeq, selectionStrategy);
				Map<Integer, TIntSet> backupList = null;

				if ((processingStrategy == GreedyGraphConstructor.PROCESSING_STRATEGY_MAXDIFF_FIRST)
						|| (processingStrategy == GreedyGraphConstructor.PROCESSING_STRATEGY_MAXDIFF_REFINED))
				{
					modifyList.sortModifyListByDegreeDifference();
					if ((backupList = new ModifyListTransferer(this.graph).transferModifylistToGraph(simulate, modifyList, processingStrategy)) != null)
						return backupList;

				}

				if (processingStrategy == GreedyGraphConstructor.PROCESSING_STRATEGY_MAXNEIGHBORS_FIRST)
				{
					modifyList.sortModifyListByCommonNeighbors();
					if ((backupList = new ModifyListTransferer(this.graph).transferModifylistToGraph(simulate, modifyList, processingStrategy)) != null)
						return backupList;
				}

				if (processingStrategy == GreedyGraphConstructor.PROCESSING_STRATEGY_RANDOM)
				{
					for (int i = 0; i < GreedyGraphConstructor.RANDOM_PROCESSING_PASSES; i++)
					{
						modifyList.randomizeModifyListPositions();
						if ((backupList = new ModifyListTransferer(this.graph).transferModifylistToGraph(simulate, modifyList, processingStrategy)) != null)
							return backupList;
					}
				}
			}
			return null;
		}

		else
		{ // Single pass

			modifyList.computeAndGetModifyList(initialSeq, solutionSeq, selectionStrategy);

			Map<Integer, TIntSet> backupList = null;

			if ((processingStrategy == GreedyGraphConstructor.PROCESSING_STRATEGY_MAXDIFF_FIRST)
					| (processingStrategy == GreedyGraphConstructor.PROCESSING_STRATEGY_MAXDIFF_REFINED))
			{
				modifyList.sortModifyListByDegreeDifference();
				if ((backupList = new ModifyListTransferer(this.graph).transferModifylistToGraph(simulate, modifyList, processingStrategy)) != null)
					return backupList;
				else
					return null;

			}

			if (processingStrategy == GreedyGraphConstructor.PROCESSING_STRATEGY_MAXNEIGHBORS_FIRST)
			{
				modifyList.sortModifyListByCommonNeighbors();
				if ((backupList = new ModifyListTransferer(this.graph).transferModifylistToGraph(simulate, modifyList, processingStrategy)) != null)
					return backupList;
				else
					return null;

			}

			if (processingStrategy == GreedyGraphConstructor.PROCESSING_STRATEGY_RANDOM)
			{
				for (int i = 0; i < GreedyGraphConstructor.RANDOM_PROCESSING_PASSES; i++)
				{
					modifyList.randomizeModifyListPositions();

					if ((backupList = new ModifyListTransferer(this.graph).transferModifylistToGraph(simulate, modifyList, processingStrategy)) != null)
						return backupList;
				}
				return null;

			}

			return null;
		}

	}

	/**
	 * Applies the differences between the initialSeq and solutionSeq compact degree sequences to the input graph.
	 * 
	 * @param initialSeq The compact degree sequence corresponding to the input graph.
	 * @param solutionSeq The k-anonymous solution degree sequence corresponding to the input graph.
	 * @param simulate Simulation mode reverts all changes made to the graph structure after construction.
	 * @param processingStrategy Determines the order in which vertices are processed during graph construction. e.g. which vertices are connected
	 *            first.
	 * @param selectionStrategy Determines the strategy to select vertices of certain degrees from the input graph, if the number of vertices of that
	 *            degree is larger than the amount of vertices which need to be modified.
	 * @param randomSelectionPasses Specifies the number of passes for SELECTION_STRATEGY_RANDOM, if this strategy has been specified. Otherwise, this
	 *            parameter is ignored.
	 * @param randomProcessingPasses Specifies the number of passes for PROCESSING_STRATEGY_RANDOM, if this strategy has been specified. Otherwise,
	 *            this parameter is ignored.
	 * @return The edges which have been added to the graph during graph construction process (non-simulation-mode) or which need to be added to the
	 *         graph (simulation mode). If graph construction was not successful, the return value is null.
	 * 
	 */
	public Map<Integer, TIntSet> constructGraph(final CompactDegreeSequence initialSeq, final CompactDegreeSequence solutionSeq,
			final boolean simulate, final int processingStrategy, final int selectionStrategy, final int randomSelectionPasses,
			final int randomProcessingPasses)
	{
		GreedyGraphConstructor.RANDOM_PROCESSING_PASSES = randomProcessingPasses;
		GreedyGraphConstructor.RANDOM_SELECTION_PASSES = randomSelectionPasses;
		return this.constructGraph(initialSeq, solutionSeq, simulate, processingStrategy, selectionStrategy);
	}

	/**
	 * Graph construction for non-compact degree sequences (deprecated).
	 * 
	 * @param initialDegreeSequence Initial degree sequence corresponding to the input graph.
	 * @param solutionSequence k-anonymous solution degree sequence.
	 * @param simulate Simulation mode reverts all changes made to the graph structure after construction.
	 * @return The edges which have been added to the graph during graph construction process (non-simulation-mode) or which need to be added to the
	 *         graph (simulation mode). If graph construction was not successful, the return value is null.
	 */
	public Map<Integer, TIntSet> constructGraph(final Vertex initialDegreeSequence[], final int solutionSequence[], final boolean simulate)
	{

		final ModifyList modifyList = new ModifyList(this.graph);

		int maxDifference = 0;
		int diff;
		for (int i = 1; i < initialDegreeSequence.length; i++)
		{
			final Vertex n = initialDegreeSequence[i];
			if ((n != null) && ((diff = solutionSequence[i] - n.getDegree()) != 0))
			{
				modifyList.add(new ModifyListEntry(n.getPosition(), n.getDegree(), diff, false));
				maxDifference = Math.max(diff, maxDifference);
			}
		}

		if (maxDifference > (modifyList.size() + 1))
		{
			if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 0)
			{
				System.out.println("Can't construct graph! Can't adjust " + modifyList.size() + " nodes when on node needs to be increased by "
						+ maxDifference);
			}
			return null;
		}

		return new ModifyListTransferer(this.graph).transferModifylistToGraph(simulate, modifyList, -1);

	}

}
