package algorithm.alternativeRealizer;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.list.TIntList;
import gnu.trove.list.array.TIntArrayList;
import gnu.trove.map.TIntIntMap;
import gnu.trove.map.hash.TIntIntHashMap;
import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;

import algorithm.dataStructures.CompactDegreeSequence;
import algorithm.dataStructures.GraphInterface;
import benchmarks.MainCLIAnonymizerAndAlternativeRealizer;

public class ModifyList implements Cloneable
{

	GraphInterface graph;
	private List<ModifyListEntry> modifyList;
	private TIntSet modifySet;

	public ModifyList(final GraphInterface graph)
	{
		this.graph = graph;
		this.modifyList = null;
	}

	public boolean add(final ModifyListEntry entry)
	{
		this.modifyList.add(entry);
		this.modifySet = null;
		return true;
	}

	private boolean alreadyContains(final int node)
	{
		return this.modifySet.contains(node);

	}

	private int amountOfAdjacentVerticesInList(final int node)
	{
		int sum = 0;
		for (final int v : this.graph.getNeighbors(node))
		{
			if (this.alreadyContains(v))
			{
				sum++;
			}
		}
		return sum;
	}

	@Override
	public ModifyList clone()
	{
		final ModifyList clone = new ModifyList(this.graph);
		clone.modifyList = new ArrayList<>();
		clone.modifySet = new TIntHashSet();
		for (final ModifyListEntry m : this.modifyList)
		{
			clone.modifyList.add(m.clone());
		}
		final TIntIterator it = this.modifySet.iterator();
		while (it.hasNext())
		{
			clone.modifySet.add(it.next());
		}

		return clone;
	}

	public void computeAndGetModifyList(final CompactDegreeSequence initialSeq, final CompactDegreeSequence solutionSeq, final int selectionStrategy)
	{
		final List<RaiseSequenceEntry> degreeRaiseList = new BlockRaiseListComputer().generateBlockRaiseListBottomUp(initialSeq, solutionSeq);

		this.modifyList = new ArrayList<>();
		this.modifySet = new TIntHashSet();
		// solution degree mapped to a set of vertex indexes which shall be
		// raise to that solution degree
		final Map<Integer, TIntSet> selectedVertices = new HashMap<>();
		for (int i = 0; i < degreeRaiseList.size(); i++)
		{ // for each vertex degree, where we have vertices which must be
			// modified...

			final RaiseSequenceEntry entry = degreeRaiseList.get(i);
			final int initDegree = entry.getFirst();
			final int solDegree = entry.getSecond();
			int amountOfNodes = entry.getThird();

			final int[] nodesWithThisDegree = this.graph.getVerticesWithDegree(initDegree);

			if ((nodesWithThisDegree != null) && (this.graph.getNumberOfVerticesWithDegree(initDegree) == amountOfNodes))
			{ // we must raise all the nodes in our graph with degree initDegree
				for (final int node : nodesWithThisDegree)
				{
					this.modifyList.add(new ModifyListEntry(node, initDegree, solDegree - initDegree, true));
					this.modifySet.add(node);
					if (selectedVertices.get(solDegree) == null)
					{
						selectedVertices.put(solDegree, new TIntHashSet());
					}
					selectedVertices.get(solDegree).add(node);
					--amountOfNodes;
				}
			}

			else if ((nodesWithThisDegree == null) || (amountOfNodes > this.graph.getNumberOfVerticesWithDegree(initDegree)))
			{ // there are less nodes with initDegree than we must raise ->
				// re-raise some nodes
				// which we have risen before
				final TIntSet alreadyRisenNodes = selectedVertices.get(initDegree);
				final TIntIterator it = alreadyRisenNodes.iterator();

				while ((amountOfNodes > 0)
						&& ((nodesWithThisDegree == null) || (this.graph.getNumberOfVerticesWithDegree(initDegree) < amountOfNodes)))
				{

					final int node = it.next();
					it.remove();
					if (selectedVertices.get(solDegree) == null)
					{
						selectedVertices.put(solDegree, new TIntHashSet());
					}
					selectedVertices.get(solDegree).add(node);

					final Iterator<ModifyListEntry> it2 = this.modifyList.iterator();
					while (it2.hasNext())
					{
						final ModifyListEntry modd = it2.next();
						if (modd.getNode() == node)
						{
							it2.remove();
							this.modifyList.add(new ModifyListEntry(node, initDegree, (solDegree - initDegree) + modd.getDegreeDifference(), true));
							break;
						}
					}

					--amountOfNodes;
				}
			}

			if (amountOfNodes > 0)
			{ // When there are more nodes available than we need --> here come
				// the strategies

				final List<RaiseCandidate> raiseCandidates = new ArrayList<>(this.graph.getNumberOfVerticesWithDegree(initDegree));

				final int neededVertices = amountOfNodes;
				if ((selectionStrategy == GreedyGraphConstructor.SELECTION_STRATEGY_LEAST_COMMON_NEIGHBORS)
						|| (selectionStrategy == GreedyGraphConstructor.SELECTION_STRATEGY_LEAST_COMMON_NEIGHBORS_REFINED)) // sorted modifyList

				{
					// Collect information about shared neighbors to nodes that
					// have already been raised.. choose the best one then..
					int amountOfPerfectCandidates = 0;
					for (final int actualVertex : nodesWithThisDegree)
					{

						int sumOfCommonNeighborsInSelection = 0;

						for (final int v : this.graph.getNeighbors(actualVertex))
						{
							if (this.alreadyContains(v))
							{
								sumOfCommonNeighborsInSelection++;
							}
						}

						raiseCandidates.add(new RaiseCandidate(actualVertex, sumOfCommonNeighborsInSelection));
						if (sumOfCommonNeighborsInSelection == 0)
						{
							amountOfPerfectCandidates++;
						}
						if (amountOfPerfectCandidates == neededVertices)
						{
							break;
						}
					}
					Collections.sort(raiseCandidates, new Comparator<RaiseCandidate>()
					{
						@Override
						public int compare(final RaiseCandidate o1, final RaiseCandidate o2)
						{
							return o1.getWeight() - o2.getWeight();
						}
					});

				}

				else if (selectionStrategy == GreedyGraphConstructor.SELECTION_STRATEGY_LEAST_NEIGHBOR_DEGREE_SUM)
				{

					// Collect information about shared neighbors to nodes that
					// have already been raised.. choose the best one then..
					int perfectVertices = 0;
					OUTERBREAK: for (final int node : nodesWithThisDegree)
					{

						int sumOfNeighborsDegree = 0;
						for (final int v : this.graph.getNeighbors(node))
						{
							sumOfNeighborsDegree += this.graph.getDegree(v);
						}
						raiseCandidates.add(new RaiseCandidate(node, sumOfNeighborsDegree));
						if (sumOfNeighborsDegree == 1)
						{
							perfectVertices++;
						}
						if (perfectVertices == amountOfNodes)
						{
							break OUTERBREAK;
						}
					}

					Collections.sort(raiseCandidates, new Comparator<RaiseCandidate>()
					{
						@Override
						public int compare(final RaiseCandidate o1, final RaiseCandidate o2)
						{
							return o1.getWeight() - o2.getWeight();
						}
					});

				}

				if (selectionStrategy != GreedyGraphConstructor.SELECTION_STRATEGY_RANDOM)
				{
					for (int x = 0; x < amountOfNodes; x++)
					{
						final RaiseCandidate raiseThisOne = raiseCandidates.get(x);
						this.modifyList.add(new ModifyListEntry(raiseThisOne.getNode(), initDegree, solDegree - initDegree, false));
						this.modifySet.add(raiseThisOne.getNode());
						if (selectedVertices.get(solDegree) == null)
						{
							selectedVertices.put(solDegree, new TIntHashSet());
						}
						selectedVertices.get(solDegree).add(raiseThisOne.getNode());
					}
				}

				else
				{

					final TIntSet selectionSet = new TIntHashSet();
					final TIntList nodesWithThisDegreeList = new TIntArrayList(nodesWithThisDegree);

					while (amountOfNodes != selectionSet.size())
					{
						final int pos = new Random().nextInt(this.graph.getNumberOfVerticesWithDegree(initDegree));
						selectionSet.add(pos);
					}
					final TIntIterator it = selectionSet.iterator();

					while (it.hasNext())
					{
						final int node = nodesWithThisDegreeList.get(it.next());
						this.modifyList.add(new ModifyListEntry(node, initDegree, solDegree - initDegree, false));
						this.modifySet.add(node);
						if (selectedVertices.get(solDegree) == null)
						{
							selectedVertices.put(solDegree, new TIntHashSet());
						}
						selectedVertices.get(solDegree).add(node);
					}
				}
			}

		}

		if ((selectionStrategy == GreedyGraphConstructor.SELECTION_STRATEGY_LEAST_COMMON_NEIGHBORS)
				|| (selectionStrategy == GreedyGraphConstructor.SELECTION_STRATEGY_LEAST_COMMON_NEIGHBORS_REFINED))
		{
			final boolean refined = (selectionStrategy == GreedyGraphConstructor.SELECTION_STRATEGY_LEAST_COMMON_NEIGHBORS_REFINED);
			this.improveSolutionByCommonNeighborsSort(refined);
		}
	}

	public ModifyListEntry get(final int i)
	{
		if ((this.modifyList != null) && (this.modifyList.size() > i))
			return this.modifyList.get(i);
		else
			return null;
	}

	private void improveSolutionByCommonNeighborsSort(final boolean refined)
	{
		boolean doItAgain = true;
		while (doItAgain)
		{
			doItAgain = false;
			for (int i = 0; i < this.modifyList.size(); i++)
			{
				if (!this.modifyList.get(i).isFixed())
				{
					final int node = this.modifyList.get(i).getNode();
					int actualSum = this.amountOfAdjacentVerticesInList(node);
					if (actualSum > 0)
					{
						final int[] nodesWithThisDegree = this.graph.getVerticesWithDegree(this.graph.getDegree(node));

						for (int l = 0; (l < nodesWithThisDegree.length) && (actualSum > 0); l++)
						{
							final int newNode = nodesWithThisDegree[l];
							if ((newNode != node) && !this.alreadyContains(newNode))
							{
								final int newSum = this.amountOfAdjacentVerticesInList(newNode);
								if (newSum < actualSum)
								{
									if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 2)
									{
										System.out.println("Swapped node #" + node + ",amount of common neighbors: " + actualSum + " with node #"
												+ newNode + ",amount of common neighbors: " + newSum);
									}

									actualSum = newSum;
									final int diff = this.modifyList.get(i).getDegreeDifference();
									this.modifyList.remove(i);
									this.modifyList.add(i, new ModifyListEntry(newNode, this.graph.getDegree(node), diff, false));
									this.modifySet.remove(node);
									this.modifySet.add(newNode);
									if (refined == true)
									{
										doItAgain = true;
									}
								}
							}

						}
					}

				}

			}
		}

	}

	protected Iterator<ModifyListEntry> iterator()
	{
		return this.modifyList.iterator();
	}

	protected void randomizeModifyListPositions()
	{

		final Map<ModifyListEntry, Integer> randomHelper = new HashMap<>();
		final Random r = new Random();

		for (int i = 0; i < this.modifyList.size(); i++)
		{
			randomHelper.put(this.modifyList.get(i), r.nextInt(Integer.MAX_VALUE));
		}

		Collections.sort(this.modifyList, new Comparator<ModifyListEntry>()
		{

			@Override
			public int compare(final ModifyListEntry o1, final ModifyListEntry o2)
			{
				return randomHelper.get(o1) - randomHelper.get(o2);
			}

		});

	}

	protected int size()
	{
		if (this.modifyList != null)
			return this.modifyList.size();
		else
			return 0;
	}

	protected void sortModifyListByCommonNeighbors()
	{

		final TIntIntMap map = new TIntIntHashMap();

		for (int i = 0; i < this.modifyList.size(); i++)
		{
			for (int j = i + 1; j < this.modifyList.size(); j++)
			{
				if (this.graph.isAdjacent(this.modifyList.get(i).getNode(), this.modifyList.get(j).getNode()))
				{
					map.adjustOrPutValue(this.modifyList.get(i).getNode(), 1, 1);
				}
			}
		}

		Collections.sort(this.modifyList, new Comparator<ModifyListEntry>()
		{

			@Override
			public int compare(final ModifyListEntry o1, final ModifyListEntry o2)
			{
				if (map.get(o2.getNode()) == map.get(o1.getNode()))
					return o2.getDegreeDifference() - o1.getDegreeDifference();
				return map.get(o2.getNode()) - map.get(o1.getNode());
			}

		});

	}

	protected void sortModifyListByDegreeDifference()
	{

		Collections.sort(this.modifyList, new Comparator<ModifyListEntry>()
		{

			@Override
			public int compare(final ModifyListEntry o1, final ModifyListEntry o2)
			{
				if (o2.getDegreeDifference() != o1.getDegreeDifference())
					return o2.getDegreeDifference() - o1.getDegreeDifference();
				else
					return ModifyList.this.graph.getDegree(o2.getNode()) - ModifyList.this.graph.getDegree(o1.getNode());
			}

		});

	}

	/**
	 * Calculates the total amount of degree increases (for debugging)
	 * 
	 * @param modifyList
	 * @return
	 */
	private int sumOfAllRaises(final List<ModifyListEntry> modifyList)
	{
		int sum = 0;
		for (final ModifyListEntry mod : modifyList)
		{
			sum += mod.getDegreeDifference();
		}
		return sum;
	}

}
