package algorithm.alternativeRealizer;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import algorithm.dataStructures.GraphInterface;
import benchmarks.MainCLIAnonymizerAndAlternativeRealizer;

public class ModifyListTransferer
{

	private final GraphInterface graph;

	public ModifyListTransferer(final GraphInterface graph)
	{
		this.graph = graph;
	}

	/**
	 * Checks whether all nodes have the solution degree and the gaph construction is done.
	 * 
	 * @param modifyList List of nodes and the amount of degree raises that remains.
	 * @return
	 */
	private boolean finished(final ModifyList modifyList)
	{
		final Iterator<ModifyListEntry> it = modifyList.iterator();
		while (it.hasNext())
		{
			final ModifyListEntry m = it.next();
			if (m.getDegreeDifference() > 0)
				return false;
		}
		return true;
	}

	/**
	 * For each pair of vertices that could not be connected by an edge, this method tries to find a different solution by a local search algorithm.
	 * It removes added edges and tries to connect the endpoints of this edge to other vertices.
	 * 
	 * @param backupList List of already added edges. Each entry consists of the the node indices of both nodes.
	 * @param modifyList List of vertices whose degree has to be raised.
	 * @param changedEdges Stores all added egdes, so that the indices match to the matrix for quick access.
	 * @param matrix Data structure which stores IntTupel for each combination of nodes which are affected.
	 */
	private void reconstructGraph(final ModifyList modifyList, final Map<Integer, TIntSet> changedEdges)
	{
		boolean reDo = false;
		if (!this.finished(modifyList))
		{
			for (int i = 0; i < modifyList.size(); i++)
			{
				final ModifyListEntry m = modifyList.get(i);
				if (m.getDegreeDifference() > 0)
				{
					for (int j = i + 1; j < modifyList.size(); j++)
					{
						final ModifyListEntry n = modifyList.get(j);
						if ((m.getDegreeDifference() > 0) && (n.getDegreeDifference() > 0))
						{
							final int node1 = Math.min(n.getNode(), m.getNode());
							final int node2 = Math.max(m.getNode(), n.getNode());
							for (int x = 0; x < modifyList.size(); x++)
							{
								if ((x != i) && (x != j))
								{
									final ModifyListEntry s = modifyList.get(x);

									for (int y = x + 1; y < modifyList.size(); y++)
									{
										if ((y != i) && (y != j) && (y != x) && (m.getDegreeDifference() > 0) && (n.getDegreeDifference() > 0))
										{
											final ModifyListEntry t = modifyList.get(y);
											final int node3 = Math.min(s.getNode(), t.getNode());
											final int node4 = Math.max(s.getNode(), t.getNode());
											if (!(changedEdges.get(node3) == null) && changedEdges.get(node3).contains(node4))
											{
												if (!this.graph.isAdjacent(node1, node3) && !this.graph.isAdjacent(node2, node4))
												{
													changedEdges.get(node3).remove(node4);
													if (changedEdges.get(Math.min(node1, node3)) == null)
													{
														changedEdges.put(Math.min(node1, node3), new TIntHashSet());
													}
													changedEdges.get(Math.min(node1, node3)).add(Math.max(node1, node3));
													if (changedEdges.get(Math.min(node2, node4)) == null)
													{
														changedEdges.put(Math.min(node2, node4), new TIntHashSet());
													}
													changedEdges.get(Math.min(node2, node4)).add(Math.max(node2, node4));
													this.graph.removeEdge(node3, node4);
													this.graph.addEdge(node1, node3);
													this.graph.addEdge(node2, node4);
													m.reduceDegreeDifference(1);
													n.reduceDegreeDifference(1);
													reDo = true;

												} else if (!this.graph.isAdjacent(node2, node3) && !this.graph.isAdjacent(node1, node4))
												{

													changedEdges.get(node3).remove(node4);
													if (changedEdges.get(Math.min(node1, node4)) == null)
													{
														changedEdges.put(Math.min(node1, node4), new TIntHashSet());
													}
													changedEdges.get(Math.min(node1, node4)).add(Math.max(node1, node4));
													if (changedEdges.get(Math.min(node2, node3)) == null)
													{
														changedEdges.put(Math.min(node2, node3), new TIntHashSet());
													}
													changedEdges.get(Math.min(node2, node3)).add(Math.max(node2, node3));
													this.graph.removeEdge(node3, node4);
													this.graph.addEdge(node1, node4);
													this.graph.addEdge(node2, node3);
													m.reduceDegreeDifference(1);
													n.reduceDegreeDifference(1);
													reDo = true;
												}
											}

										}
									}
								}

							}
						}
					}
				}
			}
		}

		if (reDo)
		{
			this.reconstructGraph(modifyList, changedEdges);
			return;
		}

	}

	/**
	 * Performs the graph construction, i.e. adds edges and reconnects edges if needed.
	 * 
	 * @param simulate If true, all changes are undoed after graph constructing.
	 * @param modifyList List of all vertices whose degree must be raised. The first value of each IntTupel is the node, the second value is the
	 *            difference between original degree and solution degree.
	 * @return A List of all edges, that have been added.
	 */
	protected Map<Integer, TIntSet> transferModifylistToGraph(final boolean simulate, ModifyList modifyList, final int processingStrategy)
	{

		// So that we don't touch the original modify list, we are going to need
		// it to be unmodified in random mode for the next run
		modifyList = modifyList.clone();

		final Map<Integer, TIntSet> changedEdges = new HashMap<>();

		boolean doItAgain = true;
		while (doItAgain)
		{
			doItAgain = false;
			OUTERBREAK: for (int i = 0; i < modifyList.size(); i++)
			{
				for (int j = i + 1; j < modifyList.size(); j++)
				{
					final ModifyListEntry n = modifyList.get(i);
					final ModifyListEntry m = modifyList.get(j);
					if ((n.getDegreeDifference() > 0) && (n != m) && (m.getDegreeDifference() > 0)
							&& !this.graph.isAdjacent(n.getNode(), m.getNode()))
					{
						this.graph.addEdge(n.getNode(), m.getNode());
						n.reduceDegreeDifference(1);
						m.reduceDegreeDifference(1);
						final int node1 = Math.min(n.getNode(), m.getNode());
						final int node2 = Math.max(m.getNode(), n.getNode());
						if (changedEdges.get(node1) == null)
						{
							changedEdges.put(node1, new TIntHashSet());
						}
						changedEdges.get(node1).add(node2);

						if (processingStrategy == GreedyGraphConstructor.PROCESSING_STRATEGY_MAXDIFF_REFINED)
						{
							modifyList.sortModifyListByDegreeDifference();
							doItAgain = true;
							break OUTERBREAK;
						}

					}
				}
			}
		}

		this.reconstructGraph(modifyList, changedEdges);
		boolean alreadyUndone = false;
		boolean failed = false;
		for (int i = 0; i < modifyList.size(); i++)
		{
			final ModifyListEntry n = modifyList.get(i);
			if (n.getDegreeDifference() != 0)
			{
				if (!alreadyUndone)
				{
					this.undoAddEdges(changedEdges);
					alreadyUndone = true;
				}
				failed = true;
				if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 3)
				{
					System.out.println("Edge insertion failure: Node " + n.getNode() + " requires to be connected to " + n.getDegreeDifference()
							+ " more edges, which is not possible. Original degree: " + this.graph.getDegree(n.getNode()) + ", Solution degree: "
							+ (this.graph.getDegree(n.getNode()) + modifyList.get(i).getDegreeDifference()));
				}
			}
		}

		if (!failed)
		{
			int cost = 0;
			final Set<Entry<Integer, TIntSet>> entrySet = changedEdges.entrySet();
			final Iterator<Entry<Integer, TIntSet>> it = entrySet.iterator();
			while (it.hasNext())
			{
				cost += it.next().getValue().size();
			}
			// if (cost != (MainCLI.cost / 2))
			// {
			// System.out.println("Error, wrong cost of " + cost + " instead of " + MainCLI.cost);
			// }
		}

		if (failed)
			return null;
		else
		{
			if (MainCLIAnonymizerAndAlternativeRealizer.logLevel > 0)
			{
				System.out.println("Graph Construction " + (simulate ? "(Simulation)" : "") + " was succesful!");
			}
			final int anonymity = this.graph.getAnonymity();
			// if (anonymity != MainCLI.k)
			// {
			//
			// System.out.println("Error: Graph is only " + anonymity + "-anonymous!");
			// }

			if (simulate)
			{
				this.undoAddEdges(changedEdges);
			}

			return changedEdges;
		}

	}

	/**
	 * Based on the input edge list, all changes are undone, so that each edge that has been added while graph constructing is removed again.
	 * 
	 * @param backupList List of edges that has to be removed
	 * @return True, if all edges could be removed.
	 */
	private boolean undoAddEdges(final Map<Integer, TIntSet> backupList)
	{
		boolean successful = true;
		for (final Entry<Integer, TIntSet> t : backupList.entrySet())
		{
			final TIntIterator valIt = t.getValue().iterator();
			while (valIt.hasNext())
			{
				successful &= this.graph.removeEdge(t.getKey(), valIt.next());
			}

		}
		if (!successful)
		{
			System.out.println("Error restoring original graph!");
		}
		return successful;
	}

}
