package algorithm.dataStructures;

import gnu.trove.iterator.TIntIntIterator;
import gnu.trove.map.TIntIntMap;
import gnu.trove.map.hash.TIntIntHashMap;
import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

import java.io.Serializable;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

public class AdjacentGraph implements Serializable, GraphInterface

{

	private int amountOfEdges = 0;

	private TIntSet[] degreeSet;

	private final List<TIntSet> graphSet;

	private final Map<Integer, String> indexToName;

	private final int noOfVerticesSize;

	/**
	 * Creates a new Adjacent Graph containing vertices and edges of input graphSet.
	 * 
	 * @param graphSet List of Integer sets, each set with index i contains all adjacent vertices to vertex with index i.
	 */
	public AdjacentGraph(final List<TIntSet> graphSet)
	{
		this(graphSet, null);
	}

	/**
	 * Creates a new Adjacent Graph containing vertices and edges of input graphSet.
	 * 
	 * @param graphSet List of Integer sets, each set with index i contains all adjacent vertices to vertex i.
	 * @param indexToName Maps the index of each vertex index in graphSet to its vertex String name.
	 */
	public AdjacentGraph(final List<TIntSet> graphSet, final Map<Integer, String> indexToName)
	{
		this.noOfVerticesSize = graphSet.size();
		this.graphSet = graphSet;
		this.calculateAmountOfEdges();
		this.fillDegreeSet();
		this.indexToName = indexToName;
	}

	@Override
	public boolean addEdge(final int node1, final int node2)
	{
		if (this.isVertexIndexInRange(node1) && this.isVertexIndexInRange(node2) && (node1 != node2))
		{
			boolean successful = true;
			final int oldDeg1 = this.getDegree(node1);
			final int oldDeg2 = this.getDegree(node2);
			successful &= this.graphSet.get(node1).add(node2);
			successful &= this.graphSet.get(node2).add(node1);
			final int newDeg1 = this.getDegree(node1);
			final int newDeg2 = this.getDegree(node2);
			if (successful)
			{
				this.amountOfEdges++;
				this.degreeSet[oldDeg1].remove(node1);
				this.degreeSet[oldDeg2].remove(node2);
				if ((this.degreeSet.length <= newDeg1) || (this.degreeSet.length <= newDeg2))
				{
					final TIntSet newDegreeset[] = Arrays.copyOf(this.degreeSet, this.degreeSet.length + 1);
					this.degreeSet = newDegreeset;
					this.degreeSet[this.degreeSet.length - 1] = new TIntHashSet();
				}
				this.degreeSet[newDeg1].add(node1);
				this.degreeSet[newDeg2].add(node2);
			}
			return successful;
		}

		return false;
	}

	private void calculateAmountOfEdges()
	{
		this.amountOfEdges = 0;
		if (this.graphSet != null)
		{
			for (final TIntSet adjacentNodes : this.graphSet)
			{
				if (adjacentNodes != null)
				{
					this.amountOfEdges += adjacentNodes.size();
				}
			}
			this.amountOfEdges /= 2;
		}
	}

	private void fillDegreeSet()
	{
		this.degreeSet = new TIntSet[this.getDelta() + 1];

		for (int i = 0; i < this.degreeSet.length; i++)
		{
			this.degreeSet[i] = new TIntHashSet();
		}

		for (int i = 0; i < this.graphSet.size(); i++)
		{
			final TIntSet node = this.graphSet.get(i);
			this.degreeSet[node.size()].add(i);
		}
	}

	private TIntSet getAdjacentNodes(final int vertexIndex)
	{
		if (this.isVertexIndexInRange(vertexIndex))
			return this.graphSet.get(vertexIndex);
		return null;
	}

	@Override
	public TIntSet getAllVerticesWithDegree(final int degree)
	{
		if ((degree >= 0) && (degree < this.degreeSet.length))
			return this.degreeSet[degree];
		else
			return null;
	}

	@Override
	public int getAnonymity()
	{
		final TIntIntMap anon = new TIntIntHashMap();
		for (final TIntSet adjacentNodes : this.graphSet)
		{
			if (adjacentNodes != null)
			{
				if (anon.containsKey(adjacentNodes.size()))
				{
					anon.put(adjacentNodes.size(), anon.get(adjacentNodes.size()) + 1);
				} else
				{
					anon.put(adjacentNodes.size(), 1);
				}
			}

		}
		int min = Integer.MAX_VALUE;
		final TIntIntIterator it = anon.iterator();
		while (it.hasNext())
		{
			it.advance();
			if (min > it.value())
			{
				min = it.value();
			}
		}
		return min;
	}

	@Override
	public int getDegree(final int vertexIndex)
	{
		if (this.isVertexIndexInRange(vertexIndex))
			return this.graphSet.get(vertexIndex).size();
		else
			return -1;
	}

	@Override
	public int getDelta()
	{
		int delta = 0;
		for (final TIntSet node : this.graphSet)
		{
			delta = Math.max(delta, node.size());
		}
		return delta;
	}

	@Override
	public String getName(final int vertexIndex)
	{
		if (this.isVertexIndexInRange(vertexIndex))
		{
			if (this.indexToName == null)
				return String.valueOf(vertexIndex);
			return this.indexToName.get(new Integer(vertexIndex));
		}
		return null;
	}

	@Override
	public int[] getNeighbors(final int vertexIndex)
	{
		if (this.isVertexIndexInRange(vertexIndex))
			return this.getAdjacentNodes(vertexIndex).toArray();
		else
			return null;
	}

	@Override
	public int getNumberOfEdges()
	{
		return this.amountOfEdges;
	}

	@Override
	public int getNumberOfVertices()
	{
		return this.noOfVerticesSize;
	}

	@Override
	public int getNumberOfVerticesWithDegree(final int degree)
	{
		if ((degree >= 0) && (degree < this.degreeSet.length))
			return this.degreeSet[degree].size();
		else
			return -1;
	}

	@Override
	public int getNumberOfVerticesWithDegreeAtLeast(final int degree)
	{
		if ((degree >= 0) && (degree < this.degreeSet.length))
		{
			int sum = 0;
			for (int i = degree; i < this.degreeSet.length; i++)
			{
				sum += this.degreeSet[i].size();
			}
			return sum;
		} else
			return -1;

	}

	@Override
	public int[] getVerticesWithDegree(final int degree)
	{
		if ((degree >= 0) && (degree < this.degreeSet.length))
			return this.getAllVerticesWithDegree(degree).toArray();
		else
			return null;

	}

	@Override
	public boolean isAdjacent(final int vertexIndex1, final int vertexIndex2)
	{
		if (this.isVertexIndexInRange(vertexIndex1) && this.isVertexIndexInRange(vertexIndex2))
		{
			final boolean t = this.graphSet.get(vertexIndex1).contains(vertexIndex2);
			if (!t)
				return false;
			return true;
		}
		return false;
	}

	private boolean isVertexIndexInRange(final int vertexIndex)
	{
		return (vertexIndex >= 0) && (vertexIndex <= this.getNumberOfVertices());
	}

	@Override
	public boolean removeEdge(final int node1, final int node2)
	{
		if (this.isVertexIndexInRange(node1) && this.isVertexIndexInRange(node2) && (node1 != node2))
		{
			boolean successful = true;
			final int oldDeg1 = this.getDegree(node1);
			final int oldDeg2 = this.getDegree(node2);
			successful &= this.graphSet.get(node1).remove(node2);
			successful &= this.graphSet.get(node2).remove(node1);
			final int newDeg1 = this.getDegree(node1);
			final int newDeg2 = this.getDegree(node2);
			if (successful)
			{
				this.amountOfEdges--;
				this.degreeSet[oldDeg1].remove(node1);
				this.degreeSet[oldDeg2].remove(node2);
				this.degreeSet[newDeg1].add(node1);
				this.degreeSet[newDeg2].add(node2);
			}
			return successful;
		}
		return false;
	}

	@Override
	public String toString()
	{
		return "AdjacentGraph [size=" + this.noOfVerticesSize + ",amountOfEdges=" + this.getNumberOfEdges() + ", degreeSet="
				+ Arrays.toString(this.degreeSet) + ", edgeSet=" + this.graphSet + "]";
	}

}
