package algorithm.dataStructures;

import java.io.Serializable;
import java.util.Arrays;

import algorithm.alternativeRealizer.DegreesChangeList;

public class CompactDegreeSequence implements Serializable
{
	/**
	 * 
	 */
	private static final long serialVersionUID = 8241107201328591868l;

	private final int[] degrees;

	private int[] degreeSum;

	private int[] numberOfEntries;

	public CompactDegreeSequence(final CompactDegreeSequence sequence, final int i)
	{
		this.degrees = Arrays.copyOf(sequence.degrees, i);
	}

	public CompactDegreeSequence(final int[] degreeSequence, final boolean isNotCompact)
	{
		if (isNotCompact)
		{
			// new DegreeSequence();
			this.degrees = new int[DegreeSequence.getDelta(degreeSequence) + 1];
			Arrays.fill(this.degrees, 0);
			for (int i = 1; i < degreeSequence.length; i++)
			{
				this.degrees[degreeSequence[i]]++;
			}
		} else
		{
			this.degrees = Arrays.copyOf(degreeSequence, degreeSequence.length);
		}
	}

	/**
	 * Calculates the result of this degree sequence minus the given one. Entry i gives the number of vertices the have been increase by i from old to
	 * this.
	 * 
	 * @param oldSeq
	 * @return
	 */
	public CompactDegreeSequence difference(final CompactDegreeSequence oldSeq)
	{
		final int[] newSeq = Arrays.copyOf(this.degrees, this.degrees.length);

		final int[] result = new int[this.degrees.length];
		Arrays.fill(result, 0);

		for (int i = 0; i < oldSeq.degrees.length; i++)
		{
			int c = oldSeq.degrees[i];
			for (int j = i; j < this.degrees.length; j++)
			{
				if (c <= 0)
				{
					break;
				}
				final int r = Math.min(c, newSeq[j]);
				newSeq[j] -= r;
				c -= r;
				result[j - i] += r;
			}
		}
		return new CompactDegreeSequence(result, false);
	}

	public DegreesChangeList differenceWithDegrees(final CompactDegreeSequence oldSeq)
	{
		final int[] newSeq = Arrays.copyOf(this.degrees, this.degrees.length);

		final DegreesChangeList result = new DegreesChangeList();

		for (int i = 0; i < oldSeq.degrees.length; i++)
		{
			int c = oldSeq.degrees[i];
			for (int j = i; j < this.degrees.length; j++)
			{
				if (c <= 0)
				{
					break;
				}
				final int r = Math.min(c, newSeq[j]);
				newSeq[j] -= r;
				c -= r;
				if (((j - i) > 0) && (r > 0))
				{
					result.add(j - i, i, r);
				}
			}
		}
		return result;
	}

	@Override
	public boolean equals(final Object obj)
	{
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (this.getClass() != obj.getClass())
			return false;
		final CompactDegreeSequence other = (CompactDegreeSequence) obj;
		if (!Arrays.equals(this.degrees, other.degrees))
			return false;
		return true;
	}

	// public CompactDegreeSequence newDifference(CompactDegreeSequence oldSeq)
	// {
	// int[] newDegrees = new int[this.degrees.length];
	//
	// int c=0;
	// for(int i=0;i<newDegrees.length;i++)
	// {
	// newDegrees[i]=this.getNumberOfVerticesInBags(0, i);
	// }
	// }

	public int get(final int i)
	{
		return this.degrees[i];
	}

	public int get(final int i, final int valueIfNotDefined)
	{
		if (i >= this.degrees.length)
			return valueIfNotDefined;
		return this.degrees[i];
	}

	public int[] getDegreeCopy()
	{
		return Arrays.copyOf(this.degrees, this.degrees.length);
	}

	public int getDegreeSumsInBags(final int fromBagIndex, final int endBagIndex)
	{
		if ((fromBagIndex > endBagIndex) || (fromBagIndex < 0) || (endBagIndex >= this.degrees.length))
			throw new IllegalArgumentException();
		if (this.degreeSum == null)
		{
			this.degreeSum = new int[this.degrees.length];
			for (int i = this.degrees.length - 1; i >= 0; i--)
			{
				this.degreeSum[i] = (i * this.degrees[i]) + (i < (this.degrees.length - 1) ? this.degreeSum[i + 1] : 0);
			}
		}

		return this.degreeSum[fromBagIndex] - (endBagIndex < (this.degrees.length - 1) ? this.degreeSum[endBagIndex + 1] : 0);
	}

	public int getLevelOfAnonymity()
	{
		int t = Integer.MAX_VALUE;
		for (int i = 0; i < this.degrees.length; i++)
			if (this.degrees[i] > 0)
			{
				t = Math.min(t, this.degrees[i]);
			}
		return t;
	}

	public int getMaxDegree()
	{
		return this.degrees.length - 1;
	}

	public int getMaxNonZeroDegree()
	{
		for (int i = this.degrees.length - 1; i >= 0; i--)
		{
			if (this.degrees[i] > 0)
				return i;
		}
		return 0;
	}

	public int getNumberOfBlocksWithSizeAtLeast(int min)
	{
		return this.getNumberOfBlocksWithSize(min, this.getTotalDegreeSum());
	}

	public int getNumberOfBlocksWithSizeAtMost(int max)
	{
		return this.getNumberOfBlocksWithSize(0, max);
	}

	public int getNumberOfBlocksWithSize(int min, int max)
	{
		int number = 0;
		for (int i = 0; i < this.degrees.length; i++)
		{
			if (this.degrees[i] >= min && this.degrees[i] <= max)
				number++;
		}
		return number;
	}

	
	public int getMaxVerticesWithSameDegree()
	{
		int max = 0;
		for (int i = 0; i < this.degrees.length; i++)
		{
			max = Math.max(max, this.degrees[i]);
		}
		return max;
	}

	/**
	 * Returns the number of elements in the bags from start to end index (values are cached).
	 * 
	 * @param fromBagIndex
	 * @param endBagIndex
	 * @return
	 */
	public int getNumberOfVerticesInBags(final int fromBagIndex, final int endBagIndex)
	{
		if ((fromBagIndex > endBagIndex) || (fromBagIndex < 0) || (endBagIndex >= this.degrees.length))
			throw new IllegalArgumentException();
		if (this.numberOfEntries == null)
		{
			this.numberOfEntries = new int[this.degrees.length];
			for (int i = this.degrees.length - 1; i >= 0; i--)
			{
				this.numberOfEntries[i] = this.degrees[i] + (i < (this.degrees.length - 1) ? this.numberOfEntries[i + 1] : 0);
			}
		}
		return this.numberOfEntries[fromBagIndex] - (endBagIndex < (this.degrees.length - 1) ? this.numberOfEntries[endBagIndex + 1] : 0);
	}

	private long getRightSum(final int i, final int r1)
	{
		long rightSide1 = (r1 * (r1 - 1)) + ((this.getNumberOfVerticesInBags(i, this.getMaxDegree()) - r1) * Math.min(r1, i));

		if (r1 >= (i - 1))
		{
			rightSide1 += this.getDegreeSumsInBags(0, i - 1);
		} else
		{
			rightSide1 += this.getNumberOfVerticesInBags(r1, i - 1) * r1;
			rightSide1 += this.getDegreeSumsInBags(0, r1 - 1);
		}
		return rightSide1;
	}

	public int getTotalDegreeSum()
	{
		return this.getDegreeSumsInBags(0, this.degrees.length - 1);
	}

	public int getTotalNumberOfVertices()
	{
		return this.getNumberOfVerticesInBags(0, this.degrees.length - 1);
	}

	@Override
	public int hashCode()
	{
		final int prime = 31;
		int result = 1;
		result = (prime * result) + Arrays.hashCode(this.degrees);
		return result;
	}

	public boolean isRealizable()
	{
		if ((this.getTotalDegreeSum() % 2) == 1)
			return false;

		int entryCounter = 0;
		int degreeCounter = 0;

		for (int i = this.degrees.length - 1; i >= 1; i--)
		{
			if (this.degrees[i] > 0)
			{
				final int r1 = entryCounter + 1;
				final int r2 = entryCounter + this.degrees[i];

				final long rightSide1 = this.getRightSum(i, r1);
				final long rightSide2 = this.getRightSum(i, r2);

				final long leftside1 = degreeCounter + i;
				final long leftside2 = degreeCounter + (this.degrees[i] * i);
				if ((rightSide1 < leftside1) || (rightSide2 < leftside2))
					return false;

				if (r1 >= i)
					return true;

				entryCounter += this.degrees[i];
				degreeCounter += i * this.degrees[i];
			}
		}

		return true;
	}

	public int noOfAddDegreeOneVerticesForBeingRealizableWRSimpleErdös()
	{
		int neededVertices = 0;

		int entryCounter = 0;
		int degreeCounter = 0;

		for (int i = this.degrees.length - 1; i >= 1; i--)
		{
			if (this.degrees[i] > 0)
			{
				final int r1 = entryCounter + 1;
				final int r2 = entryCounter + this.degrees[i];

				final long rightSide1 = this.getRightSum(i, r1);
				final long rightSide2 = this.getRightSum(i, r2);

				final long leftside1 = degreeCounter + i;
				final long leftside2 = degreeCounter + (this.degrees[i] * i);

				if (rightSide1 < leftside1)
				{
					neededVertices = (int) Math.max(leftside1 - rightSide1, neededVertices);
				}

				if (rightSide2 < leftside2)
				{
					neededVertices = (int) Math.max(leftside2 - rightSide2, neededVertices);
				}

				if (r1 >= i)
				{
					break;
				}

				entryCounter += this.degrees[i];
				degreeCounter += i * this.degrees[i];
			}
		}

		if ((((this.getTotalDegreeSum() % 2) + (neededVertices % 2)) % 2) == 0) // the same as (this.getTotalDegreeSum() + neededVertices) % 2 == 0
			return neededVertices;
		return neededVertices + 1;
	}

	/**
	 * checks whether the degree sequence fulfills the criteria of Erdös-Gallai. A degree sequence fulfills this criteria if and only if the degree
	 * sequence is realizable/graphical.
	 * 
	 * @return
	 */
	public boolean OldisRealizable()
	{
		long sumFrom_i_plus1 = 0;
		long r = 0;
		for (int i = this.degrees.length - 1; i >= 0; i--)
		{
			for (int j = 1; j <= this.degrees[i]; j++)
			{
				r++;
				final long leftside = sumFrom_i_plus1 + (j * i);

				if (((2 * r) >= i) && ((r * (r - 1)) >= leftside) && ((this.getTotalDegreeSum() % 2) == 0))
					return true;

				long rightside = (r * (r - 1)) + ((this.degrees[i] - j) * Math.min(r, i));
				for (int i2 = i - 1; i2 >= 0; i2--)
				{
					rightside += this.degrees[i2] * Math.min(i2, r);
				}
				if (leftside > rightside)
					return false;
			}
			sumFrom_i_plus1 += this.degrees[i] * i;
		}
		if ((sumFrom_i_plus1 % 2) == 0)
			return true;
		return false;
	}

	public int OldnoOfAddDegreeOneVerticesForBeingRealizable()
	{
		int neededVertices = 0;
		long sumFrom_i_plus1 = 0;
		long r = 0;
		for (int i = this.degrees.length - 1; i >= 0; i--)
		{
			for (int j = 1; j <= this.degrees[i]; j++)
			{
				r++;
				final long leftside = sumFrom_i_plus1 + (j * i);

				if (((2 * r) >= i) && ((r * (r - 1)) >= leftside))
					return (((this.getTotalDegreeSum() % 2) + (neededVertices % 2)) % 2) == 0 ? neededVertices : neededVertices + 1;

				long rightside = (r * (r - 1)) + ((this.degrees[i] - j) * Math.min(r, i));
				if (leftside > rightside)
				{
					for (int i2 = i - 1; i2 >= 0; i2--)
					{
						rightside += this.degrees[i2] * Math.min(i2, r);
					}
				}
				if (leftside > rightside)
				{
					neededVertices = Math.max(neededVertices, (int) (leftside - rightside));
				}
			}
			sumFrom_i_plus1 += this.degrees[i] * i;
		}
		if ((((sumFrom_i_plus1 % 2) + (neededVertices % 2)) % 2) == 0)
			return neededVertices;
		else
			return neededVertices + 1;
	}

	@Override
	public String toString()
	{
		String s = "[";
		for (int i = 0; i < this.degrees.length; i++)
		{
			if (this.degrees[i] > 0)
			{
				s += "(" + i + ", #:" + this.degrees[i] + ")";
			}
		}
		s += "]";
		return s;
		// return s2.length() < s.length() ? s2 : s;
	}

	public String toStringFull()
	{
		return Arrays.toString(this.degrees);

	}
	
	public String printDegreeDistribution()
	{
		String s = "";
		for (int i = 0; i < this.degrees.length; i++)
		{
			if (this.degrees[i] > 0)
				s += "" + i + "\t" + this.degrees[i] + "\r\n";
		}
		s += "";
		return s;
		// return s2.length() < s.length() ? s2 : s;
	}

}
