package algorithm.dataStructures;

import gnu.trove.set.TIntSet;

public interface GraphInterface
{

	/**
	 * Adds an undirected edge to the graph.
	 * 
	 * @param vertex1 First vertex.
	 * @param vertex2 Second vertex.
	 * @return True, if the edge has been added succefully and there was no such edge before. False otherwise.
	 */
	public boolean addEdge(int vertex1, int vertex2);

	/**
	 * Calculates and return a list with all vertex indices that have degree n.
	 * 
	 * @param n vertex degree
	 * @return Vertex index list
	 */
	public TIntSet getAllVerticesWithDegree(int n);

	/**
	 * Computes and return the degree-anonymity of the graph.
	 * 
	 * @return Degree-anonymity.
	 */
	public int getAnonymity();

	/**
	 * Returns the vertex vertex degree of a given vertex in the graph.
	 * 
	 * @param vertex The index of the vertex.
	 * @return Degree of the vertex with index n.
	 */
	public int getDegree(int vertex);

	/**
	 * Calculates the maximum vertex degree of this graph.
	 * 
	 * @return Maximum vertex degree.
	 */
	public int getDelta();

	/**
	 * Returns the name of the vertex with index n.
	 * 
	 * @param n Vertex index.
	 * @return Vertex name.
	 */
	public String getName(int n);

	/**
	 * Computes and returns an integer array containing all neighbor vertex indices of the input vertex.
	 * 
	 * @param vertex The vertex whose neighbors are returned.
	 * @return Array containing the indices of all neighbors of the input vertex.
	 */
	public int[] getNeighbors(int vertex);

	/**
	 * Return the total sum of edges in this graph
	 * 
	 * @return Amount of edges in this graph
	 */
	public int getNumberOfEdges();

	/**
	 * Returns the number of vertices in the graph.
	 * 
	 * @return Number of vertices in the graph.
	 */
	public int getNumberOfVertices();

	/**
	 * Return the number of vertices which have degree n.
	 * 
	 * @param n Vertex degree.
	 * @return Number of vertices with degree n.
	 */
	public int getNumberOfVerticesWithDegree(int n);

	/**
	 * Return the number of vertices which have at least degree n.
	 * 
	 * @param n Vertex degree.
	 * @return Number of vertices with degree >= n.
	 */
	public int getNumberOfVerticesWithDegreeAtLeast(int degree);

	/**
	 * Computes and returns an integer array containing the vertex indices of all vertices with degree n.
	 * 
	 * @param n Vertex degree.
	 * @return Integer array containing the vertex indices of all vertices with degree n.
	 */
	public int[] getVerticesWithDegree(int n);

	/**
	 * Computes whether two vertices are adjacent.
	 * 
	 * @param node1 Index of the first vertex.
	 * @param node2 Index of the second vertex.
	 * @return True if node1 and node2 are adjacent, false otherwise.
	 */
	public boolean isAdjacent(int node1, int node2);

	/**
	 * Removes an edge from the graph.
	 * 
	 * @param node1 Index of the first vertex.
	 * @param node2 Index of the second vertex.
	 * @return True, if edge deletion was successful, false otherwise.
	 */
	public boolean removeEdge(int node1, int node2);

}