package algorithm.graphRealizer;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Vector;

import org.apache.commons.lang3.mutable.MutableDouble;
import org.apache.commons.lang3.mutable.MutableInt;

import algorithm.CostWaster;
import algorithm.FastErdosGallai;
import algorithm.UpperBound;
import algorithm.dataStructures.CompactDegreeSequence;
import algorithm.dataStructures.GraphInterface;
import algorithm.dataStructures.Timer;
import dataStructures.IntTupel;
import dataStructures.Store;
import dataStructures.utils.ArrayS;
import dataStructures.utils.IntStore;

public final class GraphRealizer
{
	private final class DegreeIncrease implements Cloneable

	{
		private int degreeIncrease;

		private final int originalDegree;

		private int vertex = Integer.MAX_VALUE;

		public DegreeIncrease(final int originalDegree)
		{
			this.originalDegree = originalDegree;
			this.degreeIncrease = 0;
		}

		@Override
		protected DegreeIncrease clone()
		{
			final DegreeIncrease s = new DegreeIncrease(this.originalDegree);
			s.degreeIncrease = this.degreeIncrease;
			s.vertex = this.vertex;
			return s;
		}

		public int getDegreeIncrease()
		{
			return this.degreeIncrease;
		}

		public int getOriginalDegree()
		{
			return this.originalDegree;
		}

		public void setDegreeIncrease(final int increase)
		{
			this.degreeIncrease = increase;
		}

		public void setVertex(final int vertex)
		{
			this.vertex = vertex;
		}

		@Override
		public String toString()
		{
			return "[" + this.originalDegree + "->" + (this.degreeIncrease + this.originalDegree)
					+ (this.vertex != Integer.MAX_VALUE ? ", vert:" + this.vertex : "") + "]";
		}

	}

	private final class DegreeIncreaseCollection implements Cloneable
	{
		private final Vector<DegreeIncrease> finalDegreeIncreases;

		private final Vector<DegreeIncrease> opendedDegreeIncreases;

		public DegreeIncreaseCollection()
		{
			this.opendedDegreeIncreases = new Vector<DegreeIncrease>();
			this.finalDegreeIncreases = new Vector<DegreeIncrease>();
		}

		@Override
		protected DegreeIncreaseCollection clone()
		{
			final DegreeIncreaseCollection coll = new DegreeIncreaseCollection();
			for (final DegreeIncrease in : this.opendedDegreeIncreases)
			{
				coll.opendedDegreeIncreases.add(in.clone());
			}
			for (final DegreeIncrease in : this.finalDegreeIncreases)
			{
				coll.finalDegreeIncreases.add(in.clone());
			}
			return coll;
		}

		public void finalize(final int degree, final int numberVerticesToFinalize)
		{
			final Iterator<DegreeIncrease> iterator = this.opendedDegreeIncreases.iterator();
			int i = 1;
			while (iterator.hasNext() && (i++ <= numberVerticesToFinalize))
			{
				final DegreeIncrease in = iterator.next();
				iterator.remove();
				in.setDegreeIncrease(degree - in.getOriginalDegree());
				this.finalDegreeIncreases.add(in);
			}
		}

		public Vector<DegreeIncrease> getFinializedDegrees()
		{
			return this.finalDegreeIncreases;
		}

		public int getNoOpenedVertices()
		{
			return this.opendedDegreeIncreases.size();
		}

		public void openVertices(final int degree, final int noOfVerticesToOpen)
		{
			for (int i = 0; i < noOfVerticesToOpen; i++)
			{
				this.opendedDegreeIncreases.add(new DegreeIncrease(degree));
			}
		}

		@Override
		public String toString()
		{
			return "DegreeIncreaseCollection [vec=" + Arrays.toString(this.opendedDegreeIncreases.toArray()) + ", finalDegreeIncreases="
					+ Arrays.toString(this.finalDegreeIncreases.toArray()) + "]";
		}

	}

	public final static int advancedReorg_Penalize_Constant = 2;

	public static int log = 0;

	public static MutableInt MAX_INNERLOOP_ADD_MOST_PROMISING = new MutableInt(4);

	public static final int MINIMUM_TIME_FOR_IS_REALIZABLE = 10;

	public static final MutableInt PERCENTAGE_DISTURB_VERTEX_DEGREE_ORDER = new MutableInt(2);

	public static final MutableDouble PERCENTAGE_SHUFFLE_VERTEX_ORDER = new MutableDouble(0.1);

	public static final MutableDouble Safety_Buffer_Fac_MAX_Time_Configer = new MutableDouble(0.9d);

	public static final MutableInt SHUFFLE_DEGREE_VERTEX_MAPPING = new MutableInt(0);

	public static final MutableInt TIME_FACTOR_AS_LONG_AS_NO_SOLUTION_FOUND = new MutableInt(3);

	private static int getStepIncrease(final int x)
	{
		if (x < 0)
			throw new IllegalAccessError();

		// x += 2;

		final int i = x < UpperBound.StepIncrease_Magic_Const.intValue() ? x : (int) Math
				.round((UpperBound.StepIncrease_Magic_Const.intValue() - Math.pow(UpperBound.StepIncrease_Magic_Const.intValue(),
						UpperBound.StepIncrease_Magic_Const2.doubleValue())) + Math.pow(x, UpperBound.StepIncrease_Magic_Const2.doubleValue()));

		// System.err.println("stepInc:" + x);
		// int i = (int) Math.round(Math.sin(Math.pow(x, 0.8)) * Math.pow(x, 0.5) + Math.pow(x, 0.6) + 0.5);
		// final hint i = x <= 32 ? x : 47 + (int) Math.round(Math.pow(x, 0.5));
		return 2 << Math.min(UpperBound.StepIncrease_Magic_Const3.intValue(), i);
	}

	/**
	 * Finds the minimum amount of edges (heuristically) to be wasted such that {@code solutionSeq} becomes realizable. May take a while...but is
	 * (almost) guaranteed to return an edge set.
	 * 
	 * @param graph
	 * @param graphSequence
	 * @param solutionSeq
	 * @param k
	 * @return
	 * @throws CannotRealize
	 */
	public static Store<Integer, Iterable<IntTupel>> isRealizable(final GraphInterface graph, final CompactDegreeSequence graphSequence,
			final CompactDegreeSequence solutionSeq, final int k) throws CannotRealize
	{
		return GraphRealizer.isRealizable(graph, graphSequence, solutionSeq, Integer.MAX_VALUE, Integer.MAX_VALUE, k);
	}

	/**
	 * Checks whether the given {@code solutionSeq} can be realized in the given {@code graph} by wasting at most {@code maxWastCost} additional edges
	 * (in addition to those required by {@code solutionSeq}). If possible within the given time bound (specified by {@code maxTimeBound} in seconds)
	 * it returns a corresponding edge set (tries to minimize its size); otherwise throws a {@code CannotRealize} exception. Most likely, the method
	 * exceeds {@code maxTimeBound} by a couple of seconds.
	 * 
	 * @param graph
	 * @param graphSequence {@code CompactDegreeSequence} of the given graph (just to avoid recomputing it over and over again).
	 * @param solutionSeq
	 * @param maxWastCost
	 * @param maxTimeBound Max. time the method is supposed to take. Value must be at least 10 (below that value time prediction is too imprecise).
	 * @param k
	 * @return edge set and its size
	 * @throws CannotRealize
	 */
	public static Store<Integer, Iterable<IntTupel>> isRealizable(final GraphInterface graph, final CompactDegreeSequence graphSequence,
			final CompactDegreeSequence solutionSeq, int maxWastCost, int maxTimeBound, final int k) throws CannotRealize
	{
		if (maxTimeBound < GraphRealizer.MINIMUM_TIME_FOR_IS_REALIZABLE)
			throw new IllegalArgumentException("maxTimeBound is less than " + GraphRealizer.MINIMUM_TIME_FOR_IS_REALIZABLE + ":" + maxTimeBound);
		maxWastCost = Math.min(Integer.MAX_VALUE >> 1, maxWastCost);
		maxTimeBound = Math.min(Integer.MAX_VALUE >> 1, maxTimeBound);

		final Timer timer = new Timer(maxTimeBound);

		final int solutionDiffToOrig = solutionSeq.difference(graphSequence).getTotalDegreeSum();
		int minAddCosErdösSol = new FastErdosGallai(graph, graphSequence).computeFeasibility(solutionSeq, true);
		minAddCosErdösSol += (minAddCosErdösSol + solutionDiffToOrig) % 2;

		final SophisticatedGraphRealizerConfigGenerator generator = new SophisticatedGraphRealizerConfigGenerator(solutionDiffToOrig);

		int stepCounter = 0;

		for (int addCost = minAddCosErdösSol; addCost <= maxWastCost; addCost += GraphRealizer.getStepIncrease(++stepCounter))
		{
			if (UpperBound.printRealizerParams > 0)
			{
				System.out.println("addCost:" + UpperBound.largeIntFormater.format(addCost) + " (remaingTime: " + timer.getRemainingTime() + "s)");
			}

			if ((stepCounter % UpperBound.HIGH_QUALITY_IN_EACH_STEP.intValue()) == 0)
			{
				generator.enableHighQualityMode();
			} else
			{
				generator.disableHighQuality();
			}
			if ((timer.elapsedTimeInSeconds() + generator.setTimeForNextConfig(timer)) > maxTimeBound)
				throw new CannotRealize();

			Iterable<IntTupel> tmpTupel;
			final int maxTimeNoSolutionFound = Math.min(
					(int) Math.floor(generator.setTimeForNextConfig(timer) * GraphRealizer.TIME_FACTOR_AS_LONG_AS_NO_SOLUTION_FOUND.intValue()),
					timer.getRemainingTime());
			if ((tmpTupel = GraphRealizer.isRealizable(graph, graphSequence, solutionSeq, addCost, generator, maxTimeNoSolutionFound, k)) != null)
			{
				Store<Integer, Iterable<IntTupel>> bestSolutionSoFar = new Store<>(addCost + solutionDiffToOrig, tmpTupel);
				if (((timer.elapsedTimeInSeconds() + generator.setTimeForNextConfig(timer)) > maxTimeBound) || (addCost == minAddCosErdösSol))
					return bestSolutionSoFar;

				if (UpperBound.printRealizerParams > 2)
				{
					System.out.println("start binary search" + " (remaingTime: " + timer.getRemainingTime() + "s)");
				}

				// found solution, now doing binary search for the smallest value
				generator.enableHighQualityMode();// switch to high quality mode
				int ub = addCost;
				int counter = stepCounter;
				int lb = addCost - GraphRealizer.getStepIncrease(counter--);
				{// set ub and lb such that lb is the smallest unrealizable value
					while ((lb >= minAddCosErdösSol)
							&& ((tmpTupel = GraphRealizer.isRealizable(graph, graphSequence, solutionSeq, lb, generator, timer.getRemainingTime(), k)) != null))
					{
						bestSolutionSoFar = new Store<>(lb + solutionDiffToOrig, tmpTupel);
						if ((timer.elapsedTimeInSeconds() + generator.setTimeForNextConfig(timer)) > maxTimeBound)
							return bestSolutionSoFar;

						ub = lb;
						lb -= GraphRealizer.getStepIncrease(counter--);// reverse last steps
					}
				}

				if (UpperBound.printRealizerParams > 2)
				{
					System.out.println("end binary search" + " (remaingTime: " + timer.getRemainingTime() + "s)");
				}
				if ((timer.elapsedTimeInSeconds() + generator.setTimeForNextConfig(timer)) > maxTimeBound)
					return bestSolutionSoFar;

				while ((ub - lb) > 2)
				{
					int change = (ub - lb) >> 1;
					change += change % 2;
					final int bound = lb + change;
					if ((tmpTupel = GraphRealizer.isRealizable(graph, graphSequence, solutionSeq, bound, generator, timer.getRemainingTime(), k)) != null)
					{
						bestSolutionSoFar = new Store<>(lb + solutionDiffToOrig, tmpTupel);
						if ((timer.elapsedTimeInSeconds() + generator.setTimeForNextConfig(timer)) > maxTimeBound)
							return bestSolutionSoFar;

						ub = bound;
					} else
					{
						lb = bound;
					}
				}
				return bestSolutionSoFar;
			}

		}

		throw new CannotRealize();
	}

	private static Iterable<IntTupel> isRealizable(final GraphInterface graph, final CompactDegreeSequence origSeq, final CompactDegreeSequence sol,
			final int addCost, final SophisticatedGraphRealizerConfigGenerator generator, final int maxTime, final int k)
	{
		final Timer timer = new Timer();
		final CostWaster waster = new CostWaster(origSeq, sol, addCost, k);

		for (final CompactDegreeSequence realSol : waster.getAllRealizableSequences())
		{
			final GraphRealizerConfig nextConfig = generator.getNextConfig(addCost);
			try
			{

				final GraphRealizer realizer = new GraphRealizer(graph, origSeq, realSol, nextConfig);

				Iterable<IntTupel> tupel;
				if ((tupel = realizer.isRealizable()) != null)
				{
					if (UpperBound.printRealizerParams > 2)
					{
						System.out.println(nextConfig);
					}
					if (UpperBound.printRealizerParams > 0)
					{
						System.out.println("foundSol with waste cost:" + UpperBound.largeIntFormater.format(addCost));
					}
					return tupel;
				}
			} catch (final CannotRealize e)
			{
				if (UpperBound.printRealizerParams > 2)
				{
					System.out.println(nextConfig);
				}
				if ((timer.elapsedTimeInSeconds() + generator.setTimeForNextConfig(timer)) >= maxTime)
					return null;
			}
		}
		return null;
	}

	private final GraphRealizerConfig config;

	private Integer[] degreesWithPossibleJumps;

	private final GraphInterface graph;

	private final CompactDegreeSequence initialSeq;

	private int[] jumpDegreePosition;

	private Integer[] maxAmountOfJumps;

	private final int nonShuffleIter;

	private int[] numberOfJumps;

	private final CompactDegreeSequence solutionSeq;

	private GraphRealizer(final GraphInterface graph, final CompactDegreeSequence iniSeq, final CompactDegreeSequence solutionSeq,
			final GraphRealizerConfig config)
	{
		this.graph = graph;
		this.initialSeq = iniSeq;
		this.solutionSeq = solutionSeq;

		this.config = config;
		this.nonShuffleIter = Math.max(1, config.getMaxIter() / GraphRealizer.PERCENTAGE_DISTURB_VERTEX_DEGREE_ORDER.intValue());
	}

	private boolean advancedReorganizeSolution(final DegreeIncrease e1, final Vector<DegreeIncrease> degreeIncreases,
			final AbstractEdgeContainer newEdges) throws CannotRealize
	{
		this.config.increaseExecutedAdvancedReorganization();
		for (final DegreeIncrease deg2 : degreeIncreases)
		{

			if ((deg2 != e1) && (deg2.degreeIncrease > 0))
			{
				for (final IntTupel tup : newEdges.getAllEdges())
				{
					if (this.possibleToInsertEdge(newEdges, e1.vertex, tup.getFirst())
							&& this.possibleToInsertEdge(newEdges, deg2.vertex, tup.getSecond()))
					{
						newEdges.remove(tup);
						newEdges.addEdges(new IntTupel(e1.vertex, tup.getFirst()));
						newEdges.addEdges(new IntTupel(deg2.vertex, tup.getSecond()));
						e1.degreeIncrease--;
						deg2.degreeIncrease--;

						this.config.setExecutedAdvReorganizationTime();
						return true;
					} else if (this.possibleToInsertEdge(newEdges, e1.vertex, tup.getSecond())
							&& this.possibleToInsertEdge(newEdges, deg2.vertex, tup.getFirst()))
					{
						newEdges.remove(tup);
						newEdges.addEdges(new IntTupel(e1.vertex, tup.getSecond()));
						newEdges.addEdges(new IntTupel(deg2.vertex, tup.getFirst()));
						e1.degreeIncrease--;
						deg2.degreeIncrease--;

						this.config.setExecutedAdvReorganizationTime();
						return true;
					}
				}
			}
		}

		this.config.setExecutedAdvReorganizationTime();
		return false;
	}

	private DegreeIncreaseCollection getAdaptDegreeIncList(DegreeIncreaseCollection iniColl) throws CannotRealize
	{
		boolean realizable = false;
		while (!realizable)
		{
			try
			{
				iniColl = this.getAdaptiveDegreeIncreaseList2(this.solutionSeq);
			} catch (final IllegalAccessError e)
			{
				throw new CannotRealize(this, -1, -1);
			}
			realizable = this.isRealizable(iniColl);
		}
		return iniColl;
	}

	private DegreeIncreaseCollection getAdaptiveDegreeIncreaseList2(final CompactDegreeSequence solutionSeq)
	{
		if ((this.degreesWithPossibleJumps == null) || (this.degreesWithPossibleJumps.length == 0))
			throw new IllegalAccessError();

		if (this.jumpDegreePosition == null)
		{
			this.jumpDegreePosition = new int[1];
			this.numberOfJumps = new int[1];
			Arrays.fill(this.numberOfJumps, 1);
		} else if (this.increaseInNumberOfJumpsPossible())
		{
		} else
		{
			boolean success = false;
			for (int k = this.jumpDegreePosition.length - 1; k >= 0; k--)
			{
				if (((k == (this.jumpDegreePosition.length - 1)) && (this.jumpDegreePosition[k] < (this.degreesWithPossibleJumps.length - 1)))
						|| ((k < (this.jumpDegreePosition.length - 1)) && (this.jumpDegreePosition[k] < (this.jumpDegreePosition[k + 1] - 1))))
				{
					this.jumpDegreePosition[k]++;
					this.numberOfJumps[k] = 1;
					for (int k2 = k + 1; k2 < this.jumpDegreePosition.length; k2++)
					{
						this.jumpDegreePosition[k2] = this.jumpDegreePosition[k2 - 1] + 1;
						this.numberOfJumps[k2] = 1;
					}
					success = true;
					break;
				}
			}
			if (!success)
			{
				if (this.jumpDegreePosition.length < this.degreesWithPossibleJumps.length)
				{
					this.jumpDegreePosition = new int[this.jumpDegreePosition.length + 1];
					for (int l = 0; l < this.jumpDegreePosition.length; l++)
					{
						this.jumpDegreePosition[l] = l;
					}
					this.numberOfJumps = new int[this.jumpDegreePosition.length];
					Arrays.fill(this.numberOfJumps, 1);
				} else
					throw new IllegalAccessError();
			}
		}

		int pCount = 0;
		final DegreeIncreaseCollection coll = new DegreeIncreaseCollection();

		for (int deg = 0; deg <= solutionSeq.getMaxDegree(); deg++)
		{

			int verticesToFinalize = Math.min(solutionSeq.get(deg), coll.getNoOpenedVertices());
			int noOfVerticesToOpen = (deg <= this.initialSeq.getMaxDegree() ? this.initialSeq.get(deg) : 0)
					- (solutionSeq.get(deg) - verticesToFinalize);

			if ((pCount < this.jumpDegreePosition.length) && (deg == this.degreesWithPossibleJumps[this.jumpDegreePosition[pCount]]))
			{
				verticesToFinalize -= this.numberOfJumps[pCount];
				noOfVerticesToOpen -= this.numberOfJumps[pCount];
				pCount++;
			}
			coll.finalize(deg, verticesToFinalize);
			coll.openVertices(deg, noOfVerticesToOpen);

		}
		return coll;

	}

	private void getDegreeVertexMapping(final DegreeIncreaseCollection coll, final TreeMap<Integer, Vector<Integer>> degreeToPossibleVertices)
	{
		final HashSet<Integer> assignedVertices = new HashSet<>(coll.finalDegreeIncreases.size());

		final int[] indexCounter = new int[this.initialSeq.getMaxDegree() + 1];
		Arrays.fill(indexCounter, 0);
		for (final DegreeIncrease deg : coll.getFinializedDegrees())
		{
			Vector<Integer> s = degreeToPossibleVertices.get(deg.originalDegree);
			if (indexCounter[deg.originalDegree] == 0)
			{
				if (GraphRealizer.SHUFFLE_DEGREE_VERTEX_MAPPING.intValue() == 0)
				{
					Collections.shuffle(s);
				} else if ((GraphRealizer.SHUFFLE_DEGREE_VERTEX_MAPPING.intValue() == 1)
						|| (GraphRealizer.SHUFFLE_DEGREE_VERTEX_MAPPING.intValue() == 2))
				{
					final Vector<IntStore<Integer>> sVec = new Vector<>();
					for (final int v : s)
					{
						sVec.add(new IntStore<Integer>(this.getNoVertices(assignedVertices, v), v));
					}
					Collections.sort(sVec);

					final Vector<Integer> newVec = new Vector<>();
					for (final IntStore<Integer> vec : sVec)
					{
						newVec.add(vec.getS2());
					}
					degreeToPossibleVertices.put(deg.originalDegree, newVec);
					s = newVec;
				}
				if (GraphRealizer.SHUFFLE_DEGREE_VERTEX_MAPPING.intValue() == 2)
				{
					ArrayS.shuffleWithinSmallRange(s, GraphRealizer.PERCENTAGE_SHUFFLE_VERTEX_ORDER.floatValue());
				}
			}
			final Integer vertex = s.get(indexCounter[deg.originalDegree]++);
			assignedVertices.add(vertex);
			deg.setVertex(vertex);
		}
	}

	private DegreeIncreaseCollection getInitialDegreeIncreaseList(final CompactDegreeSequence initialSeq, final CompactDegreeSequence solutionSeq)
	{
		final DegreeIncreaseCollection coll = new DegreeIncreaseCollection();

		final ArrayList<Integer> indices = new ArrayList<Integer>();
		final ArrayList<Integer> amoungOfJumpsT = new ArrayList<Integer>();

		for (int deg = 0; deg <= solutionSeq.getMaxDegree(); deg++)
		{
			final int verticesToFinalize = Math.min(solutionSeq.get(deg), coll.getNoOpenedVertices());
			coll.finalize(deg, verticesToFinalize);
			final int noOfVerticesToOpen = (deg <= initialSeq.getMaxDegree() ? initialSeq.get(deg) : 0) - (solutionSeq.get(deg) - verticesToFinalize);
			coll.openVertices(deg, noOfVerticesToOpen);

			if ((noOfVerticesToOpen > 0) && (verticesToFinalize > 0))
			{
				indices.add(deg);
				amoungOfJumpsT.add(Math.min(noOfVerticesToOpen, verticesToFinalize));
			}
		}

		this.degreesWithPossibleJumps = indices.toArray(new Integer[0]);
		this.maxAmountOfJumps = amoungOfJumpsT.toArray(new Integer[0]);

		if (GraphRealizer.log > 1)
		{
			System.out.println("poss:" + this.degreesWithPossibleJumps.length + "_" + Arrays.toString(this.degreesWithPossibleJumps)
					+ "_amount of Jumps:" + Arrays.toString(this.maxAmountOfJumps));
		}
		return coll;
	}

	private int getNoVertices(final HashSet<Integer> assignedVertices, final int vertex)
	{
		int i = 0;
		if (assignedVertices.size() < this.graph.getDegree(vertex))
		{
			for (final int v : assignedVertices)
			{
				if (this.graph.isAdjacent(v, vertex))
				{
					i++;
				}
			}
		} else
		{
			for (final int v : this.graph.getNeighbors(vertex))
			{
				if (assignedVertices.contains(v))
				{
					i++;
				}
			}
		}
		return i;
	}

	private TreeMap<Integer, Vector<Integer>> getPossibleDegreeToVertexMapping(final DegreeIncreaseCollection coll)
	{
		// final int[] neededVertices = this.getNeededVerticesPerDegree(coll);

		final TreeMap<Integer, Vector<Integer>> degreeToPossibleVertices = new TreeMap<>();

		for (final DegreeIncrease co : coll.finalDegreeIncreases)
		{
			if (!degreeToPossibleVertices.containsKey(co.originalDegree))
			{
				final Vector<Integer> possibleVertices = new Vector<>();
				for (final int v : this.graph.getVerticesWithDegree(co.originalDegree))
				{
					possibleVertices.add(v);
				}
				degreeToPossibleVertices.put(co.originalDegree, possibleVertices);
			}
		}

		return degreeToPossibleVertices;
	}

	private boolean increaseInNumberOfJumpsPossible()
	{
		for (int i = 0; i < this.numberOfJumps.length; i++)
		{
			if (this.numberOfJumps[i] < this.maxAmountOfJumps[this.jumpDegreePosition[i]])
			{
				this.numberOfJumps[i]++;
				return true;
			}
		}
		return false;
	}

	/**
	 * Either returns edges realizing the given solution sequence or a {@code CannotRealize} exception.
	 * 
	 * @return
	 * @throws CannotRealize
	 */
	private Iterable<IntTupel> isRealizable() throws CannotRealize
	{
		DegreeIncreaseCollection degreeIncreaseCollection = this.getInitialDegreeIncreaseList(this.initialSeq, this.solutionSeq);

		final TreeMap<Integer, Vector<Integer>> degreeToPossibleVertices = this.getPossibleDegreeToVertexMapping(degreeIncreaseCollection);

		for (int outerTrials = 1; outerTrials <= this.config.outerIter; outerTrials++)
		{
			for (int innerTrials = 1; innerTrials <= (this.config.innerIter + (outerTrials == 1 ? this.config.innerLoopMostPromising : 0)); innerTrials++)
			{
				this.config.increaseExecutedIterations();

				final DegreeIncreaseCollection coll = degreeIncreaseCollection.clone();
				this.getDegreeVertexMapping(coll, degreeToPossibleVertices);

				try
				{
					final Iterable<IntTupel> realizeMapping = this.realizeMapping(coll.finalDegreeIncreases, innerTrials * outerTrials);

					if (GraphRealizer.log > 0)
					{
						System.out.print("sucess realize:" + outerTrials + " (inner:" + innerTrials + ")\n");
					}
					this.config.setExecutedLastIterationTime();
					return realizeMapping;
				} catch (final CannotRealize e)
				{
					this.config.setExecutedLastIterationTime();
				}
			}

			// update degree inc list (adaptive includes jumps)
			degreeIncreaseCollection = this.getAdaptDegreeIncList(degreeIncreaseCollection);
		}
		throw new CannotRealize();
	}

	private boolean isRealizable(final DegreeIncreaseCollection iniColl)
	{
		int[] intArray = new int[10];
		for (final DegreeIncrease deg : iniColl.finalDegreeIncreases)
		{
			if (deg.degreeIncrease >= intArray.length)
			{
				intArray = Arrays.copyOf(intArray, deg.degreeIncrease + 1);
			}
			intArray[deg.degreeIncrease]++;
		}
		return new CompactDegreeSequence(intArray, false).isRealizable();
	}

	private boolean possibleToInsertEdge(final AbstractEdgeContainer newEdges, final int vertex1, final int vertex2)
	{
		return (vertex1 != vertex2) && !newEdges.contains(new IntTupel(vertex1, vertex2)) && !this.graph.isAdjacent(vertex1, vertex2);
	}

	private Iterable<IntTupel> realizeMapping(final Vector<DegreeIncrease> finalDegreeIncreases, final int iter) throws CannotRealize
	{
		this.reorganizeVertexIncreases(finalDegreeIncreases, iter);

		final AbstractEdgeContainer newEdges = new BitSetBasedEdgeContainer(this.graph.getNumberOfVertices() + 1);
		// final AbstractEdgeContainer newEdges = AbstractEdgeContainer.getHashSetBasedEdgeContainer();

		int reorgSimpleCounter = 0;
		int reorgAdvCounter = 0;
		final int numberOfIncreaseVertices = finalDegreeIncreases.size();
		for (int i = 0; i < numberOfIncreaseVertices; i++)
		{
			final DegreeIncrease e1 = finalDegreeIncreases.get(i);
			if (e1.degreeIncrease > 0)
			{
				for (int j = i + 1; j < numberOfIncreaseVertices; j++)
				{
					final DegreeIncrease e2 = finalDegreeIncreases.get(j);
					if ((e2.degreeIncrease > 0) && this.possibleToInsertEdge(newEdges, e1.vertex, e2.vertex))
					{
						e1.degreeIncrease--;
						e2.degreeIncrease--;
						newEdges.addEdges(new IntTupel(e1.vertex, e2.vertex));
						if (e1.degreeIncrease == 0)
						{
							break;
						}
					}
				}

				while (e1.degreeIncrease > 0)
				{
					if (((reorgSimpleCounter++) >= this.config.maxSimpleReorg) || !this.simpleReorganizeSolution(e1, finalDegreeIncreases, newEdges))
					{// either no further simple reorg possible or not successfull
						if (((reorgAdvCounter++) >= this.config.maxAdvReorg) || !this.advancedReorganizeSolution(e1, finalDegreeIncreases, newEdges))
						{// either no further advances reorg possible or not successfull
							break;
						}
					}
				}
				if (e1.degreeIncrease > 0)
					throw new CannotRealize();
			}
		}

		return newEdges.getAllEdges();
	}

	private void reorganizeVertexIncreases(final Vector<DegreeIncrease> finalDegreeIncreases, final int iter)
	{
		// reorganize degree sequence
		final Comparator<? super DegreeIncrease> d = new Comparator<DegreeIncrease>()
		{

			@Override
			public int compare(final DegreeIncrease o1, final DegreeIncrease o2)
			{
				return o2.degreeIncrease - o1.degreeIncrease;
			}

		};
		// natural ordering is decreasing degree
		Collections.sort(finalDegreeIncreases, d);

		if (iter > 1)
		{
			ArrayS.shuffleEqualSortKeyElements(finalDegreeIncreases, d);
		}

		if (iter > this.nonShuffleIter)
		{
			final float per = 0.1f / ((this.config.getMaxIter() - iter) + 1);
			ArrayS.shuffleWithinSmallRange(finalDegreeIncreases, per);
		}

	}

	private boolean simpleReorganizeSolution(final DegreeIncrease e1, final Vector<DegreeIncrease> degreeIncreases,
			final AbstractEdgeContainer newEdges) throws CannotRealize
	{
		this.config.increaseExecutedSimpleReorganizeSolution();

		boolean change = false;
		if (e1.degreeIncrease >= 2)
		{
			final HashSet<IntTupel> addEdges = new HashSet<>();
			final Vector<IntTupel> removeEdges = new Vector<>();

			for (final IntTupel tup : newEdges.getAllEdges())
			{
				if (this.possibleToInsertEdge(newEdges, e1.vertex, tup.getFirst()) && !addEdges.contains(new IntTupel(e1.vertex, tup.getFirst()))
						&& this.possibleToInsertEdge(newEdges, e1.vertex, tup.getSecond())
						&& !addEdges.contains(new IntTupel(e1.vertex, tup.getSecond())))
				{
					removeEdges.add(tup);
					addEdges.add(new IntTupel(e1.vertex, tup.getFirst()));
					addEdges.add(new IntTupel(e1.vertex, tup.getSecond()));
					e1.degreeIncrease -= 2;
					// return true;
					change = true;
					if (e1.degreeIncrease < 2)
					{
						break;
					}
				}
			}
			newEdges.removeAll(removeEdges);
			newEdges.addAll(addEdges);
		}

		this.config.setExecutedSimpleReorganizationTime();
		return change;

	}

}
