/*
 * Implementation of the FKDA algorithm
 * Published in: Fast Identity Anonymization on Graphs
 * Authors: Xuesong Lu, Yi Song, and Stéphane Bressan
 */

package benchmarks;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

import algorithm.dataStructures.GraphInterface;
import dataStructures.IntTupel;

public class FKDA
{

	public enum Heuristic
	{
		BACKWARD, FORWARD, RANDOM
	}

	private GraphInterface graph;

	private Heuristic heur; // heuristic to choose in edge_creation: either FKDA.Heuristic.FORWARD, FKDA.Heuristic.BACKWARD or FKDA.Heuristic.RANDOM
	private int k; // value of anonymity, has to be positive

	private int n;

	private Set<IntTupel> result;

	private int trials; // number of times we try to find a suitable edge using the RANDOM edge_creation heuristic
						// default = n, number of nodes, has to be positive

	private NodeDegree[] v;

	public FKDA()
	{
	}

	private void add_edge(final int i, final int j)
	{
		this.result.add(new IntTupel(this.v[i].vertex, this.v[j].vertex));
		this.v[i].degree++;
		this.v[j].degree++;
	}

	/*
	 * Backward heuristic: one of three heuristics mentioned in section 3.2 under "The edge_creation Algorithm" examines the candidate vertices in
	 * ascending order of their degrees from v_n to v_j+1 to use, choose: Heuristic.BACKWARD output: index: j or -1 if there is no suitable edge (v_j,
	 * v_l)
	 */
	private int backward_edge(final int i, final int j)
	{
		int l = this.n;
		for (; l > (j + 1); l--)
		{
			if (!this.edge_exists(this.v[j].vertex, this.v[l].vertex) && (this.v[l].degree < this.v[i].degree))
				return l;
		}
		return -1;
	}

	/*
	 * Implements Algorithm 2: edge_creation algorithm see section 3.2, page 287 input: index: i number of vertices that are going to be anonymized
	 * starting from vi: na output: index: j
	 */
	private int edge_creation(final int i, final int na)
	{
		int j, l;

		for (j = i + 1; j <= ((i + na) - 1); j++)
		{
			while (this.v[j].degree < this.v[i].degree)
			{
				switch (this.heur)
				{
				case BACKWARD:
				{
					l = this.backward_edge(i, j);
					break;
				}
				case RANDOM:
				{
					l = this.random_edge(i, j);
					break;
				}
				default:
				{
					l = this.forward_edge(i, j);
					break;
				}
				}

				if (l == -1)
					return j;
				this.add_edge(j, l);
			}
		}
		Arrays.sort(this.v);
		return j;
	}

	private boolean edge_exists(final int node1, final int node2)
	{ // takes vertices, not indices!
		return this.graph.isAdjacent(node1, node2) || this.result.contains(new IntTupel(node1, node2));
	}

	/*
	 * Implements Algorithm 4: The Fast K-Degree Anonymization Algorithm see section 3.4, page 288-289 input: graph value of anonymity: k heuristic
	 * for edge_creation: heur (choices: Heuristic.FORWARD, Heuristic.BACKWARD, Heuristic.RANDOM) output: a k-degree anonymized graph
	 */

	public Set<IntTupel> fkda(final GraphInterface graph, final int k, final Heuristic heur)
	{
		return this.fkda(graph, k, heur, -1);
	}

	public Set<IntTupel> fkda(final GraphInterface graph, final int k, final Heuristic heur, final int trials)
	{
		this.init(graph, k, heur, trials);
		int i = 0;
		int na, l, j;
		while (i < this.n)
		{
			na = this.greedy_examination(i);
			j = this.edge_creation(i, na);
			if (j == (i + na))
			{
				i = i + na;
			} else
			{
				l = this.relaxed_edge_creation(i, j);
				if (l < i)
				{
					i = 0;
				}
			}
		}

		return this.result;
	}

	/*
	 * Forward heuristic: one of three heuristics mentioned in section 3.2 under "The edge_creation Algorithm" examines the candidate vertices in
	 * descending order of their degrees from v_j+1 to v_n to use, choose: Heuristic.FORWARD output: index: j or -1 if there is no suitable edge (v_j,
	 * v_l)
	 */
	private int forward_edge(final int i, final int j)
	{
		for (int l = j + 1; l < this.n; l++)
		{
			if (!this.edge_exists(this.v[j].vertex, this.v[l].vertex) && (this.v[l].degree < this.v[i].degree))
				return l;
		}
		return -1;
	}

	/*
	 * implements Algorithm 1: greedy_examination see section 3.1, page 286 input: index: i output: number of consecutive vertices that are anonymized
	 */
	private int greedy_examination(final int i)
	{
		int j = i + 1;
		for (; j < this.n; j++)
		{
			if (this.v[j].degree < this.v[i].degree)
			{
				if ((i > 0) && (this.v[i].degree == this.v[i - 1].degree))
				{
					if (!(((this.n - j) + 1) < this.k))
						return j - i;
				} else
				{
					if (!((((this.n - i) + 1) < (2 * this.k)) || (((this.n - j) + 1) < this.k)))
						return Math.max(this.k, j - i);
				}
				break;
			}
		}
		return (this.n - i) + 1;
	}

	/*
	 * setting initial values
	 */
	private void init(final GraphInterface graph, final int k, final Heuristic heur, final int trials)
	{
		this.graph = graph;
		this.set_v();
		this.n = graph.getNumberOfVertices() - 1;
		this.result = new HashSet<IntTupel>();
		if (trials == -1)
		{
			this.trials = this.n;
		} else
		{
			this.trials = trials;
		}
		this.heur = heur;
		this.k = k;
	}

	/*
	 * Random heuristic: one of three heuristics mentioned in section 3.2 under "The edge_creation Algorithm" selects a candidate at random to use,
	 * choose: Heuristic.BACKWARD output: index: j or -1 if there is no suitable edge (v_j, v_l)
	 */
	private int random_edge(final int i, final int j)
	{
		final Random rand = new Random();
		int l;
		int k = 0;
		do
		{
			l = rand.nextInt(this.n + 1);
			k++;
		} while ((k < this.trials) && ((l == j) || this.edge_exists(this.v[j].vertex, this.v[l].vertex) || (this.v[l].degree >= this.v[i].degree)));
		if ((l != j) && !this.edge_exists(this.v[j].vertex, this.v[l].vertex) && (this.v[l].degree < this.v[i].degree))
			return l;
		return -1;
	}

	/*
	 * Implements Algorithm 3: relaxed_edge_creation see section 3.3, page 288 tries to create a new edge (v_j, v_l) input: index: i index: j output:
	 * index: l or -1 if no edge was created
	 */

	private int relaxed_edge_creation(final int i, final int j)
	{
		int l;
		int dvl, dvi;
		for (l = this.n; l > 0; l--)
		{
			if (j == l)
			{
				continue; // avoid self loops
			}
			if (!this.edge_exists(this.v[j].vertex, this.v[l].vertex))
			{
				this.add_edge(j, l);
				dvl = this.v[l].degree;
				dvi = this.v[i].degree;
				if (this.v[j].degree == this.v[i].degree)
				{
					Arrays.sort(this.v);
					if (dvl > dvi)
						return -1;
					else
						return l;
				}
			}
		}
		return l;
	}

	// setter methods

	public void set_trials(final int t)
	{
		if (t > 0)
		{
			this.trials = t;
		}
	}

	private void set_v()
	{
		this.v = new NodeDegree[this.graph.getNumberOfVertices()];
		for (int i = 0; i < this.graph.getNumberOfVertices(); i++)
		{
			this.v[i] = new NodeDegree(i, this.graph.getDegree(i));
		}
		Arrays.sort(this.v);
	}
};

class NodeDegree implements Comparable<NodeDegree>
{
	int degree;

	int vertex;

	public NodeDegree(final int vertex, final int degree)
	{
		this.vertex = vertex;
		this.degree = degree;
	}

	// flipped for descending order
	@Override
	public int compareTo(final NodeDegree o)
	{
		if (this.degree > o.degree)
			return -1;
		else if (this.degree < o.degree)
			return 1;
		return 0;
	}

	@Override
	public String toString()
	{
		return "NodeDegree [vertex=" + this.vertex + ", degree=" + this.degree + "]";
	}
}