package reader;

import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import org.apache.commons.collections15.map.HashedMap;

import algorithm.dataStructures.AdjacentGraph;
import algorithm.dataStructures.Edge;
import algorithm.dataStructures.GraphInterface;

public class GraphReader
{

	File file;

	private String filename;

	InputStream is = null;

	public GraphReader(final InputStream is)
	{
		this.is = is;
	}

	public GraphReader(final String filename)
	{
		this.file = new File(filename);
	}

	public GraphReader(final URL filestream)
	{
		this.file = new File(filestream.getFile());
	}

	public String getFilename()
	{
		return this.filename;
	}

	/**
	 * Checks whether the input graph is symmetric.
	 * 
	 * @param graph graph instance
	 * @return true if graph is symmetric, false otherwise
	 */
	private boolean inputFileIsValid(final GraphInterface graph)
	{

		for (int i = 1; i < graph.getNumberOfVertices(); i++)
		{
			final int neighbors[] = graph.getNeighbors(i);
			for (final int neighbor : neighbors)
			{
				if (!graph.isAdjacent(neighbor, i))
					return false; // graph not symmetric

			}
		}

		return true;

	}

	/**
	 * Reads graph text files of DIMACs graph challenge. First line after comments is the header, containing information about amount of vertices and
	 * edges. Each following line i contains all adjacent vertices of vertex i.
	 * 
	 * @return Adjacent graph containing all edges of input file
	 */
	private GraphInterface readFileAsListOfAdjacencies()
	{
		List<TIntSet> tGraphSet = null;

		BufferedReader in = null;
		try
		{
			if (this.is == null)
			{
				in = new BufferedReader(new FileReader(this.file));
			} else
			{
				in = new BufferedReader(new InputStreamReader(this.is));
			}

			// Include header information (|V|, |E|)
			String zeile = null;
			do
			{
				zeile = in.readLine();
				if (zeile == null)
					return null;
			} while ((zeile.startsWith("%") || zeile.startsWith("#")));

			// Header
			if (!(Pattern.matches("\\d+\\s+\\d+", zeile.trim()) || Pattern.matches("\\d+\\s+\\d+\\s+\\d+", zeile.trim())))
			{
				System.out.println("File does not contain a header. Aborting");
				return null;
			}
			final String header[] = zeile.split(" ");
			tGraphSet = new ArrayList<TIntSet>(Integer.valueOf(header[0]));

			while ((zeile = in.readLine()) != null)
			{
				zeile = zeile.trim();
				if (Pattern.matches("[\\d\\s]*", zeile))
				{
					final TIntHashSet tNodeSet = new TIntHashSet();
					if (!zeile.isEmpty())
					{
						zeile = zeile.trim();
						final String nodes[] = zeile.split("\\s+");
						for (final String node : nodes)
						{
							final Integer nodeInteger = Integer.valueOf(node) - 1;
							if (nodeInteger != tGraphSet.size())
							{ // Exclude reflexive edges
								tNodeSet.add(nodeInteger);
							}
						}
					}
					tGraphSet.add(tNodeSet);
				} else if (!zeile.startsWith("%") && !zeile.startsWith("#"))
					return null;

			}
		} catch (final Exception e)
		{
			System.out.println("#Error occured while reading input graph.");
			e.printStackTrace();
			return null;
		} finally
		{
			try
			{
				in.close();
			} catch (final IOException e)
			{
				e.printStackTrace();
			}
		}

		return new AdjacentGraph(tGraphSet);
	}

	/**
	 * Reads input file as a list of edges: Each line contains two integers representing two vertices, separated by whitespace.
	 * 
	 * @return AdjacentGraph
	 */
	private GraphInterface readFileAsListOfEdges()
	{

		GraphInterface graph;

		List<TIntSet> tGraphSet = null;
		final List<Edge> edges = new LinkedList<>();
		BufferedReader in = null;
		final Map<Integer, String> indexToName = new HashedMap<Integer, String>();
		try
		{
			if (this.is == null)
			{
				in = new BufferedReader(new FileReader(this.file));
			} else
			{
				in = new BufferedReader(new InputStreamReader(this.is));
			}

			String zeile = null;

			int maxVerticeNumber = -1;
			final Map<String, Integer> nameToIndex = new HashedMap<String, Integer>();

			int index = 0;

			while ((zeile = in.readLine()) != null)
			{
				if (Pattern.matches("\\d+\\s+\\d+", zeile))
				{
					final String vertices[] = zeile.split("\\s");
					int a;
					int b;
					if (nameToIndex.containsKey(vertices[0]))
					{
						a = nameToIndex.get(vertices[0]);
					} else
					{
						a = index++;
						nameToIndex.put(vertices[0], a);
						indexToName.put(a, vertices[0]);
					}
					if (nameToIndex.containsKey(vertices[1]))
					{
						b = nameToIndex.get(vertices[1]);
					} else
					{
						b = index++;
						nameToIndex.put(vertices[1], b);
						indexToName.put(b, vertices[1]);
					}

					if (a != b)
					{
						edges.add(new Edge(a, b));
						maxVerticeNumber = Math.max(maxVerticeNumber, a);
						maxVerticeNumber = Math.max(maxVerticeNumber, b);
					}

				} else if (!zeile.startsWith("%") && !zeile.startsWith("#"))
					return null;

			}

			tGraphSet = new ArrayList<TIntSet>(Integer.valueOf(maxVerticeNumber));
			for (int i = 0; i <= maxVerticeNumber; i++)
			{
				tGraphSet.add(new TIntHashSet());
			}

			for (final Edge e : edges)
			{

				tGraphSet.get(e.getVertexA()).add(e.getVertexB());

				tGraphSet.get(e.getVertexB()).add(e.getVertexA());
			}
			graph = new AdjacentGraph(tGraphSet, indexToName);

		} catch (final Exception e)
		{
			System.out.println("#Error occured while reading input graph.");
			e.printStackTrace();
			return null;
		} finally
		{
			try
			{
				in.close();
			} catch (final IOException e)
			{
				e.printStackTrace();
			}
		}

		return graph;

	}

	/**
	 * Reads the graph input text file. Handles three different formats (each line contains an edge, each line i contains adjacent vertices of vertex
	 * i, each line contains an edge and an integer -1 or 1.
	 * 
	 * @return Adjacent graph containing all edges of input file
	 */
	public GraphInterface readGraphFile()
	{
		GraphInterface graph;

		if ((graph = this.readTripleFormat()) != null)
		{
			if (!this.inputFileIsValid(graph))
			{
				System.out.println("Graph input file is not symmetric. Aborting");
				return null;
			}
			return graph;
		}

		if ((graph = this.readFileAsListOfEdges()) != null)
		{
			if (!this.inputFileIsValid(graph))
			{
				System.out.println("Graph input file is not symmetric. Aborting");
				return null;
			}
			return graph;
		}

		if ((graph = this.readFileAsListOfAdjacencies()) != null)
		{
			if (!this.inputFileIsValid(graph))
			{
				System.out.println("Graph input file is not symmetric. Aborting");
				return null;
			}
			return graph;
		}

		return graph;
	}

	private GraphInterface readTripleFormat()
	{
		List<TIntSet> tGraphSet = null;
		final List<Edge> edges = new LinkedList<>();
		BufferedReader in = null;
		final Map<Integer, String> indexToName = new HashedMap<Integer, String>();
		try
		{
			if (this.is == null)
			{
				in = new BufferedReader(new FileReader(this.file));
			} else
			{
				in = new BufferedReader(new InputStreamReader(this.is));
			}

			String zeile = null;

			int maxVerticeNumber = -1;
			final Map<String, Integer> nameToIndex = new HashedMap<String, Integer>();

			int index = 0;
			while ((zeile = in.readLine()) != null)
			{
				if (Pattern.matches("\\d+\\s+\\d+\\s+-*\\d+", zeile))
				{
					final String vertices[] = zeile.split("\\s");
					int a;
					int b;
					if (nameToIndex.containsKey(vertices[0]))
					{
						a = nameToIndex.get(vertices[0]);
					} else
					{
						a = index++;
						nameToIndex.put(vertices[0], a);
						indexToName.put(a, vertices[0]);
					}
					if (nameToIndex.containsKey(vertices[1]))
					{
						b = nameToIndex.get(vertices[1]);
					} else
					{
						b = index++;
						nameToIndex.put(vertices[1], b);
						indexToName.put(b, vertices[1]);
					}

					if (a != b) // if a > b then we have already parsed the edge due to symmetry
					{
						edges.add(new Edge(a, b));
						maxVerticeNumber = Math.max(maxVerticeNumber, a);
						maxVerticeNumber = Math.max(maxVerticeNumber, b);
					}

				} else if (!Pattern.matches("%|#.*", zeile))
					return null;

			}

			tGraphSet = new ArrayList<TIntSet>(Integer.valueOf(maxVerticeNumber));
			for (int i = 0; i <= maxVerticeNumber; i++)
			{
				tGraphSet.add(new TIntHashSet());
			}

			for (final Edge e : edges)
			{
				tGraphSet.get(e.getVertexA()).add(e.getVertexB());
			}

		} catch (final Exception e)
		{
			System.out.println("#Error occured while reading input graph.");
			e.printStackTrace();
			return null;
		} finally
		{
			try
			{
				in.close();
			} catch (final IOException e)
			{
				e.printStackTrace();
			}
		}

		return new AdjacentGraph(tGraphSet, indexToName);
	}

}
